-- $Id: bases.lua 4252 2021-07-31 23:22:57Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2021 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

-- Base object entities for the auto-generated object entities in
-- data/general/objects/*.lua.  This file is not auto-generated, so it's
-- okay to edit it by hand.

local ActorTalents = require 'engine.interface.ActorTalents'

newEntity {
  define_as = 'ESSENCE_BASE',
  type = 'misc',
  subtype = 'alchemic',
  tval = 4,
  base_tmpl = '& Essence~ of #',
  display = '*',
  color = colors.C_G,
  level = 0,
  weight = 0.4,
  encumber = 0.4,
  cost = 2,
  stacking = true,
  check_stack_ok_known = true,
  flags = { EASY_KNOW=true },
  category = 'magic',
}

newEntity {
  define_as = 'CHEST_BASE',
  type = 'misc',
  subtype = 'chest',
  tval = 7,
  display = '~',
  color = colors.C_s,
  name_cb = function(self, name, t)
    if not self:isIDd() then return name end
    if self.chest.trap then
      if self.chest.trap:knownBy(game.player) then
	return ('%s (%s)'):format(name, self.chest.trap:getName())
      else
	return name .. ' (trapped)'
      end
    elseif self.chest.n_obj > 0 then
      return name .. ' (disarmed)'
    else
      return name .. ' (empty)'
    end
  end,
  category = 'treasure',
  destroyed_by = { [DamageType.FIRE] = true, [DamageType.ACID] = true },
  apply_damtype = function(self, who, x, y, dt, dam)
    local DamageType = require 'engine.DamageType'
    if damtype == DamageType.KILL_TRAP or dt == DamageType.KILL_DOOR then
      if self.chest and self.chest.trap then
	if self.chest.trap.removed then
	  self.chest.trap:removed(x, y, who)
	end
	self.chest.trap = nil
	self:identify(true, who)
	if self.map_remember then game.logPlayer(who, 'Click!') end
      end
    end
  end,
  on_identify = function(self, who)
    if self.chest and self.chest.trap then self.chest.trap:identify(true) end
  end,
  known_trap = function(self, x, y, who)
    -- Slight Hack(TM):  Juggle args so that this can be called either in
    -- or out of a Map:checkAllEntities() context.
    if not y and not who then who = x end
    return self.chest and self.chest.trap and self.chest.trap:knownBy(who)
  end,
  is_openable = function(self, x, y)
    -- Chest can be opened if it's not empty.
    return self.chest and self.chest.n_obj > 0
  end,
  try_open = function(self, x, y, who)
    if not self.chest then return false, false, 'malformed chest?' end
    -- If no trap, we always succeed.
    local success = not self.chest.trap
    if not success then
      -- Otherwise, try to disarm and unlock the chest.
      -- [cf. do_cmd_open_chest() in src/cmd2.c]
      if not who.skill_disarm then return false, false, 'no disarming skill?' end
      -- Oddly, this depends on the trap index rather than the trap's
      -- difficulty, as used elsewhere.
      local chance = math.max(who:getDisarmPower() - self.chest.trap.idx, 2)
      success = rng.percent(chance)
      if success then
	game.logPlayer(who, 'You have picked the lock.')
	who:gainExp(1)
      end
    end
    if success then
      if self.chest.trap then
	-- Trigger the trap.
	self.chest.trap:trigger(x, y, who, true)
      end
      -- Generate objects.
      local lev = game.zone:level_adjust_level(game.level, 'object')
      local o
      local Object = require 'mod.class.Object'
      for _ = 1, self.chest.n_obj do
	-- Small chests mostly have gold.
	if self.chest.small and rng.percent(75) then
	  local coin_type = obj_util.choose_coin_type(lev)
	  o = game.zone:makeEntityByName(game.level, 'object', coin_type)
	else
	  -- Slight Hack(TM):  No chests in chests.
	  Object.resolve_flags.force_no_chest = true
	  o = game.zone:makeEntity(game.level, 'object', nil, lev, true)
	end
	if o then
	  local dx, dy = util.drop_loc_near(x, y, o.unique)
	  if dx and dy then game.level.map:addObject(dx, dy, o) end
	end
      end
      -- Mark the chest as empty
      self.chest.trap = nil
      self.chest.n_obj = 0
      return true, false, nil
    else
      return false, true, nil
    end
  end,
}

newEntity {
  define_as = 'PARCHMENT_BASE',
  type = 'misc',
  subtype = 'parchment',
  tval = 8,
  base_tmpl = '& Parchment~ - #',
  display = '?',
  weight = 0.5,
  encumber = 0.5,
  stacking = true,
  category = 'magic',
  use_simple = {
    titles = { 'Read', 'read' },
    use = function(self, who, inven, floor, item)
      if self.use_simple.town_map then
	game:onLevelLoad('worldmap-1', obj_util.parchment_map_town, self.use_simple.town_map)
	local z_inf = game.state:getZoneInfo(self.use_simple.town_map.zone)
	if z_inf then
	  game.logPlayer(who, [[Based on this map, you now know where the town of %s is.]], z_inf.name)
	else
	  game.logPlayer(who, [[You're not sure what town this map is showing, but you're pretty sure you can get there now.]])
	end
      elseif self.use_simple.lore then
	if who.learnLore then
	  who:learnLore(self.use_simple.lore)
	else
	  game.log('[??? actor %s using lore %s]', who:getName(), self:getName{no_count=true})
	end
      else
	game.log('[Sorry, not implemented (%s)]', self:getName{no_count=true})
      end
      return { id=true, used=true, destroy=true }
    end,
    name = function(self)
      local desc = 'do something unspecified'
      if self.use_simple.town_map then
	desc = 'reveal the location of a town and its surrounding geography on the world map'
      elseif self.use_simple.lore then
	desc = 'learn the lore on the parchment'
      end
      return desc .. obj_util.consume_desc_suffix(self, 'parchment')
    end,
  },
  cant_use_if = { blind=true, confused=true, unlit=true },
}

newEntity {
  define_as = 'CORPSE_BASE',
  type = 'misc',
  subtype = 'corpse',
  tval = 9,
  display = '~',
  color = colors.C_U,
  level = 20,
  cost = 0,
  flags = { EASY_KNOW=true },
  category = 'junk',
}

newEntity {
  define_as = 'INSTRUMENT_BASE',
  type = 'equipment',
  subtype = 'instrument',
  tval = 14,
  slot = 'LAUNCHER',
  display = '/',
  color = colors.C_W,
  level = 7,
  weight = 3,
  encumber = 3,
  cost = 400,
  category = 'tools',
  resolvers.ego(),
}

newEntity {
  define_as = 'WEAPON_NAME_BASE',
  name_cb_post_ego = function(self, name, idd, t)
    if t.no_mods then return name end
    return name .. string.format(' (%dd%d)', self.combat[1], self.combat[2])
  end,
  category = 'combat',
  flags = { SHOW_MODS=true },
  resolvers.weapon(),
}

newEntity {
  base = 'WEAPON_NAME_BASE',
  define_as = 'BOOMERANG_BASE',
  type = 'throwing',
  subtype = 'boomerang',
  tval = 15,
  slot = 'LAUNCHER',
  display = '{',
  color = colors.C_y,
  weight = 6,
  encumber = 6,
  to_h = 0,
  to_d = 0,
  stacking = true,
  check_stack = function(self, o)
    -- Require both known.
    if not self:known() or not o:known() then return false end
    if not util.weapons_can_stack(self, o) then return false end
    return config.settings.tome2.stack_wpn_armor
  end,
  worn_talents = { ActorTalents.T_THROW_BOOMERANG },
}

newEntity {
  base = 'WEAPON_NAME_BASE',
  define_as = 'AMMO_BASE',
  type = 'ammo',
  slot = 'QUIVER',
  display = '{',
  name_cb_post_ego = function(self, name, idd, t)
    -- Slight Hack(TM):  We modify the parent callback to add the
    -- "(exploding)" note as necessary.
    if self.explode_dam_type then name = name .. ' (exploding)' end
    return name .. string.format(' (%dd%d)', self.combat[1], self.combat[2])
  end,
  stacking = true,
  check_stack = function(self, o)
    -- Require both known or both not known.
    if self:known() and not o:known() then return false end
    if not self:known() and o:known() then return false end
    if not util.weapons_can_stack(self, o) then return false end
    if self.explode_dam_type ~= o.explode_dam_type then return false end
    return true
  end,
  base_break_chance = 25,
  resolvers.ammo(),
}

newEntity {
  base = 'AMMO_BASE',
  define_as = 'ARROW_BASE',
  subtype = 'arrows',
  tval = 17,
  weight = 0.2, -- override as necessary
  encumber = 0.2,
  base_break_chance = 50,
  destroyed_by = {
    [DamageType.ACID] = true,
    [DamageType.FIRE] = true,
  },
}

newEntity {
  base = 'AMMO_BASE',
  define_as = 'BOLT_BASE',
  subtype = 'bolts',
  tval = 18,
  weight = 0.3, -- override as necessary
  encumber = 0.3,
  destroyed_by = { [DamageType.ACID] = true },
}

newEntity {
  base = 'AMMO_BASE',
  define_as = 'SHOT_BASE',
  subtype = 'shots',
  tval = 16,
  weight = 0.4,
  encumber = 0.4,
}

newEntity {
  define_as = 'LAUNCHER_BASE',
  type = 'shooter',
  slot = 'LAUNCHER',
  display = '}',
  tval = 19,
  to_h = 0,
  to_d = 0,
  stacking = true,
  name_cb_post_ego = function(self, name, idd, t)
    return name .. string.format(' (x%d)', self:ammoMult())
  end,
  check_stack = function(self, o)
    -- Require both known.
    if not self:known() or not o:known() then return false end
    if self.ammo_mult ~= o.ammo_mult then return false end
    return config.settings.tome2.stack_wpn_armor
  end,
  flags = { SHOW_MODS=true },
  category = 'combat',
  destroyed_by = {
    [DamageType.ACID] = true,
    [DamageType.FIRE] = true,
  },
  resolvers.weapon(),
}

newEntity {
  base = 'LAUNCHER_BASE',
  define_as = 'BOW_BASE',
  subtype = 'bow',
  color = colors.C_U,
  ammo_type = 'arrows',
  destroyed_by = { [DamageType.FIRE] = true, },
  worn_talents = { ActorTalents.T_FIRE_ARROW },
}

newEntity {
  base = 'LAUNCHER_BASE',
  define_as = 'CROSSBOW_BASE',
  subtype = 'crossbow',
  color = colors.C_s,
  ammo_type = 'bolts',
  worn_talents = { ActorTalents.T_FIRE_BOLT },
}

newEntity {
  define_as = 'TOOL_BASE',
  type = 'tool',
  slot = 'TOOL',
  category = 'tools',
  resolvers.ego(),
}

newEntity {
  base = 'TOOL_BASE',
  define_as = 'DIGGER_BASE',
  subtype = 'digger',
  tval = 20,
  display = '\\',
  stacking = true,
  check_stack = function(self, o)
    -- Require both known.
    if not self:known() or not o:known() then return false end
    if not util.weapons_can_stack(self, o) then return false end
    return config.settings.tome2.stack_wpn_armor
  end,
  category = 'tools',
  worn_talents = { ActorTalents.T_OBJ_DIG },
}

newEntity {
  base = 'WEAPON_NAME_BASE',
  define_as = 'WEAPON_BASE',
  type = 'weapon',
  slot = 'MAIN_HAND',
  to_h = 0,
  to_d = 0,
  stacking = true,
  check_stack = function(self, o)
    -- Require both known.
    if not self:known() or not o:known() then return false end
    if not util.weapons_can_stack(self, o) then return false end
    return config.settings.tome2.stack_wpn_armor
  end,
  destroyed_by = { [DamageType.ACID] = true, },
}

newEntity {
  base = 'WEAPON_BASE',
  define_as = 'HAFTED_BASE',
  subtype = 'hafted',
  tval = 21,
  display = '\\',
  destroyed_by = { [DamageType.FIRE] = true, },
}

-- Slight Hack(TM):  Separate base for two-handed hafted weapons that sets
-- the 'slot_forbid' field appropriately.
newEntity {
  base = 'HAFTED_BASE',
  define_as = 'HAFTED_2H_BASE',
  slot_forbid = 'OFF_HAND',
}

newEntity {
  base = 'WEAPON_BASE',
  define_as = 'POLEARM_BASE',
  subtype = 'polearm',
  tval = 22,
  display = '/',
  edged_weapon = true,
  destroyed_by = { [DamageType.FIRE] = true, },
}

-- Slight Hack(TM):  As above.
newEntity {
  base = 'POLEARM_BASE',
  define_as = 'POLEARM_2H_BASE',
  slot_forbid = 'OFF_HAND',
}

newEntity {
  base = 'WEAPON_BASE',
  define_as = 'SWORD_BASE',
  subtype = 'sword',
  tval = 23,
  display = '|',
  edged_weapon = true,
}

-- Slight Hack(TM):  As above.
newEntity {
  base = 'SWORD_BASE',
  define_as = 'SWORD_2H_BASE',
  slot_forbid = 'OFF_HAND',
}

newEntity {
  base = 'WEAPON_BASE',
  define_as = 'AXE_BASE',
  subtype = 'axe',
  tval = 24,
  display = '/',
  color = colors.C_s,
  edged_weapon = true,
}

-- Slight Hack(TM):  As above.
newEntity {
  base = 'AXE_BASE',
  define_as = 'AXE_2H_BASE',
  slot_forbid = 'OFF_HAND',
}

newEntity {
  define_as = 'ARMOR_BASE',
  type = 'armor',
  armor = 0,
  to_a = 0,
  stacking = true,
  check_stack = function(self, o)
    -- Require both known.
    if not self:known() or not o:known() then return false end
    -- Don't recall any egos that change base armor, but just in case
    if self.armor ~= o.armor then return false end
    return config.settings.tome2.stack_wpn_armor
  end,
  category = 'combat',
  destroyed_by = { [DamageType.ACID] = true, },
  resolvers.armor(),
}

newEntity {
  base = 'ARMOR_BASE',
  define_as = 'BOOTS_BASE',
  type = 'armor',
  subtype = 'footwear',
  tval = 30,
  slot = 'BOOTS',
  display = ']',
  category = 'tools',
  destroyed_by = { [DamageType.FIRE] = true, },
}

newEntity {
  base = 'ARMOR_BASE',
  define_as = 'GLOVES_BASE',
  type = 'armor',
  subtype = 'handwear',
  tval = 31,
  slot = 'GLOVES',
  display = ']',
  destroyed_by = { [DamageType.FIRE] = true, },
}

newEntity {
  base = 'ARMOR_BASE',
  define_as = 'HELM_BASE',
  type = 'armor',
  subtype = 'headwear',
  tval = 32,
  slot = 'HELM',
  display = ']',
}

newEntity {
  base = 'HELM_BASE',
  define_as = 'CROWN_BASE',
  tval = 33,
  armor = 0,
  category = 'treasure',
}

newEntity {
  base = 'ARMOR_BASE',
  define_as = 'SHIELD_BASE',
  type = 'armor',
  subtype = 'shield',
  tval = 34,
  slot = 'OFF_HAND',
  display = ')',
}

newEntity {
  base = 'ARMOR_BASE',
  define_as = 'CLOAK_BASE',
  type = 'armor',
  subtype = 'cloak',
  tval = 35,
  slot = 'CLOAK',
  display = '(',
  destroyed_by = { [DamageType.FIRE] = true, },
  category = 'tools',
}

newEntity {
  base = 'ARMOR_BASE',
  define_as = 'BODY_ARMOR_BASE',
  slot = 'BODY',
}

newEntity {
  base = 'BODY_ARMOR_BASE',
  define_as = 'SOFT_BODY_ARMOR_BASE',
  subtype = 'light',
  tval = 36,
  display = '(',
  destroyed_by = { [DamageType.FIRE] = true, },
}

newEntity {
  base = 'BODY_ARMOR_BASE',
  define_as = 'HARD_BODY_ARMOR_BASE',
  subtype = 'heavy',
  tval = 37,
  display = '[',
}

newEntity {
  base = 'HARD_BODY_ARMOR_BASE',
  define_as = 'DRAGON_ARMOR_BASE',
  tval = 38,
  is_dragon_armor = true,
  category = 'treasure',
  rating = 30,
  resolvers.precog_mention(),
}

newEntity {
  define_as = 'LITE_BASE',
  type = 'equipment',
  subtype = 'light',
  tval = 39,
  slot = 'LITE',
  display = '~',
  name_cb_post_ego = function(self, name, idd, t)
    if self.lite_fuel and (not t or not t.no_mods) then
      name = name .. (' (%d turns)'):format(self.lite_fuel)
    end
    return name
  end,
  stacking = true,
  check_stack = function(self, o)
    -- Require both known.
    if not self:known() or not o:known() then return false end
    -- Require both fueled with same time left or neither fueled.
    if self.lite_fuel and o.lite_fuel then
      return self.lite_fuel == o.lite_fuel
    else
      return not self.lite_fuel and not o.lite_fuel
    end
  end,
  refueled_by = function(self, fuel)
    return self.lite_fuel and self.refuel == fuel
  end,
  can_refuel = function(self, fuel)
    if fuel == 'oil' then
      return self.oil and util.getval(self.oil, self) > 0
    end
    if fuel == 'torch' then
      return self.define_as == 'WOODEN_TORCH' and self.lite_fuel > 0
    end
    return false
  end,
  category = 'tools', 
  destroyed_by = { [DamageType.FIRE] = true },
  resolvers.ego(),
  resolvers.light(),
}

newEntity {
  define_as = 'AMULET_BASE',
  type = 'jewelry',
  subtype = 'amulet',
  tval = 40,
  base_tmpl = '& Amulet~ of #',
  unid_tmpl = '& # Amulet~',
  slot = 'AMULET',
  display = '"',
  weight = 0.2,
  encumber = 0.2,
  unid_cost = 45,	-- cf. object_value_base() in src/object2.c
  stacking = true,
  check_stack = function(self, o)
    -- Require both known.
    if not self:known() or not o:known() then return false end
    -- I *think* Object:canStack() covers everything we need.
    return true
  end,
  category = 'treasure',
  resolvers.flavored(),
  resolvers.ego(),
}

newEntity {
  define_as = 'RING_BASE',
  type = 'jewelry',
  subtype = 'ring',
  tval = 45,
  base_tmpl = '& Ring~ of #',
  unid_tmpl = '& # Ring~',
  slot = 'RING',
  display = '=',
  weight = 0.2,
  encumber = 0.2,
  unid_cost = 45,	-- cf. object_value_base() in src/object2.c
  stacking = true,
  check_stack = function(self, o)
    -- Require both known.
    if not self:known() or not o:known() then return false end
    -- I *think* Object:canStack() covers everything we need.
    return true
  end,
  category = 'treasure',
  destroyed_by = { [DamageType.ELEC] = true, },
  resolvers.flavored(),
  resolvers.ego(),
}

newEntity {
  define_as = 'TRAP_BASE',
  type = 'misc',
  subtype = 'trapkit',
  tval = 46,
  base_tmpl = '& # Trap Set',
  display = '`',
  stacking = true,
  -- Nothing appears to assign an AC to trapkits, but the trap handling
  -- code appears to depend on its presence...
  armor = 0,
  to_a = 0,
  to_h = 0,
  to_d = 0,
  name_cb_post_tmpl = function(self, name, t)
    if self.ammo_mult and self:isIDd() then
      name = ('%s (x%d)'):format(name, self:ammoMult())
    end
    return name
  end,
  check_stack = function(self, o)
    -- Require both known.
    if not self:known() or not o:known() then return false end
    -- Interestingly, this uses the 'Allow weapons and armor to stack'
    -- option.
    return config.settings.tome2.stack_wpn_armor
  end,
  num_shots = function(self, max)
    if self:hasFlag('pval_flags', 'MULTI_SHOT') and self.pval > 0 then
      return max and math.min(1 + self.pval, max) or 1 + self.pval
    end
    return max and math.min(1, max) or 1
  end,
  auto_reset = function(self)
    -- Always auto-reset
    if self:hasFlag('flags', 'AUTOMATIC_99') then return true end
    -- Auto-reset 80% of the time.
    if self:hasFlag('flags', 'AUTOMATIC_5') then return rng.percent(80) end
    -- Never auto-reset
    return false
  end,
  category = 'tools',
  resolvers.weapon(),
  resolvers.trapkit(),
}

newEntity {
  define_as = 'STICK_COMMON_BASE',
  type = 'utility',
  color = colors.C_d,
  name_cb_post_ego = function(self, name, idd, t)
    if not idd or t.no_mods then return name end
    if not self.use_power.base_level or not self.use_power.max_level then
      -- If we're called from the visuals dialog, we won't have these.
      return name .. ' [?|?]'
    end
    return ('%s [%d|%d]'):format(name, self.use_power.base_level, self.use_power.max_level)
  end,
  sort_order = function(a, b)
    -- Sort by increasing max_level, then by increasing base_level, then by
    -- decreasing charges.
    if a.use_power.max_level ~= b.use_power.max_level then
      return a.use_power.max_level < b.use_power.max_level
    end
    if a.use_power.base_level ~= b.use_power.base_level then
      return a.use_power.base_level < b.use_power.base_level
    end
    if a.power ~= b.power then return a.power > b.power end
    return nil
  end,
  stacking = true,
  category = 'magic',
  use_power = {
    name = function(self)
      local t = spell.stick_talent(self)
      if not t then return 'cast an unknown spell (???)' end
      return ('cast the %s spell'):format(t.name)
    end,
    power = 1,
  },
  resolvers.flavored(),
  resolvers.wand_staff(),
  resolvers.ego(),
}

newEntity {
  base = 'STICK_COMMON_BASE',
  define_as = 'STAFF_BASE',
  subtype = 'staff',
  tval = 55,
  base_tmpl = '& Staff~ of #',
  unid_tmpl = '& # Staff~',
  display = '_',
  weight = 5,
  encumber = 5,
  unid_cost = 70,	-- cf. object_value_base() in src/object2.c
  check_stack = function(self, o)
    -- Require known or known empty for both.
    if not self:known() and not self.pseudo_id_empty then return false end
    if not o:known() and not o.pseudo_id_empty then return false end
    -- Require same number of charges
    if self.power ~= o.power then return false end
    -- Require same spell and same [M|N] numbers
    if self.use_power.spell ~= o.use_power.spell then return false end
    if self.use_power.base_level ~= o.use_power.base_level then return false end
    if self.use_power.max_level ~= o.use_power.max_level then return false end
    -- That's all we needed.
    return true
  end,
  use_power = {
    use = function(stick, who, inven, floor, item)
      if who:has('ANTIMAGIC_FIELD') then
	local this_dev = stick:getNumber() > 1 and 'these devices' or 'this device'
	game.logPlayer(who, 'Your antimagic field prevents you from using %s.', this_dev)
	return nil
      end
      local unstacked = nil
      if stick:getNumber() > 1 then
	-- Unstack one staff from stack and use it.
	unstacked, _ = stick:unstack(1)
      end
      local ret = spell.cast_stick_spell(unstacked or stick, who, inven, floor, item)
      if unstacked then
	-- Put our unstacked staff back into inventory.  First, though, we
	-- have to cheat and pre-emptively use a charge on the unstacked
	-- staff, or it'll just re-combine with the stack.
	ret = ret or {}
	if ret.used and not ret.no_power then
	  unstacked.power = unstacked.power - unstacked.use_power.power
	  -- Since we already used a charge, tell the upstream caller not
	  -- to use another one.
	  ret.no_power = true
	end
	who:addOrDropObject(inven, unstacked, nil, nil, {vocal_drop=true})
	who:sortInven()
      end
      return ret
    end,
  },
  -- cf. recharge() in src/spells2.c
  recharge = {
    try = function(self, power)
      -- Caller should have filtered this out, but just in case...
      if self:hasFlag('flags', 'NO_RECHARGE') then return false end
      local strength = math.floor((100 + power - self.level - 8*self.power)/15)
      -- I think it's possible for this to be zero or less, which should
      -- probably mean instant failure.
      if (strength <= 0 or rng.chance(strength)) and not self:hasFlag('flags', 'RECHARGE') then
	-- Recharge failed.
	return false
      end
      local amount = rng.range(1, 1 + math.floor(power/(self.level+2)))
      -- Divide recharge between staffs in a stack.
      local n = self:getNumber()
      if n > 1 then amount = math.floor(math.max(1, amount/n)) end
      self:forAllStack(function(so)
	so.power = so.power + amount
	-- "Un-identify" so we don't know the new charge level; we do know,
	-- however, that it's no longer empty.
	if not self:hasFlag('flags', 'RECHARGE') then
	  so.identified = nil
	  so.pseudo_id_empty = nil
	end
      end)
      return true
    end,
    on_fail = function(self, who, inven, item, floor)
      -- 50% chance of no effect if we have Perfect Casting; otherwise blow
      -- up one staff.
      if who:has('PERFECT_CASTING') and rng.chance(2) then
	game.log('The charging fails.')
      else
	local name = self:getName{no_count=true, no_mods=true, no_pval=true, no_inscr=true}
	local your = self:getNumber() > 1 and 'one of your' or 'your'
	game.log(game.flash.BAD, 'Wild magic consumes %s %s!', your, name)
	who:getOneObject(inven, item, floor, {vocal_left=true})
      end
    end,
  },
  -- Primarily for use by Alchemists' Extract Essences talent.
  -- [cf. PlayerAlchemy:extractEssences()]
  extract_charges = function(self, n, who)
    -- Every staff in the stack has the same number of charges.
    n = math.min(n, (self.power or 0) * self:getNumber())
    if n <= 0 then return 0 end
    local nn = n
    while nn > 0 do
      local unstack = self:getNumber() > 1
      local o = unstack and self:unstack(1) or self
      local dn = math.min(nn, self.power or 0)
      nn = nn - dn
      o.power = o.power - dn
      if unstack then who:addOrDropObject(who.INVEN_INVEN, o, nil, nil, {vocal_drop=true}) end
    end
    who:sortInven()
    return n
  end,
  -- Primarily for use by Alchemists' Recharge Wand/Staff talent.
  -- [cf. PlayerAlchemy:playerRecharge
  add_charges = function(self, n, who)
    if n <= 0 then return 0 end
    -- Slight Hack(TM):  T2's inventory stacking/sorting has the effect
    -- that charges all get added to a single staff unstacked out of a
    -- stack.  We replicate that here.
    local unstack = self:getNumber() > 1
    local staff = unstack and self:unstack(1) or self
    o.power = o.power + n
    if unstack then who:addOrDropObject(who.INVEN_INVEN, staff, nil, nil, {vocal_drop=true}) end
    who:sortInven()
    return n
  end,
  destroyed_by = {
    [DamageType.ACID] = true,
    [DamageType.FIRE] = true,
  },
}

newEntity {
  base = 'STICK_COMMON_BASE',
  define_as = 'WAND_BASE',
  subtype = 'wand',
  tval = 65,
  base_tmpl = '& Wand~ of #',
  unid_tmpl = '& # Wand~',
  num_charges = function(self)
    local total = self.power
    for _, o in ipairs(self.stacked or {}) do total = total + o.power end
    return total
  end,
  charge_display = function(self)
    -- Describe stacks of wands as having (N total charges).
    if not self.stacked then
      return ('%s charge%s'):format(self.power, self.power == 1 and '' or 's')
    end
    local charges = self:num_charges()
    return (' (%d total charge%s)'):format(charges, (charges == 1 and '' or 's'))
  end,
  display = '-',
  weight = 1,
  encumber = 1,
  unid_cost = 50,	-- cf. object_value_base() in src/object2.c
  check_stack = function(self, o)
    -- Require known or known empty for both.
    if not self:known() and not self.pseudo_id_empty then return false end
    if not o:known() and not o.pseudo_id_empty then return false end
    -- Require same spell and same [M|N] numbers
    if self.use_power.spell ~= o.use_power.spell then return false end
    if self.use_power.base_level ~= o.use_power.base_level then return false end
    if self.use_power.max_level ~= o.use_power.max_level then return false end
    -- That's all we needed.
    return true
  end,
  use_power = {
    use = function(stick, who, inven, floor, item)
      if who:has('ANTIMAGIC_FIELD') then
	local this_dev = stick:getNumber() > 1 and 'these devices' or 'this device'
	game.logPlayer(who, 'Your antimagic field prevents you from using %s.', this_dev)
	return nil
      end
      local ret = spell.cast_stick_spell(stick, who, inven, floor, item)
      if ret.used and not ret.no_power then
	-- Slight Hack(TM):  Use the charge ourself so we can redistribute
	-- charges.
	stick.power = stick.power - stick.use_power.power
	stick:on_stack()
	-- Since we already used a charge, tell the upstream caller not to
	-- use another one.
	ret.no_power = true
      end
      return ret
    end,
  },
  -- cf. recharge() in src/spells2.c
  recharge = {
    try = function(self, power, who)
      local n, strength = self:getNumber(), nil
      if n > 1 then
	strength = 100 + power - self.level - math.floor(8*self.power/n)
      else
	strength = math.floor((100 + power - self.level - 8*self.power)/15)
      end
      -- I think it's possible for this to be zero or less, which should
      -- probably mean instant failure.
      if (strength <= 0 or rng.chance(strength)) and not self:hasFlag('flags', 'RECHARGE') then
	-- Recharge failed.
	return false
      end
      local amount = rng.range(1, 1 + math.floor(power/(self.level+2)))
      -- Multiple wands in a stack increases recharging somewhat.
      if n > 1 then
	amount = amount + math.floor(rng.range(1, amount*(n-1))/2)
	amount = util.bound(amount, 1, 12)
      end
      self.power = self.power + amount
      if n > 1 then
	-- Distribute charges between wands if necessary.
	self:on_stack()
	-- "Un-identify" so we don't know the new charge level; we do know,
	-- however, that it's no longer empty.
	if not self:hasFlag('flags', 'RECHARGE') then
	  self:forAllStack(function(so)
	    so.identified = nil
	    so.pseudo_id_empty = nil
	  end)
	end
      end
      return true
    end,
    on_fail = function(self, who, inven, item, floor)
      local name = self:getName{no_count=true, no_mods=true, no_pval=true, no_inscr=true}
      if rng.percent(who:has('PERFECT_CASTING') and 75 or 80) then
	-- Blow up one wand.
	local your = self:getNumber() > 1 and 'one of your' or 'your'
	game.log(game.flash.BAD, 'Wild magic consumes %s %s!', your, name)
	who:getOneObject(inven, item, floor, {vocal_left=true})
      elseif who:has('PERFECT_CASTING') then
	-- Drain.
	game.log(game.flash.BAD, 'You save your %s from destruction but all charges are lost.', name)
	self:forAllStack(function(so) so.power = 0 end)
      else
	-- Blow up the entire stack of wands.
	local your = self:getNumber() > 1 and 'all of your' or 'your'
	game.log(game.flash.BAD, 'Wild magic consumes %s %s!', your, name)
	-- I can never remember whether this gets passed the inven table or
	-- just the inven ID...
	local inv_id = self:getInven(inven).id
	if floor then 
	  game.level.map:removeObject(who.x, who.y, item)
	else
	  who:removeObject(inv_id, item, true)
	end
      end
    end,
  },
  destroyed_by = { [DamageType.ELEC] = true, },
  -- Sneaky Hack(TM):  Distribute charges evenly among wands in the stack.
  on_stack = function(self)
    if not self.stacked then return end
    local charges = self:num_charges()
    local div = math.floor(charges/(1 + #self.stacked))
    local rnd = charges - div*(1 + #self.stacked)
    -- The first rnd wands get one extra charge, starting with self so that
    -- it will always have a charge if any do.
    self.power = div + (rnd > 0 and 1 or 0)
    for i, o in ipairs(self.stacked) do
      o.power = div + (i + 1 <= rnd and 1 or 0)
    end
  end,
  -- Primarily for use by Alchemists' Extract Essences talent.
  -- [cf. PlayerAlchemy:extractEssences()]
  extract_charges = function(self, n, who)
    n = math.min(n, self:num_charges())
    if n <= 0 then return 0 end
    -- We can totally cheat here and take all the charges out of the top
    -- wand in the stack, even if it leaves it with negative charges, and
    -- let on_stack() clean up behind us.  (If there's only one wand in the
    -- stack, then n can't be greater than the number of charges in the
    -- wand.)
    self.power = self.power - n
    if self:getNumber() > 1 then self:on_stack() end
    return n
  end,
  -- Primarily for use by Alchemists' Recharge Wand/Staff talent.
  -- [cf. PlayerAlchemy:playerRecharge
  add_charges = function(self, n, who)
    if n <= 0 then return 0 end
    -- As above, we can cheat by adding all the charges to the top wand in
    -- the stack and letting on_stack() distribute them for us.
    self.power = self.power + n
    if self:getNumber() > 1 then self:on_stack() end
    return n
  end,
}

newEntity {
  define_as = 'ROD_TIP_BASE',
  type = 'utility',
  subtype = 'rodtip',
  tval = 66,
  base_tmpl = '& Rod Tip~ of #',
  unid_tmpl = '& # Rod Tip~',
  display = '-',
  color = colors.C_d,
  weight = 1.5,
  encumber = 1.5,
  unid_cost = 90,	-- cf. object_value_base() in src/object2.c
  stacking = true,
  name_cb = function(self, name, t)
    if self.identified and self.rod_power and self.rod_power.power then
      name = name .. (' (%d)'):format(self.rod_power.power)
    end
    return name
  end,
  category = 'magic',
  use_simple = {
    titles = { 'Attach to rod', '(??? how are you seeing this?)' },
    use = function(self, who, inven, floor, item)
      if who:has('ANTIMAGIC_FIELD') then
	game.logPlayer(who, 'You cannot do this while your antimagic field is active.')
	return nil
      end
      if not self:isIDd() then
	game.logPlayer(who, 'You must identify this rod tip before you can attach it to a rod.')
	return {}
      end
      local ShowInventory = require 'mod.dialogs.ShowInventory'
      local function filter_cb(o)
	return o.type == 'utility' and o.subtype == 'rod' and not o.rod_tip and o.max_power >= math.floor(self.rod_power.power/(o:hasFlag('flags', 'CHEAPNESS') and 2 or 1))
      end
      local function action_cb(o)
	if not o then
	  who:talentDialogReturn(false)
	  return false
	end
	o.use_power = table.clone(self.rod_power, true)
	o.use_power.use = spell.use_rod
	if o:hasFlag('flags', 'CHEAPNESS') then
	  o.use_power.power = math.floor(self.rod_power.power/2)
	end
	-- Remove the rod tip from inventory/floor and attach it to the rod.
	o.rod_tip = who:getOneObject(inven, item, floor, {vocal_left=true})
	game.log('You create %s.', o:getName())
	who:sortInven()
	if who.recalcEverything then who:recalcEverything() end
	who:talentDialogReturn(true)
	return false
      end
      local ret = who:talentDialog(ShowInventory.new('Attach to rod:', who:getInven('INVEN'), filter_cb, action_cb, who))
      return { used=ret }
    end,
    name = 'attach the rod tip to a rod base (wait, how are you seeing this?)',
  },
  resolvers.flavored(),
}

newEntity {
  define_as = 'ROD_BASE',
  type = 'utility',
  subtype = 'rod',
  tval = 67,
  display = '-',
  name_cb_post_ego = function(self, name, idd, t)
    if self.rod_tip then
      -- Object:getName() will take care of adding current/max power
      -- (because if we have a tip, we must have a use_power{} table too).
      name = name .. ' of ' .. self.rod_tip.display_name
    else
      -- We don't have a use_power{} table, so Object:getName() won't show
      -- this bit (which is just as well, as we don't have power or a power
      -- cost yet), so we add max power here.
      name = name .. (' (%d)'):format(self.max_power)
    end
    return name
  end,
  price_mod = function(self, value)
    -- Include price of attached rod tip, if any.
    if self.rod_tip then value = value + self.rod_tip:getPrice() end
    return value
  end,
  extra_desc = function(self)
    -- We can assume both the rod and rod tip are identified at this point.
    return self.rod_tip and '\n#C__O#' .. self.rod_tip.desc .. '#LAST#' or nil
  end,
  sort_order = function(a, b)
    -- Rods with tips sort above rods without.
    if a.rod_tip and not b.rod_tip then return true end
    if b.rod_tip and not a.rod_tip then return false end
    -- Among tipped rods, sort by increasing tip sval.
    if a.rod_tip and b.rod_tip and a.rod_tip.sval ~= b.rod_tip.sval then
      return a.rod_tip.sval < b.rod_tip.sval
    end
    -- Between rods with the same tip (or same lack thereof), sort by
    -- increasing recharge time.
    local a_rch, b_rch = (a.max_power - a.power)/a.power_regen, (b.max_power - b.power)/b.power_regen
    if a_rch ~= b_rch then return a_rch < b_rch end
    -- I got nothin'.
    return nil
  end,
  -- cf. recharge() in src/spells2.c
  recharge = {
    try = function(self, power)
      local strength = math.floor(math.max(power - self.level, 0)/5)
      if rng.chance(strength) and not self:hasFlag('flags', 'RECHARGE') then
	-- Recharge failed.
	return false
      end
      local amount = power * rng.dice(3, 2)
      self.power = math.min(self.power + amount, self.max_power)
      return true
    end,
    on_fail = function(self, who, inven, item, floor)
      local name = self:getName{no_count=true, no_mods=true, no_pval=true, no_inscr=true}
      -- Slight Hack(TM):  If the rod tip is unique, just drain the rod.
      if self.rod_tip and self.rod_tip.unique then
	game.log('The recharging backfires - your %s is completely drained!', name)
	self.power = 0
	return
      end
      -- 1/3 chance of blowing up one rod (1/10 if we have Perfect
      -- Casting); otherwise drain the rod.
      if rng.chance(who:has('PERFECT_CASTING') and 10 or 3) then
	local your = self:getNumber() > 1 and 'one of your' or 'your'
	game.log(game.flash.BAD, 'Wild magic consumes %s %s!', your, name)
	who:getOneObject(inven, item, floor, {vocal_left=true})
      else
	game.log(game.flash.BAD, 'The recharge backfires, draining the rod further!')
	self.power = 0
      end
    end,
  },
  power_regen = 1,
  flags = { EASY_KNOW=true },
  category = 'magic',
  resolvers.rod(),
  resolvers.ego(),
}

newEntity {
  define_as = 'SCROLL_BASE',
  type = 'utility',
  subtype = 'scroll',
  tval = 70,
  base_tmpl = '& Scroll~ of #',
  unid_tmpl = '& Scroll~ titled "#"',
  display = '?',
  color = colors.C_d,
  weight = 0.5,
  encumber = 0.5,
  stacking = true,
  category = 'magic',
  unid_cost = 20,	-- cf. object_value_base() in src/object2.c
  destroyed_by = {
    [DamageType.ACID] = true,
    [DamageType.FIRE] = true,
  },
  use_simple = {
    titles = { 'Read', 'read' },
    use = function(self, who, inven, floor, item)
      game.log('[Sorry, not implemented (%s)]', self:getName{no_count=true})
    end,
    name = function(self)
      local desc = util.getval((self.use_simple.name_base or 'do something unspecified'), self)
      return desc .. obj_util.consume_desc_suffix(self, 'scroll')
    end,
  },
  cant_use_if = { blind=true, confused=true, unlit=true },
  alchemy_replace_on_use = 'SCRL_NOTHING',
  resolvers.flavored(),
  resolvers.ego(),
}

newEntity {
  define_as = 'POTION_BASE',
  type = 'utility',
  subtype = 'potion',
  tval = 71,
  base_tmpl = '& Potion~ of #',
  unid_tmpl = '& # Potion~',
  display = '!',
  weight = 0.4,
  encumber = 0.4,
  unid_cost = 20,	-- cf. object_value_base() in src/object2.c
  use_simple = {
    titles = { 'Drink', 'quaffed' },
    use = function(self, who, inven, floor, item)
      game.log('[Sorry, not implemented (%s)]', self:getName{no_count=true})
    end,
    name = function(self)
      local desc = util.getval((self.use_simple.name_base or 'do something unspecified'), self)
      return desc .. obj_util.consume_desc_suffix(self, 'potion')
    end,
  },
  stacking = true,
  category = 'magic',
  destroyed_by = { [DamageType.COLD] = true, },
  on_destroy = function(self, src, x, y)
    if self.shatter_dam then
      src = src.resolveSource and src:resolveSource() or src
      local sd = self.shatter_dam
      if sd.dice and se.dt then
	local dam = { rng.dice(sd.dice[1], sd.dice[2]), cx=x, cy=y }
	spell.ball_spell(src, nil, sd.dt, dam, sd.r, x, y, 0, spell.simple_blast_particle)
      end
      if sd.angry and target:reactionToward(src) > 0 then
	-- TODO anger target against src
      end
    end
  end,
  alchemy_replace_on_use = 'EMPTY_BOTTLE',
  resolvers.flavored(),
  resolvers.ego(),
}

newEntity {
  define_as = 'FOOD_BASE',
  type = 'food',
  subtype = 'misc',
  tval = 80,
  display = ',',
  stacking = true,
  check_stack_ok_known = true,
  category = 'tools',
  unid_cost = 5,	-- cf. object_value_base() in src/object2.c
}

newEntity {
  base = 'FOOD_BASE',
  define_as = 'MUSHROOM_BASE',
  subtype = 'mushroom',
  base_tmpl = '& Mushroom~ of #',
  unid_tmpl = '& # Mushroom~',
  resolvers.flavored(),
}

newEntity {
  define_as = 'MONEY_BASE',
  type = 'money',
  subtype = 'money',
  tval = 100,
  money_value = resolvers.money_value(),
  name_cb = function(self, name, t)
    if t.unresolved then return self.display_name end
    local s = self.money_value == 1 and '' or 's'
    return ('%d gold piece%s worth of %s'):format(self.money_value, s, self.display_name)
  end,
  display = '$',
  weight = 0,
  encumber = 0,
  auto_pickup = true,
  on_prepickup = function(self, who, id)
    if who == game.player then
      who:adjGold(self.money_value)
      game.logPlayer(who, 'You pick up %s.', self:getName())
      game.level.map:removeObject(who.x, who.y, id)
      return true
    else
      return false
    end
  end,
}

newEntity {
  define_as = 'RUNE_BASE',
  type = 'rune',
  base_tmpl = '& Rune~ [#]',
  display = '?',
  weight = 0.2,
  encumber = 0.2,
  stacking = true,
  check_stack = function(self, o)
    -- Everything stacks but runestones.
    return self.define_as ~= 'RUNE_RUNESTONE' and o.define_as ~= 'RUNE_RUNESTONE'
  end,
  category = 'magic',
}

newEntity {
  base = 'RUNE_BASE',
  define_as = 'RUNE_DAM_BASE',
  subtype = 'damage',
  tval = 104,
}

newEntity {
  base = 'RUNE_BASE',
  define_as = 'RUNE_MOD_BASE',
  subtype = 'modifier',
  tval = 105,
}

newEntity {
  define_as = 'BOOK_BASE',
  type = 'magic',
  subtype = 'spellbook',
  tval = 111,
  display = '?',
  weight = 3,
  encumber = 3,
  stacking = true,
  check_stack = function(self, o)
    -- Require both ID'd.
    if not self:known() or not o:known() then return false end
    -- Require same spells.
    if not self.carried_talents or not o.carried_talents then return false end
    if #self.carried_talents ~= #o.carried_talents then return false end
    for i = 1, #self.carried_talents do
      if self.carried_talents[i] ~= o.carried_talents[i] then return false end
    end
    -- Don't let the Mage quest book stack with anything else
    if self.unique or o.unique then return false end
    -- Caller will check for matching egos for us.
    return true
  end,
  category = 'magic',
  destroyed_by = { [DamageType.FIRE] = true, },
  resolvers.ego(),
}

newEntity {
  define_as = 'JUNK_BASE',
  tval = 11,
  category = 'junk',
  flags = { EASY_KNOW=true },
}

newEntity {
  define_as = 'SKELETON_BASE',
  tval = 1,
  display = '~',
  flags = { EASY_KNOW=true },
  category = 'junk',
  destroyed_by = { [DamageType.ACID] = true, },
}

newEntity {
  define_as = 'STACK_BY_MON_BASE',
  stacking = true,
  check_stack = function(self, o)
    -- TODO Require same associated monster.
    return false
  end,
}
