-- $Id: Map.lua 4246 2021-07-21 02:41:50Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2018 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Map = require "engine.Map"
local DamageType = require 'engine.DamageType'
local Tiles = require 'engine.Tiles'

module(..., package.seeall, class.inherit(Map))

Map.faction_danger_check = function(self, e) return e.unique or e.quest_mon end

function _M:init(w, h)
  Map.init(self, w, h)
end

function _M:loaded()
  Map.loaded(self)

  -- We replace the remembers() callback with our own callback that
  -- accounts for the :anyMarkedObject() method below.
  local function mapremember(t, x, y, v)
    if not x or not y or x < 0 or y < 0 or x >= self.w or y >= self.h then
      return
    end
    if v ~= nil then
      t[x + y * self.w] = v
      self._map:setRemember(x, y, self:isRemembered(x, y, v))
      self.changed = true
    end
    return t[x + y * self.w]
  end
  setmetatable(self.remembers, {__call = mapremember})
  -- We also replace the map() callback with the one below.
  getmetatable(self).__call = _M.call
end

function _M:call(x, y, pos, e)
  if self:isBound(x, y) and pos == Map.TERRAIN and e then
    local ee = self.map[x + y * self.w][pos]
    if ee and ee.unattr_on_replace then
      -- Sneaky Hack(TM):  If the feature we're removing defines an
      -- unattr_on_replace{} list of map attrs, unset those as we're
      -- replacing the feature.
      for _, attr in ipairs(ee.unattr_on_replace) do
	self:unattr(x, y, attr)
      end
    end
  end

  -- And fall through to the regular handling.
  return Map.call(self, x, y, pos, e)
end

-- Return the table of map attributes for the specified grid, or nil if the
-- grid has no map attributes.
function _M:all_attrs(x, y)
  if not x or not y or x < 0 or y < 0 or x >= self.w or y >= self.h then return nil end
  return self.attrs[x + y * self.w]
end

-- Clear a map attribute from a specified grid; basically the opposite of
-- map.attrs().
function _M:unattr(x, y, k)
  if not x or not y or x < 0 or y < 0 or x >= self.w or y >= self.h then return end

  local idx = x + y * self.w
  if self.attrs[idx] then
    self.attrs[idx][k] = nil
    if table.count(self.attrs[idx]) == 0 then self.attrs[idx] = nil end
  end
end

-- Clear all map attributes from the specified grid.
function _M:unattr_all(x, y)
  if not x or not y or x < 0 or y < 0 or x >= self.w or y >= self.h then return nil end
  self.attrs[x + y * self.w] = nil
end

-- Vaguely Annoying Hack(TM):  emgine.interface.GameTargeting:init()
-- expects to find a viewport in engine.Map; we'll put it in both places.
function _M:setViewPort(x, y, w, h, tile_w, tile_h, fontname, fontsize, allow_backcolor)
  Map.setViewPort(self, x, y, w, h, tile_w, tile_h, fontname, fontsize, allow_backcolor)
  Map.setViewPort(Map, x, y, w, h, tile_w, tile_h, fontname, fontsize, allow_backcolor)
end

-- Returns a list of (x+y*w) indexes for remembered grids.
function _M:exportRemembers()
  local xys = table.keys(self.remembers)
  table.sort(xys)
  return xys
end

-- Flag as remembered the provided list of (x+y*w) indexes [as returned by
-- :exportRemembers() above]
function _M:importRemembers(xys)
  for _, xy in ipairs(xys) do
    local x, y = xy % self.w, math.floor(xy/self.w)
    self.remembers[xy] = true
    self._map:setRemember(x, y, true)
  end
  self.changed = true
end

-- Returns a list of (x+y*w) indexes for seen grids.
function _M:exportSeens()
  local xys = table.keys(self.has_seens)
  table.sort(xys)
  return xys
end

-- Flag as seen the provided list of (x+y*w) indexes [as returned by
-- :exportSeens() above]
function _M:importSeens(xys)
  for _, xy in ipairs(xys) do
    self.has_seens[xy] = true
  end
  self.changed = true
end

---- Map-effect-related modifications (these used to live in
---- mod.class.MapEffects).

-- We override Map:displayEffects() to add special handling for effects
-- with multi-color backgrounds.
function _M:displayEffects(z, prevfbo, nb_keyframes)
  local sx, sy = self._map:getScroll()
  for e, _ in pairs(self.z_effects[z]) do
    -- Dont bother with obviously out of screen stuff
    if e.overlay and e.overlay.zdepth == z and e.x + e.radius >= self.mx and e.x - e.radius < self.mx + self.viewport.mwidth and e.y + e.radius >= self.my and e.y - e.radius < self.my + self.viewport.mheight then
      -- Slight Hack(TM):  If the effect is a ShimmerOverlay, we'll fetch
      -- its tiles below.
      local s
      if not e.overlay.tileForXY then
	s = self.tilesEffects:get(e.overlay.display, e.overlay.color_r, e.overlay.color_g, e.overlay.color_b, e.overlay.color_br, e.overlay.color_bg, e.overlay.color_bb, e.overlay.image, e.overlay.alpha)
      end

      -- If we dont have a special fbo/shader or no shader image to use, just display with simple quads
      if not self.fbo or not e.overlay.effect_shader then
	-- Now display each grids
	-- Slight Hack(TM), redux:  ShimmerOverlays will give us different
	-- tiles for each grid.
	for lx, ys in pairs(e.grids) do
	  for ly, _ in pairs(ys) do
	    if self.seens(lx, ly) then
	      if e.overlay.tileForXY then s = e.overlay:tileForXY(lx, ly) end
	      s:toScreen(self.display_x + sx + (lx - self.mx) * self.tile_w * self.zoom, self.display_y + sy + (ly - self.my) * self.tile_h * self.zoom, self.tile_w * self.zoom, self.tile_h * self.zoom)
	    end
	  end
	end
      -- We have a fbo/shader pair, so we display everything inside it and apply the shader to get nice borders and such
      else
	if not e.overlay.effect_shader_tex then
	  e.overlay.effect_shader_tex = {}
	  if type(e.overlay.effect_shader) == "table" then
	    for i = 1, #e.overlay.effect_shader do
	      e.overlay.effect_shader_tex[i] = Tiles:loadImage(e.overlay.effect_shader[i]):glTexture()
	      e.overlay.effect_shader_tex.cur = 1
	      e.overlay.effect_shader_tex.cnt = 0
	      e.overlay.effect_shader_tex.max = e.overlay.effect_shader.max
	    end
	  else
	    e.overlay.effect_shader_tex[1] = Tiles:loadImage(e.overlay.effect_shader):glTexture()
	    e.overlay.effect_shader_tex.cur = 1
	    e.overlay.effect_shader_tex.cnt = 0
	    e.overlay.effect_shader_tex.max = 1
	  end
	end

	self.fbo:use(true, 0, 0, 0, 0)
	-- Now display each grids
	-- Slight Hack(TM), re-redux:  ShimmerOverlays will give us different
	-- tiles for each grid.
	for lx, ys in pairs(e.grids) do
	  for ly, _ in pairs(ys) do
	    if self.seens(lx, ly) then
	      if e.overlay.tileForXY then s = e.overlay:tileForXY(lx, ly) end
	      s:toScreen((lx - self.mx) * self.tile_w * self.zoom, (ly - self.my) * self.tile_h * self.zoom, self.tile_w * self.zoom, self.tile_h * self.zoom)
	    end
	  end
	end
	self.fbo:use(false, prevfbo)
	e.overlay.effect_shader_tex[e.overlay.effect_shader_tex.cur]:bind(1, false)
	self.fbo_shader.shad:use(true)
	self.fbo_shader.shad:uniTileSize(self.tile_w, self.tile_h)
	self.fbo_shader.shad:uniScrollOffset(0, 0)
	self.fbo:toScreen(self.display_x + sx, self.display_y + sy, self.viewport.width, self.viewport.height, self.fbo_shader.shad, 1, 1, 1, 1, true)
	self.fbo_shader.shad:use(false)

	e.overlay.effect_shader_tex.cnt = e.overlay.effect_shader_tex.cnt + nb_keyframes
	if e.overlay.effect_shader_tex.cnt >= e.overlay.effect_shader_tex.max then
	  e.overlay.effect_shader_tex.cnt = e.overlay.effect_shader_tex.cnt - e.overlay.effect_shader_tex.max
	  e.overlay.effect_shader_tex.cur = util.boundWrap(e.overlay.effect_shader_tex.cur + 1, 1, #e.overlay.effect_shader_tex)
	end
      end
    end
  end
end

local function getEffectName(self)
  local name = self.name or (type(self.damtype) == 'number' and DamageType.dam_def[self.damtype].name .. ' area effect') or 'area effect'

  if self.src then
    return (self.src.getName and self.src:getName{no_count=true} or self.src.name or 'something') .. "'s " .. name
  end

  return name
end

local function resolveSource(self)
  return self.src and self.src.resolveSource and self.src:resolveSource() or self
end

-- We override Map:addEffect() to add custom grid functionality:  if the
-- 'dir' argument is a function, it will be called with the effect as a
-- parameter and should return the grids to use for the effect, in the
-- [x][y] format returned by f.i. core.fov.circle_grids().
function _M:addEffect(src, x, y, duration, damtype, dam, radius, dir, angle, overlay, update_fct, selffire, friendlyfire)
  if selffire == nil then selffire = true end
  if friendlyfire == nil then friendlyfire = true end

  local grids

  if type(angle) == "table" then
    -- Custom grids
    grids = angle
    angle = nil
  elseif type(dir) == 'function' then
    -- Custom grids computed by function.  We pass in a mockup of the
    -- effect.
    grids = dir({
      src=src, x=x, y=y, duration=duration, damtype=damtype, dam=dam, radius=radius, dir=dir, angle=angle,
      overlay=overlay and overlay.__CLASSNAME and overlay,
      -- Empty grids{} for the mockup.
      grids = {},
      update_fct=update_fct, selffire=selffire, friendlyfire=friendlyfire,
    })
  elseif type(dir) == "table" then
    -- Handle any angle
    grids = core.fov.beam_any_angle_grids(x, y, radius, angle, dir.source_x or src.x or x, dir.source_y or src.y or y, dir.delta_x, dir.delta_y, true)
  elseif dir == 5 then
    -- Handle balls
    grids = core.fov.circle_grids(x, y, radius, true)
  else
    -- Handle beams
    grids = core.fov.beam_grids(x, y, radius, dir, angle, true)
  end

  local e = {
    src=src, x=x, y=y, duration=duration, damtype=damtype, dam=dam, radius=radius, dir=dir, angle=angle,
    overlay=overlay and (overlay.__ATOMIC or overlay.__CLASSNAME) and overlay,
    grids = grids,
    update_fct=update_fct, selffire=selffire, friendlyfire=friendlyfire,
    getName=getEffectName, resolveSource=resolveSource
  }

  local overlay_particle = nil
  if overlay and not overlay.__CLASSNAME then
    overlay_particle = overlay
  elseif overlay then
    if overlay.overlay_particle then overlay_particle = overlay.overlay_particle end
  end

  while overlay_particle do
    e.particles = e.particles or {}
    if overlay_particle.only_one then
      e.particles[#e.particles+1] = self:particleEmitter(x, y, 1, overlay_particle.type, overlay_particle.args, nil, overlay_particle.zdepth)
      e.particles_only_one = true
    else
      e.fake_overlay = overlay_particle
      for lx, ys in pairs(grids) do
	for ly, _ in pairs(ys) do
	  e.particles[#e.particles+1] = self:particleEmitter(lx, ly, 1, overlay_particle.type, overlay_particle.args, nil, overlay_particle.zdepth)
	end
      end
    end
    overlay_particle = overlay_particle.overlay_particle
  end
  -- If nothing set, display on the last z-layer
  if e.overlay and not e.overlay.zdepth then e.overlay.zdepth = self.zdepth - 1 end

  table.insert(self.effects, e)
  if e.overlay then self.z_effects[e.overlay.zdepth][e] = true end

  self.changed = true
  return e
end

-- Similarly, we override Map:processEffects() to support the same dir()
-- grids method that we added to Map:addEffect().  In addition, we allow
-- the update_fct() method to recompute the grids itself:  if the method
-- sets the 'computed_grids' field of the passed-in effect, the caller (ie.
-- this method) will clear said field and leave the grids untouched.
function _M:processEffects(update_shape_only)
  local todel = {}
  for i, e in ipairs(self.effects) do
    -- Run damage and decrease duration only on certain ticks
    if not update_shape_only then
      for lx, ys in pairs(e.grids) do
	for ly, _ in pairs(ys) do
	  local act = game.level.map(lx, ly, Map.ACTOR)
	  if act and act == e.src and not ((type(e.selffire) == "number" and rng.percent(e.selffire)) or (type(e.selffire) ~= "number" and e.selffire)) then
	    -- Self-fire check; no damage
	  elseif act and e.src and e.src.reactionToward and (e.src:reactionToward(act) >= 0) and not ((type(e.friendlyfire) == "number" and rng.percent(e.friendlyfire)) or (type(e.friendlyfire) ~= "number" and e.friendlyfire)) then
	    -- Friendly-fire check; no damage
	  else
	    -- Otherwise hit
	    e.src.__project_source = e -- intermediate projector source
	    if type(e.damtype) == 'function' then
	      e.damtype(e, act, lx, ly)
	    elseif e.damtype then
	      DamageType:get(e.damtype).projector(e.src, lx, ly, e.damtype, e.dam)
	    end
	    e.src.__project_source = nil
	  end
	end
      end
      e.duration = e.duration - (e.decrease or 1)
    end

    if e.duration <= 0 then
      table.insert(todel, i)
    elseif e.update_fct then
      if e:update_fct(update_shape_only) then
	if e.computed_grids then
	  e.computed_grids = nil
	elseif type(e.dir) == 'function' then
	  e.grids = e:dir()
	elseif type(e.dir) == "table" then
	  e.grids = core.fov.beam_any_angle_grids(e.x, e.y, e.radius, e.angle, e.dir.source_x or e.src.x or e.x, e.dir.source_y or e.src.y or e.y, e.dir.delta_x, e.dir.delta_y, true)
	elseif e.dir == 5
	  then e.grids = core.fov.circle_grids(e.x, e.y, e.radius, true)
	else
	  e.grids = core.fov.beam_grids(e.x, e.y, e.radius, e.dir, e.angle, true)
	end
	if e.particles then
	  if e.particles_only_one then
	    e.particles[1]:shiftCustom(self.tile_w * (e.particles[1].x - e.x), self.tile_h * (e.particles[1].y - e.y))
	    e.particles[1].x = e.x
	    e.particles[1].y = e.y
	  else
	    for j, ps in ipairs(e.particles) do self:removeParticleEmitter(ps) end
	    e.particles = {}
	    for lx, ys in pairs(e.grids) do
	      for ly, _ in pairs(ys) do
		e.particles[#e.particles+1] = self:particleEmitter(lx, ly, 1, e.fake_overlay.type, e.fake_overlay.args, nil, e.zdepth)
	      end
	    end
	  end
	end
      end
    end
  end

  if #todel > 0 then table.sort(todel) end
  for i = #todel, 1, -1 do
    local e = table.remove(self.effects, todel[i])
    if e.particles then
      for j, ps in ipairs(e.particles) do self:removeParticleEmitter(ps) end
    end
    if e.overlay then
      self.z_effects[e.overlay.zdepth][e] = nil
    end
  end
end

-- We override Map:updateMap() with slight modifications from T4's
-- mod.class.AsciiMap so that the "highest" of terrain/trap/object/actor
-- will be displayed alone instead of stacking them all semi-transparently.
function _M:updateMap(x, y)
  if not x or not y or x < 0 or y < 0 or x >= self.w or y >= self.h then
    return
  end

  -- Update minimap if any
  local mos = {}

  self._map:setImportant(x, y, false)
  if not self.updateMapDisplay then
    local g = self(x, y, _M.TERRAIN)
    local o = self(x, y, _M.OBJECT)
    local a = self(x, y, _M.ACTOR)
    local t = self(x, y, _M.TRAP)
    local p = self(x, y, _M.PROJECTILE)

    -- Stack depth for various entities.
    local zs = { t=4, o=7, a=10, p=13 }
    if config.settings.tome2 and config.settings.tome2.tiles == 'ascii' then
      -- In ASCII mode, collapse most of these to the same depth
      -- (projectiles are still at z=2 as per AsciiMap, but we don't use
      -- any of those anyway).
      zs = { t=1, o=1, a=1, p=2 }
    end

    if g then
      -- Update path caches from path strings
      for i = 1, #self.path_strings do
	local ps = self.path_strings[i]
	self._fovcache.path_caches[ps]:set(x, y, g:check("block_move", x, y, self.path_strings_computed[ps] or ps, false, true))
      end

      g:getMapObjects(self.tiles, mos, 1)
      g:setupMinimapInfo(g._mo, self)
    end
    if t then
      -- Handles trap being known
      if not self.actor_player or t:knownBy(self.actor_player) then
	t:getMapObjects(self.tiles, mos, zs.t)
	t:setupMinimapInfo(t._mo, self)
      else
	t = nil
      end
    end
    if o then
      o:getMapObjects(self.tiles, mos, zs.o)
      o:setupMinimapInfo(o._mo, self)
      if self.object_stack_count then
	local mo = o:getMapStackMO(self, x, y)
	if mo then mos[9] = mo end
      end
    end
    if a then
      -- Handles invisibility and telepathy and other such things
      if not self.actor_player or self.actor_player:canSee(a) then
	a:getMapObjects(self.tiles, mos, zs.a)
	a:setupMinimapInfo(a._mo, self)

--	self._map:setImportant(x, y, true)
      end
    end
    if p then
      p:getMapObjects(self.tiles, mos, zs.p)
      p:setupMinimapInfo(p._mo, self)
    end
  else
    self:updateMapDisplay(x, y, mos)
  end

  -- Update entities checker for this spot
  -- This is to improve speed, we create a function for each spot that checks entities it knows are there
  -- This avoid a costly for iteration over a pairs() and this allows luajit to compile only code that is needed
  local ce, sort = {}, {}
  local fstr = [[if m[%s] then p = m[%s]:check(what, x, y, ...) if p then return p end end ]]
  ce[#ce+1] = [[return function(self, x, y, what, ...) local p local m = self.map[x + y * self.w] ]]
  for idx, e in pairs(self.map[x + y * self.w]) do sort[#sort+1] = idx end
  table.sort(sort, searchOrderSort)
  for i = 1, #sort do ce[#ce+1] = fstr:format(sort[i], sort[i]) end
  ce[#ce+1] = [[end]]
  local ce = table.concat(ce)
  self._check_entities[x + y * self.w] = self._check_entities_store[ce] or loadstring(ce)()
  self._check_entities_store[ce] = self._check_entities[x + y * self.w]

  -- Cache the map objects in the C map
  self._map:setGrid(x, y, mos)

  -- Update FOV caches
  if self:checkAllEntities(x, y, "block_sight", self.actor_player) then
    self._fovcache.block_sight:set(x, y, true)
  else
    self._fovcache.block_sight:set(x, y, false)
  end
  if self:checkAllEntities(x, y, "block_esp", self.actor_player) then
    self._fovcache.block_esp:set(x, y, true)
  else
    self._fovcache.block_esp:set(x, y, false)
  end
  if self:checkAllEntities(x, y, "block_sense", self.actor_player) then
    self._fovcache.block_sense:set(x, y, true)
  else
    self._fovcache.block_sense:set(x, y, false)
  end
end

---- Marked-object-related modifications.  This roughly mimics T2's method
---- of allowing the map to "remember" that an object was present on a grid
---- even if the grid itself wasn't memorized, by setting an object's
---- 'marked' field and checking for it in map_info().  We implement this
---- in T4 by making the C map's :setRemember() state reflect both the
---- remembers{} table and the presence or absence of any marked objects on
---- the grid.

-- Override Map:addObject() to update the :setRemember() state.  And as
-- long as we're here, also stack the added object with existing objects on
-- the grid if possible.
function _M:addObject(x, y, o)
  local i = self.OBJECT
  while true do
    local oo = self(x, y, i)
    if not oo then
      -- Empty spot; we can drop our object here.
      self(x, y, i, o)
      self._map:setRemember(x, y, self:isRemembered(x, y))
      return true, i - self.OBJECT + 1
    elseif oo:canStack(o) then
      -- Our object will merge with this floor object.
      oo:stack(o)
      self._map:setRemember(x, y, self:isRemembered(x, y))
      return true, i - self.OBJECT + 1
    else
      -- Keep looking.
      i = i + 1
    end
  end
end

-- Override Map:removeObject() to update the :setRemember() state.
function _M:removeObject(x, y, i)
  local ret = Map.removeObject(self, x, y, i)
  if ret then
    self._map:setRemember(x, y, self:isRemembered(x, y))
  end
  return ret
end

-- Convenience method; returns true if remembers{} is set for the specified
-- grid or any object on the specified grid is marked to remember.  The
-- optional 'v' paramater is used by the remembers{} metatable callback,
-- and contains the value passed to the callback to be put in remembers{}.
function _M:isRemembered(x, y, v)
  if v == nil then v = self.remembers(x, y) end
  if v then return true end

  -- Walk the object stack.
  local i = self.OBJECT
  while true do
    local oo = self(x, y, i)
    if not oo then break end			-- Ran out of objects.
    if oo.map_remember then return true end	-- Found a marked object.
    -- Keep looking.
    i = i + 1
  end

  -- Slight Hack(TM):  A mimic carrying a marked object also counts.
  local a = self(x, y, Map.ACTOR)
  if a and a:has('MIMIC') and a:findInInventoryBy(a:getInven('INVEN'), 'map_remember', true) then
    return true
  end

  return false
end

-- Applies the provided callback to all objects on the specified grid.
-- The callback will receive four parameters:  the object, the x and y
-- coordinates, and the stack index.
function _M:objectsApply(x, y, cb)
  local i = self.OBJECT
  while true do
    local o = self(x, y, i)
    if not o then break end
    cb(o, x, y, i)
    -- Next object.
    i = i + 1
  end
end

-- Override Map:redisplay() to pass :isRemembered() to _map:setRemember()
function _M:redisplay()
  self:checkMapViewBounded()
  self._map:setScroll(self.mx, self.my, 0)
  for i = 0, self.w - 1 do
    for j = 0, self.h - 1 do
      self._map:setSeen(i, j, self.seens(i, j))
      self._map:setRemember(i, j, self:isRemembered(i, j))
      self._map:setLite(i, j, self.lites(i, j))
      self:updateMap(i, j)
    end
  end
end

-- Override Map:applyLite() to additionally mark objects on the seen grid.
function _M:applyLite(x, y, v)
  engine.Map.applyLite(self, x, y, v)
  if x >= 0 and x < self.w and y >= 0 and y < self.h then
    self:objectsApply(x, y, function(o, _, _, _) o:markForMap(true) end)
    self._map:setRemember(x, y, self:isRemembered(x, y))
  end
end

-- Similarly with Map:applyExtraLite().
function _M:applyExtraLite(x, y, v)
  engine.Map.applyExtraLite(self, x, y, v)
  if x >= 0 and x < self.w and y >= 0 and y < self.h and self.infovs[x + y * self.w] then
    self:objectsApply(x, y, function(o, _, _, _) o:markForMap(true) end)
    self._map:setRemember(x, y, self:isRemembered(x, y))
  end
end
