-- $Id: MathomHouse.lua 3002 2017-07-26 01:36:38Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Store = require 'engine.Store'
local Dialog = require 'engine.ui.Dialog'
local ShowStore = require 'mod.dialogs.ShowStore'
local GetQuantity = require "engine.dialogs.GetQuantity"

module(..., package.seeall, class.inherit(Store))

_M.museums_def = {}

function _M:loadMuseums(f)
  self.museums_def = self:loadList(f)
end

function _M:init(t, no_default)
  Store.init(self, t, no_default)
  -- There's no obvious way to override the 't.body = {INVEN=10000}' in
  -- Store.init(), so we patch that up here.
  self:getInven('INVEN').max = 255
  -- Can't take anything out of a Mathom-house.
  self.allow_buy = false
end

function _M:canRestock()
  -- Mathom-houses don't restock.
  return false
end

function _M:loadup(level, zone, force_nb)
  -- Nor should they be filled with an initial stock.
end

-- Mostly the same as Store.interact(), with a few tweaks.
function _M:interact(who, name)
  local store, inven = self:getInven("INVEN"), who:getInven("INVEN")
  local d
  local action_cb = function(what, o, item)
    if what == "buy" then
      Dialog:simplePopup('???', "#{italic}#*blink*#{normal}# ...Okay, this shouldn't be possible.")
    else
      if o:getNumber() > 1 then
	local q
	q = GetQuantity.new(nil, nil, o:getNumber(), o:getNumber(), function(qty) self:doSell(who, o, item, qty, d) end)
	game:registerDialog(q)
      else
	self:doSell(who, o, item, 1, d)
      end
    end
  end
  local desc_cb = function(what, o) return self:descObject(who, what, o) end
  local descprice_cb = function(what, o) return '' end
  d = ShowStore.new((name or self.name), self, store, who, inven, nil, nil,
		    action_cb, desc_cb, descprice_cb,
		    self.allow_sell, self.allow_buy,
		    function(item) end, self, who)
  game:registerDialog(d)
end

function _M:trySell(who, o, item, qty)
  -- Do we have room in the Mathom-house?
  local inven = self:getInven('INVEN')
  local can_add = #inven < inven.max
  if not can_add and o and o:stackable() then
    -- Maybe we can stack it with something?
    local stack_limit = self:invenStackLimit('INVEN')
    for i, io in ipairs(inven) do
      if o:canStack(io) and io:getNumber() + qty <= stack_limit then
	can_add = true
	break
      end
    end
  end
  if not can_add then
    Dialog:simpleLongPopup('Mathom-house full', 'Alas, we have no storage space left for your generous contribution.  Thank you anyway, though.', 600)
    return nil, nil
  end

  return qty, 0
end

function _M:doBuy(who, o, item, qty, store_dialog)
  Dialog:simplePopup('???', 'Seriously, how did you get here?')
end

function _M:doSell(who, o, item, qty, store_dialog)
  qty = math.min(qty, o:getNumber())
  qty, _ = self:trySell(who, o, item, qty)
  if not qty then return end
  local o_name = o:getName{ force_qty=qty }
  local o_id_name = o:getName{ force_qty=true, force_id=true }
  local function callback(ret)
    -- "No" is on the default button.
    if not ret then
      self:transfer(who, self, item, qty)
      if store_dialog then store_dialog:updateStore() end
      game.log('%s donates to Mathom-house:  %s', who.name, o_id_name)
    end
  end
  local msg = ('(#{bold}#NOTE:#{normal}#  Items donated to the Mathom-house cannot be retrieved.)\n\nDonate %s?'):format(o_name)
  Dialog:yesnoLongPopup('Donate to Mathom-house', msg, 600, callback, 'No', 'Yes', nil, true)
end

-- We replace the parent transfer() to identify objects as we donate them.
-- [We can't do that in onSell() because we need a ref to the object in
-- dest for that.]
function _M:transfer(src, dest, item, nb)
  local src_inven, dest_inven = src:getInven("INVEN"), dest:getInven("INVEN")
  for i = 1, nb do
    local o = src:removeObject(src_inven, item)
    o:identify(true)
    dest:addObject(dest_inven, o)
  end
  src:sortInven(src_inven)
  dest:sortInven(dest_inven)
end

function _M:sortInven(inven)
  if not inven then inven = self.inven[self.INVEN_INVEN] end
  inven = self:getInven(inven)
  if not inven then return end
  obj_util.sort_inven(inven)
  self.changed = true
end
