-- $Id: NPCEgo.lua 3333 2018-04-20 00:32:38Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2018 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Entity = require 'engine.Entity'
local Tiles = require 'engine.Tiles'

module(..., package.seeall, class.inherit(Entity))

_M.egos_def = {}

-- Static
function _M:loadEgos(f)
  self.egos_def = self:loadList(f)
end

-- Static
function _M:get(id)
  return self.egos_def[id]
end

function _M:applyMod(val, mod)
  if _G.type(mod) == 'number' then return val + mod end
  if _G.type(mod) == 'table' then
    if mod.set then return mod.set end
    if mod.pct then return math.floor((val * mod.pct)/100) end
  end
  return math.max(val, 0)
end

-- cf. mego_ok() in src/monster2.c
-- Static
function _M:okEgo(m, ego)
  if m.unique then return false end
  if ego.required then
    for f, _ in pairs(ego.required.flags or {}) do
      if not m.flags[f] then return false end
    end
    for f, _ in pairs(ego.required.drop_corpse or {}) do
      if not m.drop_corpse or not m.drop_corpse[f] then return false end
    end
    if ego.required.symbols then
      local ok_sym = false
      for c, _ in pairs(ego.required.symbols) do
	ok_sym = ok_sym or c == m.display
      end
      if not ok_sym then return false end
    end
  end
  if ego.forbidden then
    for f, _ in pairs(ego.forbidden.flags or {}) do
      if m.flags[f] then return false end
    end
    for c, _ in pairs(ego.forbidden.symbols or {}) do
      if c == m.display then return false end
    end
  end
  return true
end

-- cf. pick_ego_monster() in src/monster2.c
-- Static
function _M:tryAddEgo(m)
  -- No ego uniques, thanks...
  if m.unique then return end

  local RF = require('mod.class.NPC').resolve_flags
  if RF.force_no_ego then return end
  local ego = RF.use_ego and self.egos_def[RF.use_ego]

  if ego then
    -- Make sure the pre-set ego is actually applicable.
    if not self:okEgo(m, ego) then ego = nil end
  elseif not game.zone.wilderness and not game.zone.town and m.level and m.level > 0 and rng.percent(18) then
    for _ = 1, #self.egos_def + 10 do
      local ee = rng.table(self.egos_def)
      local doit = self:okEgo(m, ee)
      if doit then
	-- Check for OoD egos.
	local lvl = self:applyMod(m.level, ee.mod and ee.mod.level or 0)
	local llvl = game.zone:level_adjust_level(game.level, 'actor')
	lvl = lvl - (math.floor(llvl/2) + rng.range(1, math.floor(llvl/2)))
	doit = rng.chance(math.max(1, lvl))
      end
      -- Check rarity
      doit = doit and rng.chance(ee.rarity)
      if doit then
	-- And there was much rejoicing. ("Yaaay...")
	ego = ee
	break
      end
    end
  end

  m.ego = ego and ego.define_as
end

local function find_resolver(m, tag)
  for i, v in ipairs(m) do
    if _G.type(v) == 'table' and v.__resolver == tag then return i, v end
  end
  return nil
end

-- cf. race_info_idx() in src/monster2.c
-- Static
function _M:applyEgo(m)
  local ego = m.ego and self.egos_def[m.ego]
  if not ego then return end

  for f, mod in pairs(ego.mod or {}) do
    if f == 'life' then
      m.hd[1] = self:applyMod(m.hd[1], mod[1])
      m.hd[2] = self:applyMod(m.hd[2], mod[2])
      if _G.type(m.max_life) == 'number' or mod.force_max then
	m.max_life = m.hd[1]*m.hd[2]
      elseif m.max_life.__resolver == 'dice' then
	m.max_life[1] = self:applyMod(m.max_life[1], mod[1])
	m.max_life[2] = self:applyMod(m.max_life[2], mod[2])
      end
    elseif f == 'attacks' then
      if m.attacks then
	for i, m_att in ipairs(mod) do
	  if m.attacks[i] or m_att.method then
	    m.attacks[i] = m.attacks[i] or {}
	    m.attacks[i].type = mod[i].type or m.attacks[i].type
	    m.attacks[i].method = mod[i].method or m.attacks[i].method
	    if m_att.dam then
	      m.attacks[i].dam = m.attacks[i].dam or { 0, 0 }
	      m.attacks[i].dam[1] = self:applyMod(m.attacks[i].dam[1], mod[i].dam[1])
	      m.attacks[i].dam[2] = self:applyMod(m.attacks[i].dam[2], mod[i].dam[2])
	    end
	  end
	end
      end
    else
      m[f] = self:applyMod(m[f], mod)
    end
  end

  for f, mod in pairs(ego.add or {}) do
    m[f] = m[f] or {}
    if f == 'drop_tags' then
      local i, rr = find_resolver(m, 'mon_inven')
      if rr and _G.type(rr[1]) == 'table' then
	-- Somewhat Clumsy Hack(TM):  Add to the inventory tag list any tag
	-- in drop_tags that isn't already present.
	local tags = {}
	for _, ff in ipairs(rr[1]) do tags[ff] = true end
	for _, ff in ipairs(mod) do
	  if not tags[ff] then table.insert(rr[1], ff) end
	end
      end
    elseif _G.type(mod) == 'table' then
      for ff, val in pairs(mod) do m[f][ff] = val end
    end
  end

  for f, mod in pairs(ego.remove or {}) do
    if f == 'random_move' then
      local rm = m.ai_state.random_move or 0
      if mod >= 50 and rm >= 50 then rm = rm - 50 end
      if mod % 2 == 1 and rm % 2 == 1 then rm = rm - 25 end
      if rm == 0 then rm = nil end
      m.ai_state.random_move = rm
    elseif f == 'drop_flags' then
      local i, rr = find_resolver(m, 'mon_inven')
      if rr and _G.type(rr[1]) == 'table' then
	for ff, _ in pairs(mod) do rr[1][ff] = nil end
      end
    elseif m[f] and _G.type(mod) == 'table' then
      for ff, _ in pairs(mod) do m[f][ff] = nil end
      if not next(m[f]) then m[f] = nil end
    end
  end

  if ego.talents then
    local i, rr = find_resolver(m, 'talents')
    if ego.talents.talent_in then
      m.ai_state.talent_in = ego.talents.talent_in
    end
    if ego.talents.unlearn_existing and not ego.talents.add then
      if i then table.remove(m, i) end
    end
    if ego.talents.add then
      if not rr then
	table.insert(m, resolvers.talents(ego.talents.add))
      else
	if ego.talents.unlearn_existing then rr[1] = {} end
	for t_id, v in pairs(ego.talents.add) do rr[1][t_id] = v end
      end
    end
  end

  -- Sneaky Hack(TM):  If we just added the EVIL flag, switch to the
  -- Enemies faction.  [There are currently no NPC egos that add the GOOD
  -- or NEUTRAL flag; if we ever add any, that will need a faction change
  -- too.]
  if m:has('EVIL') then m.faction = 'enemies' end

  local ego_part = ego.name:lower()
  if not m.ego_particle or m.ego_particle.args.ego ~= ego_part then
    m:removeParticles(m.ego_particle)
    if fs.exists(Tiles.prefix..'npc/ego/'..ego_part..'.png') then
      local ts = config.settings.tome2.tiles
      m.ego_particle = m:addParticles(engine.Particles.new('npc_ego', 1, {tileset=ts, ego=ego_part, size=16}))
    else
      self.ego_particle = nil
    end
  end
end
