-- $Id: Player.lua 4259 2021-08-02 23:13:02Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2021 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require "engine.class"
require "mod.class.Actor"
require 'mod.class.interface.PlayerCombat'
require 'mod.class.interface.PlayerSkills'
require 'mod.class.interface.PlayerFates'
require 'mod.class.interface.PlayerSanity'
require 'mod.class.interface.PlayerAlchemy'
require 'mod.class.interface.PlayerLore'
require "engine.interface.PlayerRest"
require "engine.interface.PlayerRun"
require "mod.class.interface.PlayerExplore"
require "engine.interface.PlayerMouse"
require "engine.interface.PlayerHotkeys"
local Map = require "mod.class.Map"
local Tiles = require 'engine.Tiles'
local Dialog = require 'engine.ui.Dialog'
local DeathDialog = require "mod.dialogs.DeathDialog"
local TacticExplorDialog = require 'mod.dialogs.TacticExplorDialog'
local Astar = require"engine.Astar"
local DirectPath = require "engine.DirectPath"
local DamageType = require 'engine.DamageType'
local Birther = require 'engine.Birther'
local Object = require 'mod.class.Object'
local ShowInvenFloor = require 'mod.dialogs.ShowInvenFloor'
local ShowEquipInvenFloor = require 'mod.dialogs.ShowEquipInvenFloor'
local ShowPickupFloor = require 'mod.dialogs.ShowPickupFloor'
local Automatizer = require 'mod.class.Automatizer'

--- Defines the player
-- It is a normal actor, with some redefined methods to handle user interaction.<br/>
-- It is also able to run and rest and use hotkeys
module(..., package.seeall, class.inherit(mod.class.Actor,
					  mod.class.interface.PlayerCombat,
					  mod.class.interface.PlayerSkills,
					  mod.class.interface.PlayerFates,
					  mod.class.interface.PlayerSanity,
					  mod.class.interface.PlayerAlchemy,
					  mod.class.interface.PlayerLore,
					  engine.interface.PlayerRest,
					  engine.interface.PlayerRun,
					  mod.class.interface.PlayerExplore,
					  engine.interface.PlayerMouse,
					  engine.interface.PlayerHotkeys))

local exp_chart = {
  [1]	= 0,
  [2]	= 10,
  [3]	= 25,
  [4]	= 45,
  [5]	= 70,
  [6]	= 100,
  [7]	= 140,
  [8]	= 200,
  [9]	= 280,
  [10]	= 380,
  [11]	= 500,
  [12]	= 650,
  [13]	= 850,
  [14]	= 1100,
  [15]	= 1400,
  [16]	= 1800,
  [17]	= 2300,
  [18]	= 2900,
  [19]	= 3600,
  [20]	= 4400,
  [21]	= 5400,
  [22]	= 6800,
  [23]	= 8400,
  [24]	= 10200,
  [25]	= 12500,
  [26]	= 17500,
  [27]	= 25000,
  [28]	= 35000,
  [29]	= 50000,
  [30]	= 75000,
  [31]	= 100000,
  [32]	= 150000,
  [33]	= 200000,
  [34]	= 275000,
  [35]	= 350000,
  [36]	= 450000,
  [37]	= 550000,
  [38]	= 700000,
  [39]	= 850000,
  [40]	= 1000000,
  [41]	= 1250000,
  [42]	= 1500000,
  [43]	= 1800000,
  [44]	= 2100000,
  [45]	= 2400000,
  [46]	= 2700000,
  [47]	= 3000000,
  [48]	= 3500000,
  [49]	= 4000000,
  [50]	= 4500000,
}

-- cf. PY_REGEN_* in src/defines.h
local LIFE_REGEN_BASE = 1442
local MANA_REGEN_BASE = 524
local REGEN_NORMAL = 197
local REGEN_WEAK = 98
local REGEN_FAINT = 33

function _M:init(t, no_default)
  t.display=t.display or '@'
  t.color_r=t.color_r or 230
  t.color_g=t.color_g or 230
  t.color_b=t.color_b or 230

  t.player = true
  t.type = t.type or "humanoid"
  t.subtype = t.subtype or "player"
  t.faction = t.faction or "players"

  t.lite = t.lite or 0

  mod.class.Actor.init(self, t, no_default)
  mod.class.interface.PlayerCombat.init(self, t)
  mod.class.interface.PlayerSkills.init(self, t)
  mod.class.interface.PlayerFates.init(self, t)
  mod.class.interface.PlayerSanity.init(self, t)
  mod.class.interface.PlayerAlchemy.init(self, t)
  engine.interface.PlayerHotkeys.init(self, t)

  self.descriptor = {}
  self.max_level = 50
  self.max_exp = self.exp
  self.level_hiwater = self.level
  self.speed = 0
  self.regen_frac = REGEN_NORMAL
  self.hp_mod = 0

  -- Start off just below full.
  self.food = util.FOOD.full - 1
  self.food_regen_mod = 0

  -- We need this until we can fire off the birth_piety() resolver, which
  -- will remove it if we don't need it.
  self.last_piety_regen = 0

  self.death_count = 0

  self.one_turn_fov = {}
end

function _M:getExpChart(level)
  if not level or level < 1 then return 0 end
  return math.floor((exp_chart[level] * (self.xp_mult or 100))/100)
end

-- cf. src/dungeon.c:regenhp()
function _M:regenLife()
  if self.no_life_regen then return end

  local frac = self.regen_frac + (self.life_regen_mod or 0)
  -- Spectacularly Dodgy Hack(TM):  Yavanna prayer increases regeneration
  -- based on WIS, but only if you're standing on grass.
  if self:isPrayingTo('Yavanna') then
    local g = game.level.map(self.x, self.y, Map.TERRAIN)
    if g and g.grassy then
      frac = frac + 200 + self:getStat('WIS', {scale=800})
    end
  end

  local d = self.max_life * frac + LIFE_REGEN_BASE
  self.life = util.bound(self.life + d/65536, self.die_at, self.max_life)
end

-- cf. src/dungeon.c:regenmana()
function _M:regenMana()
  local frac = self.regen_frac
  frac = frac + self:getStatRelated('int_mana_regen')*3

  local d = self.max_mana * frac + MANA_REGEN_BASE
  self.mana = util.bound(self.mana + d/65536, 0, self.max_mana)
end

-- cf. src/dungeon.c lines 1707-1770
function _M:regenFood()
  if self:has('GORGED') then
    -- Digest quickly when gorged.
    self:incFood(-100)
  else
    -- Normal food regenration rate is loosely based on speed
    -- [cf. src/dungeon.c lines 1713-1728]
    local amt = 2*self.global_speed + (self.food_regen_mod or 0)
    self:incFood(-amt)
  end
end

-- We override the ActorResource-generated :incFood() method to handle
-- changing "states".  [cf. set_food() in src/xtra2.c]
function _M:incFood(amt)
  local data = {
    [util.FOOD.starve] = { down_eff = self.EFF_STARVING },
    [util.FOOD.fainting] = { up_msg='You are still weak.', down_msg='You are getting faint from hunger!', down_eff=self.EFF_FAINTING },
    [util.FOOD.weak] = { up_msg='You are still hungry.', down_msg='You are getting weak from hunger!' },
    [util.FOOD.hungry] = { up_msg='You are no longer hungry.', down_msg='You are getting hungry.' },
    [util.FOOD.full] = { up_msg='You are full!', down_msg='You are no longer full.' },
    [util.FOOD.max] = { up_msg='You have gorged yourself!', down_msg='You are no longer gorged.', up_eff=self.EFF_GORGED },
  }
  local changed = false

  local old_food = self.food
  self.food = util.bound(self.food + amt, self.min_food, self.max_food)
  util.check_threshold(old_food, self.food, {
    util.FOOD.starve, util.FOOD.fainting, util.FOOD.weak,
    util.FOOD.hungry, util.FOOD.full, util.FOOD.max,
  }, function(before, after, threshold, thresholds)
    changed = true
    local msg_f = after > before and 'up_msg' or 'down_msg'
    local eff_add_f = after > before and 'up_eff' or 'down_eff'
    local eff_rm_f = after > before and 'down_eff' or 'up_eff'
    if data[threshold][msg_f] then game.log(data[threshold][msg_f]) end
    for _, th in ipairs(thresholds) do
      if data[th][eff_add_f] then
	self:setEffect(data[th][eff_add_f], 1, {})
      end
      if data[th][eff_rm_f] then
	self:removeEffect(data[th][eff_rm_f])
      end
    end
    if after < util.FOOD.weak and game.zone.worldmap then
      -- Drop out of world map.
      -- Questionable Hack(TM):  The CHANGE_LEVEL virtual binding cares
      -- which actual key we used, so we have to fake a keypress.
      game.key.virtuals.CHANGE_LEVEL('>', false, false, false, false, '>', false, nil)
    end
  end)

  self.changed = true
  if changed then
    self:restStop('food level changed')
    self:runStop('food level changed')
  end
  return changed
end

-- A convenience wrapper for the above.
function _M:setFood(v)
  return self:incFood(v - self.food)
end

function _M:getDesc(t)
  t = t or {}
  local desc = ''
  if not t.no_name then
    local sd = game:getSaveDescription()
    local idx = sd.description:find('\n')
    local name = idx and sd.description:sub(1, idx-1) or sd.name
    desc = desc .. name .. '\n\n'
  end
  desc = desc .. '[further player description to follow]'

  return desc
end

function _M:energyToMove()
  local ret = mod.class.Actor.energyToMove(self)
  if game.zone.worldmap then
    -- Moving on the worldmap takes a *lot* more time.
    -- [cf. src/cmd2.c line 2726]
    local map = game.level.map
    ret = ret * math.floor((map.w + map.h)/2)
  end
  -- TODO maybe allow hooks and actor callbacks to influence this?
  return ret
end

function _M:stay()
  if core.key.modState('caps') then
    if game.level then game.level.map:centerViewAround(self.x, self.y) end
  else
    self:useEnergy()
    self:autoSearch()
    self:fireCallbacks('on_rest_step')
  end
end

function _M:moveDir(dir, force)
  -- Slight Hack(TM) for DeathMold movement.  :move() will clean it up for
  -- us.
  self.moving_dir = true
  local moved = mod.class.Actor.moveDir(self, dir, force)
  if moved then self:autoSearch() end

  return moved
end

function _M:move(x, y, force)
  local moved = mod.class.Actor.move(self, x, y, force)
  if moved then
    game.level.map:moveViewSurround(self.x, self.y, 8, 8)

    self:describeFloor(self.x, self.y)
  end

  -- Remember not to describe this grid again.
  self.old_x, self.old_y = self.x, self.y

  return moved
end

function _M:canPushPast(actor)
  -- We can push past if:
  --   - the actor is friendly or neutral to us
  --     [NOTE:  the original T2 code only allowed friendlies here]
  --   - we are not confused, stunned or hallucinating
  --   - we can see the actor
  --   - we and the actor can both move into each other's grid
  -- [cf. src/cmd1.c lines 3478-3481,3499-3500]
  return actor:reactionToward(self) >= 0 and
	 not (self:has('CONFUSED') or self:has('STUN') or self:has('HALLUCINATION')) and
	 self:canSee(actor) and
	 self:canMove(actor.x, actor.y, true) and
	 actor:canMove(self.x, self.y, true)
end

function _M:tryOpenDoor(g, x, y, dry_run)
  if dry_run then
    -- TODO Disallow if we're possessing a monster without the OPEN_DOOR flag.
    return true
  end

  if not g.try_open_door or g:try_open_door(self) then
    g:openDoor(x, y, self)
    self:useEnergy()
    return true
  end
  return false
end

function _M:describeFloor(x, y)
  if self.old_x == self.x and self.old_y == self.y then return end

  -- Add or remove terrain effects depending on the grid.
  local t_effs = self:terrainEffects()
  for id, def in pairs(self.tempeffect_def) do
    if def.terrain_effect then
      if t_effs[id] and not self:hasEffect(id) then
	self:setEffect(id, 1, t_effs[id])
      elseif self:hasEffect(id) and not t_effs[id] then
	self:removeEffect(id)
      end
    end
  end

  if self:has('AUTO_ID') then
    -- Do an auto-ID sweep first.
    game.level.map:objectsApply(x, y, function(o, _, _, _) o:identify(true, self) end)
  end
  Automatizer:checkFloor(x, y)

  -- Auto-pickup stuff from floor.
  local i = 1
  local obj = game.level.map:getObject(x, y, i)
  while obj do
    if util.getval(obj.auto_pickup, obj, self) and self:pickupFloor(i, true) then
      -- Nothing to do.
    else
      i = i + 1
    end
    obj = game.level.map:getObject(x, y, i)
  end

  -- i is now one higher than the number of objects on the floor; bump it
  -- down.
  i = i - 1

  if i > 0 then
    if config.settings.tome2.floor_pickup == 'Yes' then
      -- Always pick up.
      if i == 1 and not self:okPickupWeight(game.level.map:getObject(x, y, 1)) then
	-- If there's only one item and it's too heavy, ask.
	local o = game.level.map:getObject(x, y, 1)
	Dialog:yesnoPopup('Pick Up Item', ('Pick up %s?'):format(o:getName()),
			  function(ret) if ret then self:pickupFloor(1, true) end end)
      else
	-- Otherwise, just pick up everything until the encumbrance check
	-- squawks.
	while game.level.map:getObject(x, y, 1) do
	  local o = game.level.map:getObject(x, y, 1)
	  if not self:okPickupWeight(o) then
	    game.logPlayer(self, 'Not picking up %s (too heavy).', o:getName())
	    break
	  end
	  if not self:pickupFloor(1, true) then
	    -- Out of inventory room.  pickupFloor() already squawked for
	    -- us.
	    break
	  end
	end
      end
    elseif config.settings.tome2.floor_pickup == 'Ask' then
      -- Prompt to pick up.
      if i == 1 then
	-- For one item, ask directly.
	local o = game.level.map:getObject(x, y, 1)
	Dialog:yesnoPopup('Pick Up Item', ('Pick up %s?'):format(o:getName()),
			  function(ret) if ret then self:pickupFloor(1, true) end end)
      else
	-- For multiple items, prompt to go to a pickup-from-floor dialog.
	local d
	local all_cb = function(o, item)
	  self:pickupFloor(item, true)
	  d:used()
	end
	local yn_cb = function(ret)
	  if ret then d = self:showPickupFloor(self:weightDynamicTitle('Pickup'), nil, all_cb) end
	end
	Dialog:yesnoPopup('Pick Up Items', ('Pick up %d items?'):format(i), yn_cb)
      end
    else
      -- No pick up.
      if i == 1 then
	game.log('On floor:  %s', game.level.map:getObject(x, y, 1):getName())
      else
	game.log('There are %d objects here.', i)
      end
    end
  end

  local g = game.level.map(x, y, Map.TERRAIN)
  if g and g.change_level then
    local name = g:getName(x, y)
    game.logPlayer(self, '#YELLOW_GREEN#There is %s here.#LAST#', name:a_an())
    self:addFlyer(60, 0, -1.5, ('Level change (%s)!'):format(name),
		  colors.simple(colors.YELLOW_GREEN), true)
  end
  if g and g.enter_msg then
    game.logPlayer(self, '#LIGHT_BLUE#%s#LAST#', g.enter_msg)
  end

  -- Ambush!
  if g and g.zone_level and rng.percent(g.zone_level - self.level*2) then
    self:setEffect(self.EFF_WILD_AMBUSH, 1, {})
    game:changeLevel(1, g.terrain_type == 'TOWN' and g.change_zone or 'wilderness', { entry_point={x=99,y=33} })
  end
end

-- Returns a map of the terrain effects we should have by virtue of being
-- on the current grid and the corresponding parameters to use if we need
-- to add them.
function _M:terrainEffects()
  local effs = {}

  if game.level.map.attrs(self.x, self.y, 'det_trap') then
    effs[self.EFF_DET_TRAPS] = {}
  end
  local g = game.level.map(self.x, self.y, Map.TERRAIN)
  if g.effect and (not g.effect_precond or g:effect_precond(self)) then
    effs[g.effect] = g.effect_params or {}
  end
  self:fireCallbacks('on_terrain_effect', self.x, self.y, g, effs)

  return effs
end

-- Slight Hack(TM):  God descriptors are indexed by the full god name with
-- all the fancy accented characters (ie. 'Manwë Súlimo'), but they also
-- define aliases like short name (ie. 'Manwe') or unaccented full name
-- (ie. 'Manwe Sulimo') for :isFollowing() and :isPrayingTo() below.  We
-- collect and cache those here.
local god_aliases = {}
local function god_by_alias(god)
  if not next(god_aliases) then
    local Birther = require('engine.Birther')
    for _, d in ipairs(Birther.birth_descriptor_def.god) do
      for _, alias in ipairs(d.aliases or {}) do
	god_aliases[alias] = d.name
      end
    end
  end
  return god_aliases[god] or god
end

function _M:isFollowing(god)
  god = god_by_alias(god)
  return self.descriptor.god == god
end

function _M:isPrayingTo(god)
  god = god_by_alias(god)
  if self.descriptor.god ~= god then return false end
  local d = Birther:getBirthDescriptor('god', self.descriptor.god)
  return d.prayer_talent and self.sustain_talents[d.prayer_talent]
end

function _M:isPraying()
  return self:isPrayingTo(self.descriptor.god)
end

function _M:inBearForm()
  local eff = self:hasEffect(self.EFF_MIMIC_FORM)
  return eff and eff.mimic and eff.mimic.name == 'Bear'
end

function _M:abandonVala(god)
  if not self:isFollowing(god) then return end
  -- Uninstall any callbacks.
  local d = Birther:getBirthDescriptor('god', self.descriptor.god)
  if d.callbacks then self:uninstallCallbacks(d, 'callbacks', d.type) end
  -- The 'Nobody' birth descriptor doesn't have any callbacks, so we don't
  -- have to install them.
  self.descriptor.god = 'Nobody'
  self.piety = 0
  self.last_piety_regen = nil
  self:recalcEverything()
end

function _M:act()
  self:stepOneTurnFOV()
  game:tweakVisuals()

  if not mod.class.Actor.act(self) then return end

  self:updateMainShader()

  -- Clean log flasher
  game.flash:empty()

  -- Resting? Running? Paralyzed? Otherwise pause
  if not self:hasEffect(self.EFF_PARALYZED) and not self:restStep() and not self:runStep() and self.player then
    game.paused = true
  end
end

-- [NOTE TO SELF:  Preparation for 1.6, in which engine.Actor:useEnergy()
--  no longer does the game.paused fiddling done below.  AFAICT it's
--  harmless to do this twice while we're still on the 1.5.x engine.
--  cf. <https://git.net-core.org/tome/t-engine4/commit/be857c398e1fe1e91e53b9a21ab483c8ca7da9d7>]
function _M:useEnergy(val)
  mod.class.Actor.useEnergy(self, val)
  if self.player and self.energy.value < game.energy_to_act then
    game.paused = false
  end
end

function _M:on_global_turn()
  if game.state.birth.fates and self.level > 10 and rng.chance(50000) then
    self:gainFate()
  end
end

-- "Superloaded" from engine.interface.PlayerHotkeys to add hotkey feedback
-- animation, as per the T4 module.
function _M:activateHotkey(id)
  local zone = table.get(game, 'hotkeys_display', 'clics', id)
  if config.settings.tome2.visual_hotkeys and zone and self.hotkey[id] then
    local px = game.hotkeys_display.display_x + zone[1] + zone[3]/2
    local py = game.hotkeys_display.display_y + zone[2] + zone[4]/2
    game:addParticle(px, py, 'hotkey_feedback', {w=zone[3], h=zone[4]})
  end

  return engine.interface.PlayerHotkeys.activateHotkey(self, id)
end

-- Handle thaumaturgy spell hotkeys.
function _M:hotkeyThaumaturgy(idx)
  local sp = self.thaumaturgy_spells and self.thaumaturgy_spells[idx]
  if sp then spell.thaumaturgy.cast(self, sp) end
end

-- Precompute FOV form, for speed
local fovdist = {}
for i = 0, 30 * 30 do
  fovdist[i] = math.max((20 - math.sqrt(i)) / 14, 0.6)
end
local wild_fovdist = {}
for i = 0, 30 * 30 do
  wild_fovdist[i] = math.max((5 - math.sqrt(i)) / 1.4, 0.6)
end

function _M:playerFOV()
  -- Clean FOV before computing it
  game.level.map:cleanFOV()

  if game.zone.worldmap then
    -- Slight Hack(TM):  On the worldmap, just process radius-2 vision.
    self:computeFOV(2, "block_sight", function(x, y, dx, dy, sqdist)
      game.level.map:applyLite(x, y, wild_fovdist[sqdist])
    end, true, true, true)
    return
  end

  -- Compute ESP FOV, using cache
  if self.esp and next(self.esp) then
    self:computeFOV(20, 'block_esp', function(x, y)
      game.level.map:applyESP(x, y, 0.6)
    end, true, true, true)
  end

  if self:has('BLIND') then
    -- Compute FOV even if we're blind; apparently NPCs use it to target us.
    self:computeFOV(self.sight or 20, "block_sight")
  else
    -- Compute both the normal and the lite FOV, using cache
    self:computeFOV(self.sight or 20, "block_sight", function(x, y, dx, dy, sqdist)
      game.level.map:apply(x, y, fovdist[sqdist])
    end, true, false, true)
    self:computeFOV(self.lite, "block_sight", function(x, y, dx, dy, sqdist)
      game.level.map:applyLite(x, y)
    end, true, true, true)
    -- Apply light effects for actors with the HAS_LITE flag.
    for _, e in pairs(game.level.entities) do
      if e ~= self and e.computeFOV and e:has('HAS_LITE') then
	e:computeFOV(1, 'block_sight', function(x, y, dx, dy, sqdist)
	  game.level.map:applyExtraLite(x, y, fovdist[sqdist])
	end, true, true)
      end
    end
  end

  -- All the one-turn FOV mods are magical, so I think we're okay doing
  -- them even when blind.
  for _, mod in ipairs(self.one_turn_fov) do
    spell.one_turn_fov_actor = mod.actor
    self:computeFOV(mod.r, mod.block, mod.callback, true, true, true)
    spell.one_turn_fov_actor = nil
  end
end

-- Used to set FOV behavior for "one turn" (ie. until the next time the
-- player acts).
function _M:setOneTurnFOV(t)
  t.actor, t.pass = self, 1
  self.one_turn_fov[#self.one_turn_fov+1] = t
  game.level.map.changed = true
end

-- Manage flags set by :setOneTurnFOV().
function _M:stepOneTurnFOV()
  -- Clumsy Hack(TM):  Our "one-turn" FOV mods actually have to survive one
  -- call to Player:act() and be removed on the second.  To that end, we
  -- keep a "pass" count on each mod, discarding it when it hits pass 3.
  if #self.one_turn_fov > 0 then
    local new_fov = {}
    for _, t in ipairs(self.one_turn_fov) do
      t.pass = t.pass + 1
      if t.pass <= 2 then new_fov[#new_fov+1] = t end
    end
    self.one_turn_fov = new_fov
    -- Step monsters with the DETECTED effect.  If we've done this right,
    -- one or more monsters will have this effect when and only when one or
    -- more FOV mods are present.
    for _, e in pairs(game.level.entities) do
      local eff = e.hasEffect and e:hasEffect(e.EFF_DETECTED)
      if eff then
	eff.pass = eff.pass + 1
	if eff.pass >= 3 then e:removeEffect(e.EFF_DETECTED) end
      end
    end
  end
end

--- Called before taking a hit, overload mod.class.Actor:onTakeHit() to stop resting and running
function _M:onTakeHit(value, src)
  self:runStop("taken damage")
  self:restStop("taken damage")

  local ret = mod.class.Actor.onTakeHit(self, value, src)
  if self.life - ret < self.max_life * config.settings.tome2.hp_warning_pct/100 then
    self:addFlyer(30, (rng.range(0,2)-1) * 0.5, 2, "LOW HEALTH!", {255,0,0}, true)
  end

  -- Hit direction warning
  if src.x and src.y and (self.x ~= src.x or self.y ~= src.y) then
    local range = core.fov.distance(src.x, src.y, self.x, self.y)
    if range > 1 then
      local angle = math.atan2(src.y - self.y, src.x - self.x)
      game.level.map:particleEmitter(self.x, self.y, 1, "hit_warning", {angle=math.deg(angle)})
    end
  end

  return ret
end

function _M:die(src)
  if self.game_ender then
    engine.interface.ActorLife.die(self, src)

    -- Record who killed us.
    local ssrc = src.resolveSource and src:resolveSource() or src
    if ssrc.recordPlayerKill then ssrc:recordPlayerKill() end

    game.paused = true
    self.energy.value = game.energy_to_act
    game:registerDialog(DeathDialog.new(self))
  else
    mod.class.Actor.die(self, src)
  end
end

function _M:setName(name)
  self.name = name
  game.save_name = name
end

-- Annoying Hack(TM):  All the other talent-use callbacks we can hook into
-- happen too early for our purposes [before self.sustain_talents{} has
-- been updated, for instance], so we hijack the whole method and do our
-- recalcEverything() call afterward.
function _M:useTalent(id, who, force_level, ignore_cd, force_target, silent, no_confirm)
  local success = engine.interface.ActorTalents.useTalent(self, id, who, force_level, ignore_cd, force_target, silent, no_confirm)

  if success then
    -- Our activated talents can get away with doing their own
    -- recalcEverything(), I think.  Also, if we're called from
    -- recalcEverything() ourself, assume it'll take care of any
    -- consequences and don't call it recursively.
    local t = self:getTalentFromId(id)
    if t.mode == 'sustained' and not self.recalcing_everything then
      self:recalcEverything()
    end
  end

  return success
end

function _M:postUseTalent(ab, ret)
  local rret = mod.class.Actor.postUseTalent(self, ab, ret)

  Automatizer:checkInventory(self)
  Automatizer:checkFloor(self.x, self.y)

  return rret
end

--- Notify the player of available cooldowns
function _M:onTalentCooledDown(tid)
  local t = self:getTalentFromId(tid)

  self:addFlyer(30, -0.3, -3.5, ("%s available"):format(t.name:capitalize()), {0,255,00})
  game.log("#00ff00#Talent %s is ready to use.", t.name)
end

function _M:levelup(down)
  local newlev = self:levelupMsg(down)

  self:recalcEverything()

  if newlev then
    self.skill_points = self.skill_points + 6
    if not self.silent_levelup then
      game.log("#00FFFF#%s has %d skill points to spend.  Press '%s' to use them.#LAST#", self.name, self.skill_points, util.getFormattedKeyBinds('LEVELUP', "' or '"))
    end
  end

  return newlev
end

--- Tries to get a target from the user
function _M:getTarget(typ)
  return game:targetGetForPlayer(typ)
end

--- Sets the current target
function _M:setTarget(target)
  return game:targetSetForPlayer(target)
end

function _M:canSeeInvis(target)
  local sq_inf = self.infravision and self.infravision * self.infravision or 0
  if self.fov.actors[target] and target.learnFlags then
    -- Learn flags if we have LOS.
    if self:has('SEE_INVIS') then
      target:learnFlags(true, 'flags', 'INVISIBLE')
    end
    if self.fov.actors[target].sqdist <= sq_inf then
      target:learnFlags(true, 'flags', 'COLD_BLOOD')
    end
  end
  return target:hasEffect(target.EFF_DETECTED) or self:has('SEE_INVIS') or self.fov.actors[target] and self.fov.actors[target].sqdist <= sq_inf
end

local function spotHostiles(self)
  if not self.x then return {} end
  local seen = {}

  -- Check for visible monsters, only see LOS actors, so telepathy wont prevent resting
  local function opaque_cb(_, x, y) return game.level.map:opaque(x, y) end
  local function see_cb(_, x, y)
    local actor = game.level.map(x, y, game.level.map.ACTOR)
    if actor and self:reactionToward(actor) < 0 and self:canSee(actor) and game.level.map.seens(x, y) then
      seen[#seen+1] = {
	x = x,
	y = y,
	actor = actor,
	entity = actor,
	name = actor:getName(),
      }
    end
  end
  core.fov.calc_circle(self.x, self.y, game.level.map.w, game.level.map.h, 20, opaque_cb, see_cb, nil)
  return seen
end

-- Overloaded from PlayerRest:restStep() to call our on_rest_step()
-- callbacks.
function _M:restStep()
  if not self.resting then return false end

  local ret, msg = self:restCheck()
  if ret and self.resting and self.resting.rest_turns and self.resting.cnt > self.resting.rest_turns then
    ret = false
    msg = nil
  end
  if not ret then
    self:restStop(msg)
    return false
  else
    self:useEnergy()
    self.resting.cnt = self.resting.cnt + 1
    self:fireCallbacks('on_rest_step')
    return true
  end
end

--- Can we continue resting ?
-- We can rest if no hostiles are in sight, and if we need life/mana/stamina (and their regen rates allows them to fully regen)
function _M:restCheck()
  local actually_resting = self.resting.past == _t'rested'
  if actually_resting and game.level.map:checkEntity(self.x, self.y, Map.TERRAIN, 'no_rest') then
    local g = game.level.map(self.x, self.y, Map.TERRAIN)
    return false, 'cannot rest on ' .. g:getName(self.x, self.y)
  end

  local seen = spotHostiles(self)
  if #seen > 0 then
    for _, node in ipairs(seen) do
      node.entity:addParticles(engine.Particles.new('notice_enemy', 1))
    end
    local dir = game.level.map:compassDirection(seen[1].x - self.x, seen[1].y - self.y)
    local onscreen = game.level.map:isOnScreen(seen[1].x, seen[1].y)
    local msg = ('hostile spotted to the %s (%s%s)'):format(dir or '???', seen[1].name, onscreen and '' or ' - offscreen')
    return false, msg
  end

  if self.resting.rest_turns then
    -- Slight Hack(TM):  We re-use the resting infrastructure to repeat
    -- certain actions like attempting to open locked doors.
    if self.repeat_action then return self:repeat_action() end
    -- Otherwise, if the player requested resting for a specific number of
    -- turns, don't stop until we're done with that.
    return true
  end

  local p = {}
  self:fireCallbacks('on_rest_check', p)
  if p.stop then
    -- Some callback requested emergency rest stop.
    return false, p.stop_msg
  elseif p.stop == false then
    -- Some callback requested continued resting.
    return true
  end

  -- Check resources, make sure they CAN go up, otherwise we will never stop
  -- [NOTE:  The only way life/mana regen can go negative is if the player
  --  has the DRAIN_LIFE/DRAIN_MANA flags, and those AFAICT don't actually
  --  prevent resting, though they do their own check to stop resting if
  --  player life/mana hits zero.]
  if self.life < self.max_life then return true end
  if self.mana and self.mana < self.max_mana then return true end

  -- Wait for symbiote to finish regenerating.
  -- TODO Maybe do this two-step, the way T4 does?
  local sym_inv = self:getInven(self.INVEN_SYMBIOTE)
  local sym_o = sym_inv and sym_inv[1]
  local symb = sym_o and sym_o.symbiote
  if symb and symb.life < symb.max_life then return true end

  return false, "all resources and life at maximum"
end

function _M:onRestStart()
  -- Slight Hack(TM):  Some repeat_action() methods need extra data, passed
  -- in via self.rest_info{}.  We move it into self.resting{} here to make
  -- it accessible to repeat_action() [and to pre-clean it up].
  self.resting.check_info = self.rest_info
  self.rest_info = nil
  self:fireCallbacks('on_rest_start')
end

function _M:onRestStop()
  self:fireCallbacks('on_rest_stop')
  -- Remove any repeating action.
  self.repeat_action = nil
end

-- Basically a duplicate of engine.interface.PlayerRun:checkRunDir(),
-- except that we check 'block_run' instead of 'block_move' for grids.
function _M:checkRunDir(dir, dist)
	dist = dist or 1
	local x, y = self.x, self.y
	for i = 1, dist do x, y = util.coordAddDir(x, y, dir) end
	-- don't treat other actors as terrain or as something to notice
	-- (let the module handle this)
	if game.level.map(x, y, game.level.map.ACTOR) and not game.level.map:checkEntity(x, y, game.level.map.TERRAIN, "block_move") then
	  return false
	end
	return (game.level.map:checkAllEntities(x, y, "block_run", self) or not game.level.map:isBound(x, y)) and true or false
end

--- Can we continue running?
-- We can run if no hostiles are in sight, and if we no interesting terrains are next to us
function _M:runCheck()
  local seen = spotHostiles(self)
  if #seen > 0 then
    local dir = game.level.map:compassDirection(seen[1].x - self.x, seen[1].y - self.y)
    local onscreen = game.level.map:isOnScreen(seen[1].x, seen[1].y)
    local msg = ('hostile spotted to the %s (%s%s)'):format(dir or '???', seen[1].name, onscreen and '' or ' - offscreen')
    return false, msg
  end

  -- Notice any noticeable terrain
  local noticed = false
  self:runScan(function(x, y, dir)
    -- Stop at border of town/wilderness zone.
    if dir == 'ahead' and game.level.map.attrs(x, y, 'wild_border') then
      noticed = noticed or 'wilderness border'
      return
    end
    -- Stop on adjacent interesting terrains
    local grid = game.level.map(x, y, Map.TERRAIN)
    if grid and grid.notice then
      noticed = noticed or "interesting terrain"
      return
    end
    -- Stop on leaving trap-detected zone if configured.
    if config.settings.tome2.disturb_detect then
      if game.level.map.attrs(self.x, self.y, 'det_trap') and not game.level.map.attrs(x, y, 'det_trap') then
	noticed = noticed or 'leaving trap-detected area'
	return
      end
    end
    -- Stop on interesting objects.
    if dir == 'self' and not game.level.map.attrs(x, y, 'obj_seen') and game.level.map:getObject(x, y, 1) then
      game.level.map.attrs(x, y, 'obj_seen', true)
      noticed = noticed or 'object seen'
      return
    end
  end)
  if noticed then return false, noticed end

  self:playerFOV()

  return engine.interface.PlayerRun.runCheck(self)
end

function _M:runMoved()
  -- Auto-search each step while running, and stop if we find something.
  local found_something, msg = self:autoSearch()
  if found_something then self:runStop(msg) end
end

function _M:runStopped()
  game.level.map.clean_fov = true
  self:playerFOV()
  local seen = spotHostiles(self)
  if #seen > 0 then
    for _, node in ipairs(seen) do
      node.entity:addParticles(engine.Particles.new('notice_enemy', 1))
    end
  end

  -- If we stopped on an object, mark it as seen.
  if game.level.map:getObject(self.x, self.y, 1) then
    game.level.map.attrs(self.x, self.y, 'obj_seen', true)
  end
end

--- Move with the mouse
function _M:mouseMove(tmx, tmy)
  local function hostile_cb()
    local seen = spotHostiles(self)
    return #seen > 0
  end
  return engine.interface.PlayerMouse.mouseMove(self, tmx, tmy, hostile_cb)
end

function _M:canLearnTalent(t, offset)
  -- The player doesn't learn talents directly, getting them instead from
  -- skill levels and equipment; the T-Engine's talent requirement infra-
  -- structure is thus superfluous.
  return true
end

function _M:playerSetPetName(a)
  if not a or a.unique or a.faction ~= 'pets' then return end
  local GetText = require 'engine.dialogs.GetText'
  local cur_name = a:getName{indef_art=true, cite_given_name=true, no_your=true}
  local function cb(text)
    if text and text:match('^%s*$') then text = nil end
    a:setName(text)
  end
  local d = GetText.new('Set Name:  '..cur_name, 'New name (or empty to clear)', 0, 50, cb)
  if a.given_name then d.c_box:setText(a.given_name) end
  game:registerDialog(d)
end

-- cf. do_cmd_open_chest(), do_cmd_open_aux(), easy_open_door(),
-- do_cmd_disarm_chest(), do_cmd_disarm_aux(), all in src/cmd2.c
function _M:getDisarmPower()
  local chance = self.skill_disarm
  if self:hasEffect(self.EFF_BLIND) or not game.level.map.seens(self.x, self.y) then
    chance = math.floor(chance/10)
  end
  if self:hasEffect(self.EFF_CONFUSED) or not self:hasEffect(self.EFF_HALLUCINATION) then
    chance = math.floor(chance/10)
  end
  return chance
end

function _M:playerOpen()
  local function match(x, y)
    return game.level.map:checkAllEntities(x, y, 'is_openable')
  end
  local tx, ty = util.select_adjacent(self, match, 'You see nothing here to open.', 'There is nothing to open there.')
  if not tx or not ty then return end
  local g = game.level.map(tx, ty, Map.TERRAIN)
  local a = game.level.map(tx, ty, Map.ACTOR)
  local oo = game.level.map:checkAllEntitiesNoStop(tx, ty, 'is_openable')
  -- Slight Hack(TM):  Get one of the chests out of oo.
  local o = nil
  for e, ok in pairs(oo) do
    if ok and e:getEntityKind() == 'object' and e.type == 'misc' and e.subtype == 'chest' then
      o = e
      break
    end
  end
  if not g.door_opened and not o then
    game.log('You see nothing there to open.')
    return
  elseif a then
    game.log('There is a monster in the way!')
    -- Slight Hack(TM):  Move in the direction of the monster to attack it.
    local dir = (self.x - tx + 1)*3 + ty - self.y + 2
    self:moveDir(dir)
  elseif o then
    -- Slight Hack(TM):  We hijack the resting infrastructure to try this
    -- multiple times.
    self.repeat_action = function(self)
      local success, can_retry, msg = o:try_open(tx, ty, self)
      if success then return false, nil end
      return can_retry, msg
    end
    self:restInit(99, 'attempting to open chest', 'attempted to open chest')
  elseif g and g.door_opened then
    if not g.try_open_door then
      -- Door is not locked or jammed; just open it.
      g:openDoor(tx, ty, self)
      self:useEnergy()
      return
    end
    -- Slight Hack(TM):  As above.
    self.repeat_action = function(self)
      local success, can_retry = g:try_open_door(self)
      if success then
	g:openDoor(tx, ty, self)
	return false, nil
      end
      return can_retry, nil
    end
    self:restInit(99, 'attempting to unlock door', 'attempted to unlock door')
  end
end

function _M:playerCloseDoor()
  local function match(x, y)
    return game.level.map:checkEntity(x, y, Map.TERRAIN, 'door_closed')
  end
  local tx, ty = util.select_adjacent(self, match, 'There are no open doors here.', 'There is no open door there.')
  if not tx or not ty then return end
  local g = game.level.map(tx, ty, Map.TERRAIN)
  local a = game.level.map(tx, ty, Map.ACTOR)
  if not g.door_closed then
    game.log('There is no open door there.')
    return
  elseif a then
    game.log('There is a monster in the way!')
    -- Slight Hack(TM):  Move in the direction of the monster to attack it.
    local dir = (self.x - tx + 1)*3 + ty - self.y + 2
    self:moveDir(dir)
  else
    g:closeDoor(tx, ty)
    self:useEnergy()
  end
end

-- cf. do_cmd_bash() and do_cmd_bash_aux() in src/cmd2.c
function _M:playerBashDoor()
  local function match(x, y)
    return game.level.map:checkEntity(x, y, Map.TERRAIN, 'door_opened')
  end
  local tx, ty = util.select_adjacent(self, match, 'There is nothing to bash here.', 'There is nothing to bash there.')
  if not tx or not ty then return end
  local g = game.level.map(tx, ty, Map.TERRAIN)
  local a = game.level.map(tx, ty, Map.ACTOR)
  if not g.door_opened then
    game.log('You see nothing there to bash.')
    return
  elseif a then
    game.log('There is a monster in the way!')
    -- Slight Hack(TM):  Move in the direction of the monster to attack it.
    local dir = (self.x - tx + 1)*3 + ty - self.y + 2
    self:moveDir(dir)
  else
    game.log('You smash into the door!')
    local power = g.door_jam_power or g.door_lock_power or 0
    local chance = math.max(self:getStatRelated('str_bash') - power*10, 1)
    if rng.percent(chance) then
      game.log('The door crashes open!')
      -- 50% chance of breaking the door.
      local repl = rng.chance(2) and 'BROKEN_DOOR' or g.door_opened
      g:openDoor(tx, ty, self, repl)
      -- This will use energy for us.
      self:move(tx, ty)
    else
      local save = self:getStatRelated('dex_safe') + self.level
      if rng.percent(save) then
	game.log('The door holds firm.')
	self:useEnergy()
	-- TODO allow repeat command
      else
	self:setEffect(self.EFF_PARALYZED, 2 + rng.range(1, 2), { merge='combine', msg='You are off-balance.' })
	self:useEnergy()
	-- TODO stop repeat command
      end
    end
  end
end

function _M:playerDisarm(force_trigger)
  local function match(x, y)
    return game.level.map:checkEntity(x, y, Map.TRAP, 'known_by', self) or
	   game.level.map:checkAllEntities(x, y, 'known_trap', self)
  end
  local act = force_trigger and 'trigger' or 'disarm'
  local no_match_msg = 'You see nothing to '..act..' here.'
  local bad_choice_msg = 'You see nothing to '..act..' there.'
  local tx, ty = util.select_adjacent(self, match, no_match_msg, bad_choice_msg)
  if not tx or not ty then return end
  local tr = game.level.map(tx, ty, Map.TRAP)
  local a = game.level.map(tx, ty, Map.ACTOR)
  local oo = game.level.map:checkAllEntitiesNoStop(x, y, 'known_trap', self)
  -- Slight Hack(TM):  Get one of the chests out of oo.
  local o = nil
  for e, _ in pairs(oo) do
    if e:getEntityKind() == 'object' and util.getval(e.known_trap, e, self) then
      o = e
      break
    end
  end
  local power = self:getDisarmPower()
  if (not tr or not tr:knownBy(self)) and not o then
    game.log(bad_choice_msg)
  elseif a then
    game.log('There is a monster in the way!')
    -- Slight Hack(TM):  Move in the direction of the monster to attack it.
    local dir = (self.x - tx + 1)*3 + ty - self.y + 2
    self:moveDir(dir)
  elseif o then
    if force_trigger then
      o.chest.trap:trigger(tx, ty, self)
      return
    end
    -- Slight Hack(TM):  We hijack the resting infrastructure to try this
    -- multiple times.
    self.repeat_action = function(self)
      if o.chest.trap:disarm(tx, ty, self) then
	game.log('You have disarmed the chest.')
	return false, 'successfully disarmed trap on chest'
      end
      -- Did we trigger the trap while trying to disarm it?
      if rng.range(1, power) > 5 then
	-- No, we can keep trying.
	return true, nil
      end
      -- Uh-oh...
      o.chest.trap:trigger(tx, ty, self)
      return false, 'triggered chest trap while trying to disarm it'
    end
    self:restInit(99, 'attemping to disarm chest', 'attempted to disarm chest')
  elseif tr and tr:knownBy(self) then
    if force_trigger then
      -- Pretend we failed to disarm the trap and move onto it, which will
      -- trigger it.
      self.disarm_failure = true
      self:move(tx, ty)
      self.disarm_failure = nil
      return
    end
    -- Slight Hack(TM):  As above.
    self.repeat_action = function(self)
      if tr:disarm(tx, ty, self) then
	game.log('You have disarmed the %s.', tr:getName())
	-- Signal on_end_cb() below to move us onto the newly disarmed
	-- grid.
	self.resting.disarm_success = true
	return false, 'successfully disarmed trap'
      end
      -- Did we trigger the trap while trying to disarm it?
      if rng.range(1, power) > 5 then
	-- No, we can keep trying.
	return true, nil
      end
      -- Uh-oh... Signal on_end_cb() below to move us onto and trigger the
      -- trap.
      self.resting.disarm_failure = true
      return false, 'triggered trap while trying to disarm it'
    end
    local function on_end_cb(_, _)
      -- Move if we either disarmed or triggered the trap.  In the latter
      -- case, signal Trap:block_move() [q.v.] not to block us.
      if self.resting.disarm_success or self.resting.disarm_failure then
	self.disarm_failure = self.resting.disarm_failure
	self:move(tx, ty)
	self.disarm_failure = nil
      end
    end
    self:restInit(99, 'attemping to disarm trap', 'attempted to disarm trap', on_end_cb)
  end
end

function _M:playerSearch()
  local f = function(qty)
    if qty > 1 then
      self.repeat_action = function(self)
	local success, msg = self:searchOne()
	return not success, msg
      end
      self:restInit(qty-1, 'searching', 'searched')
    end
  end
  local GetQuantity = require 'engine.dialogs.GetQuantity'
  game:registerDialog(GetQuantity.new('Search', 'Turns:', 0, 99, f))
end

function _M:searchOne()
  local prob, success, msg = self.skill_searching, false, nil
  -- TODO Also check for no light here.
  if self:hasEffect(self.EFF_BLIND) then
    prob = math.floor(prob/10)
  end
  if self:hasEffect(self.EFF_CONFUSED) or self:hasEffect(self.HALLUCINATION) then
    prob = math.floor(prob/10)
  end

  local grids = core.fov.circle_grids(self.x, self.y, 1, true)
  for x, yy in pairs(grids) do
    for y, _ in pairs(yy) do
      if rng.percent(prob) then
	-- Find trap
	local t = game.level.map(x, y, Map.TRAP)
	if t and not t:knownBy(self) and self:canSee(t) then
	  t:setKnown(self, true)
	  game.level.map:updateMap(x, y)
	  game.logPlayer(self, 'You have found a trap (%s).', t:getName())
	  success = true
	  msg = 'found a trap'
	end
	-- Find secret door, converting it to a closed door.
	local g = game.level.map(x, y, Map.TERRAIN)
	if g.define_as == 'SECRET_DOOR' then
	  game.level.map(x, y, Map.TERRAIN, game.zone.grid_list.DOOR)
	  game.level.map.lites(x, y, true)
	  game.logPlayer(self, 'You have found a secret door.')
	  success = true
	  msg = 'found a secret door'
	end
	-- Search for traps on chests.  As I read the T2 code (src/cmd1.c
	-- lines 581-610), if there are multiple chests stacked in the
	-- grid, this will find traps on all of them at once.
	local oo = game.level.map:checkAllEntitiesNoStop(x, y, 'on_open_chest')
	for _, o in pairs(oo) do
	  if o.trap and not o.trap:knownBy(self) then
	    o.trap:setKnown(self, true)
	    game.logPlayer(self, 'You have discovered a trap on the chest!')
	    success = true
	    msg = 'found a trap on a chest'
	  end
	end
      end
    end
  end

  return success, msg
end

function _M:autoSearch()
  -- A chance to auto-search on each step, depending on our Perception.
  if self.skill_perception >= 50 or rng.chance(50 - self.skill_perception) then
    self:searchOne()
  end
  -- Always auto-search if we have the Auto-Search sustain on.
  if self:isTalentActive(self.T_AUTO_SEARCH) then
    self:searchOne()
  end
end

function _M:showInvenFloor(title, inven, filter, action)
  if not inven then return end
  local d = ShowInvenFloor.new(title, inven, filter, action, self)
  game:registerDialog(d)
  return d
end

function _M:showEquipInvenFloor(title, filter, action, on_select, t)
  local d = ShowEquipInvenFloor.new(title, self, filter, action, on_select, t)
  game:registerDialog(d)
  return d
end

-- A utility method for removing and returning one object from an inventory
-- or the floor.  Intended for use in callbacks provided to ShowInvenFloor
-- and ShowEquipInvenFloor [cf. :playerUseItem() and :doWear()].
-- t.num can be used to get more than one object.
-- t.vocal_get describes the item removed; t.vocal_left describes the stack
-- left (if any).  Returns the unstacked object and the 'last' return value
-- from Actor:removeObject() or Map:getObject() as appropriate.
function _M:getOneObject(inven, item, floor, t)
  t = t or {}
  local pfx = floor and 'You see' or 'You have'
  local o, oo, last

  if floor then
    local fx, fy = t.floor_x or self.x, t.floor_y or self.y
    oo = game.level.map:getObject(fx, fy, item)
    -- :unstack() takes care of capping this at :getNumber() and defaulting
    -- to 1 for us.
    o, last = oo:unstack(t.num)
    if o and last then game.level.map:removeObject(fx, fy, item) end
  else
    local inv = self:getInven(inven)
    -- As above, since :removeObject() calls :unstack() internally.
    o, last = self:removeObject(inv, item, t.num)
    if not last then oo = inv[item] end
    self:sortInven(self:getInven(inven))
  end

  if o and t.vocal_get then
    local get_pfx = type(t.vocal_get) == 'string' and t.vocal_get or pfx
    game.log("%s %s.", get_pfx, o:getName())
  end
  if o and t.vocal_left then
    if last then
      game.log("%s no more %s.", pfx, o:getName{no_count=true})
    else
      game.log("%s %s.", pfx, oo:getName())
    end
  end

  return o, last
end

function _M:weightDynamicTitle(pfx)
  return function()
    local total_wgt_x10 = self:totalCarriedWeightX10()
    local cap_x10 = self:carryingCapacityX10()
    local pct = math.floor((total_wgt_x10*1000)/(cap_x10/2))
    local color = pct > 1000 and '#ff0000#' or pct > 900 and '#FF8a00#' or pct > 750 and '#fcff00#' or '#00ff00#'
    return ('%s %s(%.1flb, %.1f%% capacity)#LAST#'):format(pfx, color, total_wgt_x10/10, pct/10)
  end
end

function _M:playerWear()
  local inven = self:getInven('INVEN')
  local function filter_cb(o)
    return o:wornInven() and self:getInven(o:wornInven()) and true or false
  end
  local function action_cb(o, item, floor)
    self:doWear(inven, floor, item, o)
    return true
  end
  self:showInvenFloor(self:weightDynamicTitle('Wear/Wield'), inven, filter_cb, action_cb)
end

function _M:playerTakeoff()
  local action_cb = function(o, inven, item)
    self:doTakeoff(inven, item, o)
    return true
  end
  self:showEquipment(self:weightDynamicTitle('Take Off'), nil, action_cb)
end

function _M:playerPickup()
  if game.level.map:getObject(self.x, self.y, 2) then
    local d
    local action = function(o, item)
      self:pickupFloor(item, true)
      self.changed = true
      d:used()
    end
    d = self:showPickupFloor(self:weightDynamicTitle('Pickup'), nil, action)
  else
    self:pickupFloor(1, true)
    self:useEnergy()
    self:recalcEverything()
    self.changed = true
  end
end

-- Duplicated with modification from engine.interface.ActorInventory, to
-- allow picked-up ammo to stack back into the quiver.
function _M:pickupFloor(i, vocal, no_sort)
  local o = game.level.map:getObject(self.x, self.y, i)
  if not o then
    if vocal then game.logSeen(self, "There is nothing to pick up here.") end
    return
  end

  -- Our addition:  If the object being picked up stacks with what's in our
  -- quiver, pick it up into that inventory.
  local inv_id = self.INVEN_INVEN
  local quiver = self:getInven(self.INVEN_QUIVER)
  if quiver and #quiver > 0 and quiver[1]:canStack(o) then
    inv_id = self.INVEN_QUIVER
  end

  local inven = self:getInven(inv_id)
  if not inven then return end
  local prepickup = o:check("on_prepickup", self, i)
  if not prepickup then
    local num, o_name = o:getNumber(), o:getName{do_color=true}
    local ok, slot, ro = self:addObject(inv_id, o)
    if ok then
      local newo = inven[slot] -- get exact object added or stack (resolves duplicates)
      game.level.map:removeObject(self.x, self.y, i)
      if ro then -- return remaining stack to floor
	game.level.map:addObject(self.x, self.y, ro)
	num = num - ro:getNumber()
      end
      if not no_sort then self:sortInven(inv_id) end
      -- Apply checks to whole stack (including already carried) assuming homogeneous stack
      -- num added passed to functions to allow checks on part of the stack
      newo:check("on_pickup", self, num)
      self:check("on_pickup_object", newo, num)

      slot = self:itemPosition(self.INVEN_INVEN, newo, true) or 1
      local letter
      if inv_id == self.INVEN_QUIVER then
	letter = 'quiver'
      else
	letter = ShowPickupFloor:makeKeyChar(slot)
      end

      if vocal then game.logSeen(self, "%s picks up (%s): %s.", self.name:capitalize(), letter, o_name) end
      return inven[slot], num
    else
      if vocal then game.logSeen(self, "%s has no room for: %s.", self.name:capitalize(), o_name) end
      return
    end
  elseif prepickup == "skip" then
    return
  else
    return true
  end
end

function _M:playerDrop()
  local function action(o, inven, item)
    -- In case we click on one of the header lines.
    if not o or not inven or not item then return end
    if o:getNumber() > 1 then
      local n = o:getNumber()
      local function nb_cb(qty)
	qty = util.bound(qty, 1, n)
	self:doDrop(inven, item, nil, qty)
      end
      local GetQuantity = require 'engine.dialogs.GetQuantity'
      game:registerDialog(GetQuantity.new('Drop how many?', '1 to '..n, n, n, nb_cb))
    else
      self:doDrop(inven, item)
    end
  end
  self:showEquipInven(self:weightDynamicTitle('Drop'), nil, action)
end

function _M:playerUseItem(object, item, inven, floor)
  if self.no_inventory_access then return end
  if not game.zone or game.zone.worldmap then
    game.logPlayer(self, "You cannot use items on the world map.") return
  end

  local function use_cb(o, inven, floor, item)
    if not o then return end
    local co = coroutine.create(function()
      if o:hasConfig('ask_use') then
	local text = ([[Use %s?]]):format(wear, o:getName())
	local function ok_cb(ok) who:talentDialogReturn(ok) end
	ret = self:talentDialog(ui_util.yesnoPopup('Use Check', text, ok_cb))
	if not ret then return
	end
      end

      self.changed = true
      -- Questionable Hack(TM):  In case this is a teleporting object that
      -- we're using from the floor and will need to consume afterward,
      -- remember where we were.
      local orig_x, orig_y = self.x, self.y

      local ret = o:use(self, inven, floor, item) or {}
      if not ret.used then return end
      if ret.id then o:identify(true, self) end
      if ret.destroy then
	local repl = o.alchemy_replace_on_use
	local xo = self:getOneObject(inven, item, floor, {vocal_left=true, floor_x=orig_x, floor_y=orig_y})
	if game:hasEntity(xo) then game:removeEntity(xo) end
	if repl and self:getSkillLevel('ALCHEMY') > 0 then
	  Object.resolve_flags.force_power = 0
	  xo = game.zone:makeEntityByName(game.level, 'object', repl)
	  if xo then
	    xo:identify(true, self)
	    self:addOrDropObject('INVEN', xo, nil, nil, {vocal_drop=true})
	    self:sortInven()
	  end
	end
      end
    end)
    local ok, ret = coroutine.resume(co)
    if not ok then
      print(debug.traceback(co))
      if ret then error(ret) end
    end
    return true;
  end

  if object and item then
    return use_cb(object, inven, floor, item)
  else
    self:showEquipInvenFloor('Use which item?', function(o) return o:canUseObject() end, use_cb)
  end
end

function _M:playerEat(object, item, inven, floor)
  if self.no_inventory_access then return end
  if not game.zone or game.zone.worldmap then
    game.logPlayer(self, "You cannot eat food on the world map.") return
  end

  local function eat_cb(o, inven, floor, item)
    if not o or not o.food then return end
    if o.food.add then
      self:incFood(o.food.add)
    elseif o.food.set and o.food.set > self.food then
      self:setFood(o.food.set)
    end
    local id = o.food.bottle or o.food.message
    if o.food.message then game.logPlayer(self, o.food.message) end
    if o.food.on_eat then id = o.food.on_eat(o, self) end
    if id then o:identify(true, self) end
    self:getOneObject(inven, item, floor, {vocal_left=true})
    if o.food.bottle then
      local bottle = game.zone:makeEntityByName(game.level, 'object', 'EMPTY_BOTTLE')
      if o then
	o:identify(true, self)
	self:addOrDropObject('INVEN', o)
	self:sortInven()
      end
    end
    self:useEnergy()
  end

  if object and item then
    return eat_cb(object, inven, floor, item)
  else
    self:showEquipInvenFloor('Eat which item?', function(o) return o.food end, eat_cb)
  end
end

local function icky_weapon(who, o)
  if not o or type(o) ~= 'table' or o.type ~= 'weapon' then return false end
  if who:isFollowing('Eru') and o.edged_weapon and
     not (o.convey_flags and o.convey_flags.BLESSED)
  then
    return true
  end
  if who:getSkillLevel('SORCERY') > 0 and o.define_as ~= 'MAGE_STAFF' then
    return true
  end
  return false
end

-- Returns a list of all worn/wielded items in the same artifact set as the
-- reference object.  'force_with' and 'force_without' can be used to
-- pretend that the reference object is resp. isn't worn/wielded.
-- Notwitstanding the preceding, always returns an empty list unless the
-- reference object is actually an artifact that is a member of an artifact
-- set.
function _M:getSetArtifacts(o, force_with, force_without)
  if not o.set_info then return {} end
  local ret = {}
  if force_with then
    ret[1] = o
    ret[o.define_as] = true
  end
  self:inventoryApplyAll(function(inven, _, oo)
    if inven.worn and oo.set_info and oo.set_info.name == o.set_info.name then
      -- If force_with is true, we've already added the reference object;
      -- if force_without is true, we don't want to include the reference
      -- object.
      if oo ~= o or (not force_with and not force_without) then
	ret[#ret+1] = oo
	ret[oo.define_as] = true
      end
    end
  end)
  return ret
end

function _M:artifactSetChanged(set_pre, set_post)
  local n_pre, n_post = #set_pre, #set_post
  if n_pre == n_post then return end

  local work = {}
  for _, o in ipairs(set_pre) do
    work[o.define_as] = work[o.define_as] or { false, false, o }
    work[o.define_as][1] = o.set_info[n_pre] and n_pre or false
  end
  for _, o in ipairs(set_post) do
    work[o.define_as] = work[o.define_as] or { false, false, o }
    work[o.define_as][2] = o.set_info[n_post] and n_post or false
  end

  local gained, lost, upped, downed = {}, {}, {}, {}
  for id, data in pairs(work) do
    local target = (data[1] and not data[2] and lost) or
		   (data[2] and not data[1] and gained) or
		   (data[1] and data[2] and data[1] > data[2] and downed) or
		   (data[1] and data[2] and data[1] < data[2] and upped)
    if target then target[id] = true end
  end

  local function names(set)
    return table.concatNice(table.mapv(function(id)
      return work[id][3]:getName{no_count=true, no_mods=true, no_pval=true, no_inscr=true}
    end, set), ', ', ' and ')
  end

  if #upped > 0 then
    -- It should be impossible to have some upped with none gained.
    game.logPlayer(self, 'Your %s resonate%s more strongly with your %s!', names(upped), (#upped > 1 and '' or 's'), names(gained))
  elseif #gained > 0 then
    game.logPlayer(self, 'Your %s resonate with each other!', names(gained))
  end
  if #downed > 0 then
    -- Similarly, it should be impossible to have some downed with none lost.
    game.logPlayer(self, 'Your %s resonate%s less strongly as you remove your %s.', names(downed), (#downed > 1 and '' or 's'), names(lost))
  elseif #lost > 0 then
    game.logPlayer(self, 'Your %s stop resonating with each other.', names(lost))
  end
end

-- Loosely copied from T4's Actor:doWear().
function _M:doWear(inven, floor, item, o, dst, force_inven, force_item)
  if self.no_inventory_access then return end
  dst = dst or self

  -- Vaguely Annoying Hack(TM):  Wrap the whole method in a coroutine so
  -- we can implement the ask_wear config, and so that on_canwear() and
  -- actor callbacks can use :talentDialog().
  local function cb()
    if o:hasConfig('ask_wear') then
      local wear = o.type == 'weaon' and 'Wield' or 'Wear'
      local text = ([[%s %s?]]):format(wear, o:getName())
      local function ok_cb(ok) who:talentDialogReturn(ok) end
      ret = self:talentDialog(ui_util.yesnoPopup('Wield/Wear Check', text, ok_cb))
      if not ret then return
      end
    end

    -- Slight Hack(TM):  Anything except ammo we only want to wield one of.
    local no_unstack = o.type and o.type == 'ammo' and true or false

    local oo
    if not no_unstack then
      oo = self:getOneObject(inven, item, floor)
    elseif floor then
      oo = game.level.map:getObject(self.x, self.y, item)
      game.level.map:removeObject(self.x, self.y, item)
    else
      oo, _ = self:removeObject(inven, item, no_unstack)
    end

    -- Vaguely Dodgy Hack(TM):  ActorInventory:wearObject() will try to stack
    -- our wield candidate with what we're currently wielding; we need to
    -- tell Object:canStack() that only ammo can stack at this point.
    Object.stack_wielding = true
    local ro, rs = self:wearObject(oo, true, true, force_inven, force_item) -- removed object and remaining stack if any
    Object.stack_wielding = false
    if ro then
      if type(ro) == "table" then dst:addOrDropObject(floor and self.INVEN_INVEN or inven, ro, true) end -- always give full stack back

      -- Slight Hack(TM):  Message for wielding/replacing bad weapons.
      if icky_weapon(self, oo) then
	game.logPlayer(self, game.flash.BAD, 'You feel uncomfortable wielding this weapon.')
      elseif icky_weapon(self, ro) then
	game.logPlayer(self, game.flash.GOOD, 'You feel more comfortable wielding this weapon.')
      end
      -- Check for artifact set changes.
      local set_pre = self:getSetArtifacts(oo, false, true)
      local set_post = self:getSetArtifacts(oo)
      self:artifactSetChanged(set_pre, set_post)
    else
      if floor then
	game.level.map:addObject(self.x, self.y, oo)
      else
	dst:addOrDropObject(inven, oo, true)
      end
    end
    if type(rs) == "table" then
      local rrs, added, slot
      repeat -- handles a case of stacking limits causing part of a stack to be discarded
	rrs = rs
	added, slot, rs = dst:addOrDropObject(floor and self.INVEN_INVEN or inven, rs)
      until not added or not rs
      if not added then
	game.logPlayer(self, "You had to drop %s due to lack of space.", rrs:getName{do_color = true})
	if rrs and not game.zone.worldmap then game.level.map:addObject(self.x, self.y, rrs) end -- extra stack discarded in worldmap
      end
    end

    self:sortInven()
    util.live_update_inventory_dialogs()
    self:recalcEverything()
    self:useEnergy()
    self.changed = true
  end

  local co = coroutine.create(cb)
  local success, err = coroutine.resume(co)
  if not success then
    if err then error(err) end
    print(debug.traceback(co))
  end
end

function _M:doTakeoff(inven, item, o)
  if self:takeoffObject(inven, item, true) then
    self:addOrDropObject(self.INVEN_INVEN, o, true)
  end
  self:sortInven()
  self:recalcEverything()
  self:useEnergy()
  self.changed = true
end

-- Split this out of :doTakeoff() so that PlayerCombat:oneBlowWithWeapon()
-- can call it after destroying a weapon for hitting a Nazgul.  Also allows
-- us to consolidate with :doDrop() below.
function _M:onTakeoff(o, inven_id)
  mod.class.Actor.onTakeoff(self, o, inven_id)
  -- Slight Hack(TM):  Message for unwielding bad weapons.
  if icky_weapon(self, o) then
    game.logPlayer(self, game.flash.GOOD, 'You feel more comfortable no longer wielding this weapon.')
  end
  -- Check for artifact set changes.
  local set_pre = self:getSetArtifacts(o, true, false)
  local set_post = self:getSetArtifacts(o)
  self:artifactSetChanged(set_pre, set_post)
end

function _M:doDrop(inven, item, on_done, nb)
  -- Vaguely Annoying Hack(TM):  Wrap the whole method in a coroutine so
  -- we can implement the ask_drop config, and so that on_drop() and actor
  -- callbacks can use :talentDialog().
  local function cb()
    local inv = self:getInven(inven)
    local o = inv[item]
    local ret

    if o:hasConfig('ask_drop') then
      local name = o:getName{force_qty=nb}
      local text = ([[Drop %s?]]):format(name)
      local function ok_cb(ok) who:talentDialogReturn(ok) end
      ret = self:talentDialog(ui_util.yesnoPopup('Drop Check', text, ok_cb))
      if not ret then
	if on_done then on_done() end
	return
      end
    end

    if nb == nil or nb >= inv[item]:getNumber() then
      ret = self:dropFloor(inven, item, true, true)
    else
      ret = self:dropFloor(inven, item, true, nb)
    end
    self:sortInven(inven)
    util.live_update_inventory_dialogs()
    self:recalcEverything()
    self:useEnergy()
    self.changed = true
    if on_done then on_done() end
  end

  local co = coroutine.create(cb)
  local success, err = coroutine.resume(co)
  if not success then
    if err then error(err) end
    print(debug.traceback(co))
  end
end

function _M:onAddObject(o)
  mod.class.Actor.onAddObject(self, o)
  if not game.creating_player then
    -- We need to recalc speed delta for carried weight, if nothing else.
    self:recalcEverything()
  end
end

function _M:onWear(o, inven_id)
  mod.class.Actor.onWear(self, o, inven_id)
  -- This is moving an object from one inventory to another, so it won't
  -- trigger a carried-weight change.
  if o.worn_talents and not game.creating_player then
    self:recalcEverything()
  end
end

function _M:onRemoveObject(o, inven_id)
  mod.class.Actor.onRemoveObject(self, o)
  if not game.creating_player then
    -- We need to recalc speed delta for carried weight, if nothing else.
    self:recalcEverything()
  end
end

function _M:gainExp(value, regain)
  -- If our exp has been drained, increase max_exp by a fraction of what
  -- we've gained.
  if value > 0 and self.exp < self.max_exp and not regain then
    self.max_exp = self.max_exp + value/5
  end

  -- Siphon off some of the gained experience for any Becoming artifacts.
  -- TODO Sentient weapons also siphon off exp in this general area, by a
  -- different mechanism.
  local div = 1
  if value > 0 then
    local exp_arts = {}
    self:inventoryApplyAll(function(inven, _, o)
      if inven.worn and o.art_exp then exp_arts[#exp_arts+1] = o end
    end)
    div = 1 + #exp_arts
    for _, o in ipairs(exp_arts) do
      o.art_exp = o.art_exp + 2*value/(div*3)
    end
  end

  mod.class.Actor.gainExp(self, value/div)

  self.max_exp = math.max(self.max_exp, self.exp)
end

function _M:restoreExp()
  if self.exp >= self.max_exp then return false end
  game.logPlayer(self, 'You feel your life energies returning.')
  self:gainExp(self.max_exp - exp, true)
  return true
end

-- cf. weight_limit() in src/xtra1.c
function _M:carryingCapacityX10()
  -- TODO Allow hooks to influence this?
  return self:getStatRelated('str_carry')*100
end

-- cf. calc_total_weight() in src/object2.c
function _M:totalCarriedWeightX10()
  local wgt = 0
  self:inventoryApplyAll(function(_, _, o) wgt = wgt + o:stackWeightX10() end)
  return wgt
end

-- cf. calc_bonuses() in src/xtra1.c and can_carry_heavy() in src/object1.c
function _M:speedDeltaForWeight(total_wgt_x10, cap_x10)
  total_wgt_x10 = total_wgt_x10 or self:totalCarriedWeightX10()
  cap_x10 = cap_x10 or self:carryingCapacityX10()
  print(('[DEBUG] wgt %d vs. cap %d'):format(total_wgt_x10, cap_x10))
  if total_wgt_x10 <= cap_x10/2 then
    print('[DEBUG] punting')
    return 0
  end
  print(('[DEBUG] |%.1f - %.1f|/|%d/10| = %.1f'):format(cap_x10/2, total_wgt_x10, cap_x10/10, math.floor((cap_x10/2 - total_wgt_x10)/math.floor(cap_x10/10))))
  return math.floor((cap_x10/2 - total_wgt_x10)/math.floor(cap_x10/10))
end

-- cf. can_carry_heavy() in src/object1.c
function _M:okPickupWeight(o)
  local cap_x10 = self:carryingCapacityX10()
  local total_wgt_x10 = totalCarriedWeightX10()
  local o_wgt_x10 = o:stackWeightX10()
  local old_spd = self:speedDeltaForWeight(total_wgt_x10, cap_x10)
  local new_spd = self:speedDeltaForWeight(total_wgt_x10 + o_wgt_x10, cap_x10)
  return new_spd < old_spd
end

function _M:on_pickup_object(o, num)
  mod.class.Actor.on_pickup_object(self, o, num)

  -- Slight Hack(TM):  Clear the temporary auto_pickup flag set by
  -- PlayerCombat:ammoDropFloor().
  o:forAllStack(function(so)
    if so.auto_pickup == 'temp' then so.auto_pickup = nil end
  end)
  -- Apply AUTO_ID.
  if self:has('AUTO_ID') then o:identify(true, self) end
  -- Pseudo-ID stuff, if we're not confused.
  if not self:has('CONFUSED') then
    for _, inven in pairs(self.inven) do
      for _, oo in ipairs(inven) do
	if oo:pseudoID() then
	  local name = oo:getName{no_inscr=true}
	  local where = not inven.worn and 'in your pack' or
			    oo.type == 'weapon' and 'you are wielding' or
			    oo.type == 'ammo' and 'in your quiver' or
			    'you are wearing'
	  game.logPlayer(self, 'You feel the %s %s is %s...', name, where, oo.pseudo_id)
	end
      end
    end
    Automatizer:checkInventory(self)
  end

  if o.quest_on_pickup then
    local q = self:hasQuest(o.quest_on_pickup)
    if q and not q:isEnded() then q:check('on_quest_pickup_floor', o) end
  end
  if o.achievement_on_pickup then
    world:gainAchievement(o.achievement_on_pickup, self, o)
  end
end

-- Slight Hack(TM):  We handle piety ourselves, since it only regenerates
-- every N game turns (piety_regen is set to 0).
function _M:regenResources()
  engine.interface.ActorResource.regenResources(self)

  -- NOTE:  piety_regen_step is in engine turns, which are 1/10 of a
  -- normal-speed player turn.
  if self.piety_regen_step then
    while game.turn >= self.last_piety_regen + self.piety_regen_step do
      self.last_piety_regen = self.last_piety_regen + self.piety_regen_step
      -- Piety doesn't regenerate if we're resting or on the worldmap;
      -- it also doesn't regenerate in town or wilderness, except for Eru.
      local ok = not self.resting and not game.zone.worldmap and
		 (self:isFollowing('Eru') or
		  (not game.zone.town and not game.wilderness))
      if ok then self:incPiety(self:piety_regen_f()) end
    end
  end
end

-- Vaguely Annoying Hack(TM):  Since piety can affect stats, speed and
-- flags, we have to do a recalcEverything() when it changes.  This isn't
-- so bad, since normal piety regeneration only happens every N game turns,
-- and recalcing when we gain/lose piety for a monster kill or other
-- one-time event isn't expected to be too onerous.
function _M:incPiety(d)
  mod.class.Actor.incPiety(self, d)
  self:recalcEverything()
end

-- A convenience method for changing piety conditionaly on your god and/or
-- your praying status.
function _M:incPietyFor(god, d, if_praying)
  if self:isFollowing(god) and (not if_praying or self:isPraying()) then
    self:incPiety(d)
  end
end

function _M:getLuck()
  return util.bound((self.luck or 0) + (self.adj_luck or 0), -30, 30)
end

function _M:getLuckScale(min, max)
  return min + math.floor((self:getLuck() + 30)*(max - min)/60)
end

-- Utility function to alter player fields that have both a "real" value
-- and a "displayed" value that may or may not also change if the provided
-- delta is considered "visible".
function _M:adjustDis(field, delta, visible, why)
  local dis_field = 'dis_' .. field
  delta = delta or 0
  self[field] = (self[field] or 0) + delta
  self[dis_field] = (self[dis_field] or 0) + (visible and delta or 0)
  if field == 'to_h' then
    print(('[COMBAT] adjustDis:  %+d to_h (%s) -> %d'):format(delta, (why or '???'), self.to_h))
  end
end

-- This is *not* called by recalcEverything(); it should be called when the
-- character is created, and when a Possessor incarnates or deincarnates
-- into a corpse (TODO).
function _M:recalcAppearance()
  local r = self.descriptor.race:lower()
  local sr = self.descriptor.subrace:lower()
  local cg = self.descriptor.class_group:lower()
  local c = self.descriptor.class:lower()
  for _, try in ipairs({ r..'_'..c, r..'_'..cg, c, cg, r }) do
    if fs.exists(Tiles.prefix..'player/'..try..'.png') then
      self.image = 'player/'..try..'.png'
      print("[PLAYER] using tile '"..try.."' for player")
      break
    end
  end
  if config.settings.tome2.tiles == 'ascii' then
    -- This isn't used in ASCII mode, but the Chat dialog expects it to be
    -- present, so we fake one up.
    self.image = 'player/'..r..'.png'
  end

  if not self.ego_particle or self.ego_particle.args.subrace ~= sr then
    self:removeParticles(self.ego_particle)
    if fs.exists(Tiles.prefix..'player/mod/'..sr..'.png') then
      local ts = config.settings.tome2.tiles
      self.ego_particle = self:addParticles(engine.Particles.new('player_mod', 1, {tileset=ts, subrace=sr, size=16}))
    else
      self.ego_particle = nil
    end
  end

  self.display, self.color_r, self.color_g, self.color_b = '@', 230, 230, 230
  if config.settings.tome2.ascii_player_symbols then
    for _, type in ipairs({'race', 'subrace'}) do
      local d = Birther:getBirthDescriptor(type, self.descriptor[type])
      self.display = d.opt_display or self.display
      if d.opt_color then
	self.color_r, self.color_g, self.color_b = d.opt_color.r, d.opt_color.g, d.opt_color.b
      end
    end
  end

  self.changed = true
end

-- cf. calc_hitpoints() in src/xtra1.c
function _M:recalcHP()
  self.max_life = self.base_hp_by_lev[self.level] +
		  math.floor(self:getStatRelated('con_maxhp')*self.level/2)
  self.max_life = math.max(self.max_life, self.level + 1)

  -- Assorted permanent hp modifications.
  self.max_life = math.max(1, self.max_life + self.hp_mod)

  -- Sorcery skill dings hp.
  local sorc = self:getSkillLevel('SORCERY')
  if sorc > 0 then
    self.max_life = math.max(1, self.max_life - math.floor(self.max_life*sorc/100))
  end

  -- TODO Implement Melkor hp sacrifices.
  self.max_life = math.max(1, self.max_life - 10*(self.melkor_sacrifice or 0))

  -- Heroism/superheroism.
  if self:hasEffect(self.EFF_HEROISM) then self.max_life = self.max_life + 10 end
  if self:hasEffect(self.EFF_SUPERHEROISM) then self.max_life = self.max_life + 30 end

  -- HP-multiplying equipment.
  self.max_life = self.max_life + math.floor((self.max_life * self.hp_mult)/5)
  self.max_life = math.max(1, self.max_life)

  -- TODO Handle monster possession.
  -- TODO Handle Necromancer "undeath" effect.

  self:fireCallbacks('on_player_calc_max_life')
  self.max_life = math.max(1, self.max_life)

  self.life = math.min(self.life, self.max_life)
end

function _M:recalcMana()
  local mana = self:getSkillLevel('MAGIC', 200) +
	       math.floor(self:getStatRelated('int_wis_mana')*self.level/4)
  if mana > 0 then mana = mana + 1 end

  -- Modifiers from race and class.
  mana = math.floor(mana * self.base_mana_mult)

  if self:isFollowing('Eru') then
    local m = math.floor(math.min(self.piety, 35000)/1000)
    mana = math.floor(mana * (1 + m/100))
  end

  if self.gloves_lower_mana then
    local reduce = false
    for _, o in ipairs(self:getInven('GLOVES') or {}) do
      if not (o.convey_flags and o.convey_flags.FREE_ACT) and
	 not (o.pval_flags and o.pval_flags.DEX and o.pval and o.pval > 0) and
	 not (o.flags and o.flags.SPELL_CONTAIN)
      then
	reduce = true
	break
      end
    end
    if reduce then mana = math.floor((mana * 3)/4) end
  end

  -- Mana multiplier from equipment.
  mana = math.floor(mana * (1 + self.mana_mult/5));

  -- Penalize heavy armor.
  local aw = self:armorWeight()
  local sk_c = self:getSkillLevel('COMBAT')
  mana = mana - math.max(0, math.floor(aw - sk_c -  20))

  self.max_mana = math.max(mana, 0)
  self:fireCallbacks('on_player_calc_max_mana')
  self.max_mana = math.max(mana, 0)
  self.mana = math.min(self.mana, self.max_mana)
end

function _M:recalcGods()
  local d = Birther:getBirthDescriptor('god', self.descriptor.god)
  if d.apply then d:apply(self) end
end

local pval_mult = {
  skill_searching = 5,
  skill_perception = 5,
  adj_digging = 20,
  invisibility = 10,
}

function _M:applyFlagBundle(bndl)
  if bndl.lite and (not bndl.fuel_lite or bndl.lite_fuel > 0) then
    self.lite = (self.lite or 0) + bndl.lite
  end
  if bndl.antimagic then
    -- This bundle may or may not be an Object, but we can fake it.
    local am_pow, am_r = Object.antimagicMods(bndl, self)
    self.antimagic = self.antimagic + am_pow
    self.antimagic_rad = self.antimagic_rad + am_r
  end
  for stat, v in pairs(bndl.stat_flags or {}) do
    local val = (type(v) == 'number' and v) or bndl.pval or 0
    self:adjustStatForEquip(stat, val)
  end
  for dt, _ in pairs(bndl.convey_resist_dam or {}) do
    self.resist_dam[dt] = true
  end
  for dt, _ in pairs(bndl.convey_immune_dam or {}) do
    self.immune_dam[dt] = true
  end
  for dt, _ in pairs(bndl.convey_vuln_dam or {}) do
    self.vuln_dam[dt] = true
  end
  for dt, method in pairs(bndl.convey_melee_retal or {}) do
    self.melee_retal[dt] = method
  end
  for flag, _ in pairs(bndl.convey_flags or {}) do
    self.flags[flag] = true
  end
  for flag, _ in pairs(bndl.esp_flags or {}) do
    self.esp[flag] = true
  end
  for field, v in pairs(bndl.pval_flags or {}) do
    local val = (type(v) == 'number' and v) or bndl.pval or 0
    self[field] = self[field] + val * (pval_mult[field] or 1)
  end
  for field, _ in pairs(bndl.cum_flags or {}) do
    self[field] = self[field] + 1
  end
  for _, tid in ipairs(bndl.talents or {}) do
    self:accumTargetTalent(tid)
  end
  for _, ab in ipairs(bndl.abilities or {}) do
    self.abilities[ab] = true
  end
end

function _M:applyLeveledFlagBundles(bndls, vs)
  for lev, bndl in pairs(bndls or {}) do
    if lev <= vs then self:applyFlagBundle(bndl) end
  end
end

function _M:queryLeveledFlagBundles(bndls, vs, group, field)
  for lev, bndl in pairs(bndls or {}) do
    if lev <= vs and bndl[group] and bndl[group][field] then return true end
  end
  return false
end

function _M:recalcEquip()
  local sets = {}
  self:inventoryApplyAll(function(inven, _, o)
    if inven.worn and o.set_info then
      sets[o.set_info.name] = (sets[o.set_info.name] or 0) + 1
    end
  end)

  -- Clear this preemptively
  self.icky_weapon = nil

  -- TODO Add light modifiers from birth descriptors.
  self:inventoryApplyAll(function(inven, _, o)
    if inven.worn then
      self.ac = self.ac + (o.armor or 0)
      self:adjustDis('to_a', o.to_a, o:isIDd())
      -- Slightly Convoluted Hack(TM):  to-hit/to-dam bonuses for
      -- weapon-related stuff will be applied per weapon during combat
      -- handling.  (cf. src/xtra1.c:calc_bonuses() lines 3034-3048).
      local wpn_hack = o.type == 'weapon' or o.type == 'ammo' or o.type == 'tool' or o.type == 'shooter' or o.subtype == 'boomerang'
      if not wpn_hack then
	self:adjustDis('to_h', o.to_h, o:isIDd(), o:getName())
	self:adjustDis('to_d', o.to_d, o:isIDd())
      end
      if self:isFollowing('Eru') and o.edged_weapon and not o:hasFlag('convey_flags', 'BLESSED') then
	-- Eru non-blessed weapon penalty.  Note that unlike T2, we
	-- require the weapon itself to have the BLESSED flag (mostly
	-- because the flag from other weapons might not have been
	-- applied yet...).
	self:adjustDis('to_h', -15, true, 'Eru edged weapon')
	self:adjustDis('to_d', -15, true)
	self.icky_weapon = true
      end
      -- Objects are at least in part shaped like flag bundles as handled
      -- by :applyFlagBundle() above, so we can reuse it.
      self:applyFlagBundle(o)
      -- Apply flags from completed or partly completed artifact sets.
      if o.set_info then
	for n = 1, sets[o.set_info.name] do
	  if o.set_info[n] then self:applyFlagBundle(o.set_info[n]) end
	end
      end
      if o.worn_talents then self:applyObjectTalents(o, true) end
    else
      if o.carried_talents then self:applyObjectTalents(o, false) end
    end
  end)
end

function _M:applyObjectTalents(o, worn)
  local talents = worn and o.worn_talents or o.carried_talents
  if talents then
    -- Elegant Hack(TM):  To ensure that spellbooks assign their talents in
    -- the right order, the worn_talents{}/carried_talents{} subtable lists
    -- its talents in numerical order and maps talent requirements by
    -- talent ID.
    for _, tid in ipairs(talents) do
      local t, req, add = self:getTalentFromId(tid), talents[tid], true
      if req and req.talent then add = add and self:knowTalent(req_talent) end
      if req and req.skill then
	local tag = req.skill.tag
	local sk = self:getSkillLevel(tag)
	local sk_raw = self.skills[tag] and self.skills[tag].value or 0
	add = add and ((req.skill.level and sk >= req.skill.level) or
		       (not req.skill.level and sk_raw > 0))
      end
      if t.is_spell then
	add = add and spell.spell_skill_level_ok(self, tid)
      end
      if t.god then add = add and self:isFollowing(t.god) end
      if add then self:accumTargetTalent(tid) end
    end
  end
end

function _M:applyTemporaryEffects()
  -- Apply or remove effects triggered by flags.
  for id, def in pairs(self.tempeffect_def) do
    if def.equip_flags then
      local eq = def.equip_flags
      local eff = self:hasEffect(id)
      local val = (#eq == 1 and self[eq[1]]) or (#eq == 2 and self[eq[1]] and self[eq[1]][eq[2]])
      local change = (eff and not val and not eq.sticky) or (val and not eff) or (eq.power and val and eff and val ~= eff.power)
      if change then
	if not val or (eq.power and val == 0) then
	  self:removeEffect(id)
	elseif eq.power then
	  self:setEffect(id, 1, {power=val})
	else
	  self:setEffect(id, 1, {})
	end
      end
    end
  end

  -- Slight Hack(TM):  Collect a list of effects to walk through before
  -- walking through them, in case effects cancel each other (in
  -- particular, the temporary effect from the Yavanna 'Tree Roots' spell
  -- cancels stun).  ActorTemporaryEffects:callEffect() will take care of
  -- skipping temporary effects that don't have an apply() function.
  local effs = table.keys(self.tmp)
  for _, eff_id in ipairs(effs) do self:callEffect(eff_id, 'apply') end
end

function _M:applySustainedTalents()
  for t_id, p in pairs(self.sustain_talents) do
    local t = self:getTalentFromId(t_id)
    if t.apply then t.apply(self, t) end
    if type(p) == 'table' and p.apply then p.apply(self, t, p) end
  end
end

function _M:applyFlags()
  -- NOTE:  All food_regen_mod adjustments below are divided by 10 from
  -- their original implementations in src/dungeon.c lines 1709-1763, as
  -- those are based on ten normal-speed turns.
  self.regen_frac = self.food < util.FOOD.starve and 0 or
		    self.food < util.FOOD.fainting and REGEN_FAINT or
		    self.food < util.FOOD.weak and REGEN_WEAK or REGEN_NORMAL
  if self:has('REGENERATE') then
    self.regen_frac = self.regen_frac * 2
    self.food_regen_mod = self.food_regen_mod + 3
  end
  if self.resting or self:has('AUTO_SEARCH') then
    self.regen_frac = self.regen_frac * 2
  end

  self.food_regen_mod = self.food_regen_mod + self.invisibility/20
  if self:has('SLOW_DIGEST') then
    self.food_regen_mod = self.food_regen_mod - 1
  end

  local shooters = self:getInven('LAUNCHER')
  if shooters and #shooters > 0 then
    -- Flags that add extra might depending on wielded launcher.
    if self.flags.XTRA_MIGHT_BOW and shooters[1].subtype == 'bow' then
      self.extra_might = self.extra_might + 1
    end
    if self.flags.XTRA_MIGHT_SLING and shooters[1].subtype == 'sling' then
      self.extra_might = self.extra_might + 1
    end
  end

  if self:has('AC_LEVEL') then
    local ac_bonus = 20 + math.floor((self.level or 0)/5)
    self:adjustDis('to_a', ac_bonus, true)
  end
end

function _M:applyCrossEffects()
  -- Flight conveys levitation.
  if self.flags.FLY then self.flags.FEATHER = true end
  -- Chaos resistance conveys confusion resistance.
  if self.resist_dam[DamageType.CHAOS] then
    self.resist_dam[DamageType.CONFUSION] = true
  end
  if self.tmp_resist_dam[DamageType.CHAOS] then
    self.tmp_resist_dam[DamageType.CONFUSION] = true
  end
  -- Magical breathing conveys water breathing.
  if self.flags.MAGIC_BREATH then self.flags.WATER_BREATH = true end
end

function _M:updateMainShader()
  local cfg = config.settings.tome2
  if game.fbo_shader and cfg.fullscreen_effects then
    local effects = {}

    -- Low HP.
    if not self.shader_old_life or self.life ~= self.shader_old_life then
      self.shader_old_life = self.life
      if self.life < (self.max_life * cfg.hp_warning_pct)/100 then
	game.fbo_shader:setUniform('hp_warning', 1 - (self.life/self.max_life))
      else
	game.fbo_shader:setUniform('hp_warning', 0)
      end
    end

    -- Effects
    for id, def in pairs(self.tempeffect_def) do
      if def.shader then
	local effect = def:shader(self, self:hasEffect(id))
	if effect then effects[effect] = true end
      end
    end

    game.posteffects_use = table.keys(effects)
    table.insert(game.posteffects_use, game.fbo_shader.shad)
  end
end

local to_zero = {
  'ac', 'to_h', 'to_d', 'to_a', 'dis_to_h', 'dis_to_d', 'dis_to_a',
  'speed', 'lite', 'extra_might', 'extra_shots', 'extra_crit',
  'adj_luck', 'spellpower', 'mana_mult', 'hp_mult', 'infravision',
  'adj_digging', 'extra_blows', 'invisibility', 'drain_mana', 'drain_hp',
  'life_regen_mod', 'food_regen_mod', 'antimagic', 'antimagic_rad',
}
local to_nil = {
  'no_life_regen',
}
local to_clear_tables = {
  'resist_dam', 'tmp_resist_dam', 'immune_dam', 'vuln_dam', 'melee_retal',
  'esp',
}
function _M:recalcEverything()
  print('[PLAYER] recalcEverything')
  self.recalcing_everything = true
  self:preRecalcForSkills()
  for _, spec in pairs(self.stats) do spec.equip_adj = 0 end
  for _, f in ipairs(to_zero) do self[f] = 0 end
  for _, f in ipairs(to_nil) do self[f] = nil end
  self.flags = table.clone(self.base_flags)
  self.abilities = table.clone(self.bought_abilities)
  for _, f in ipairs(to_clear_tables) do self[f] = {} end
  --self.on_melee_attack = {}

  self:recalcEquip()
  --self:applyLevelGains()
  self:applyLeveledFlagBundles(self.gain_at_level, self.level)
  self:recalcForSkills()
  self:recalcGods()

  -- Temporary mimic forms can add talents, so we do this here.
  self:applyTemporaryEffects()

  -- Don't forget our birth talents...
  for type, name in pairs(self.descriptor) do
    local desc = Birther:getBirthDescriptor(type, name)
    for tid, _ in pairs(desc.talents or {}) do self:accumTargetTalent(tid) end
  end

  -- Sync our known talents with what we've determined we should have.
  local known_talents = {}
  for tid, _ in pairs(self.talents) do known_talents[tid] = true end
  for tid, _ in pairs(known_talents) do
    if not self.target_talents[tid] then
      local t = self:getTalentFromId(tid)
      if self.sustain_talents[tid] then
	-- If we were sustaining the talent we're about to lose, deactivate
	-- it.  The self.recalcing_everything flag will prevent us from
	-- being called recursively.
	self:useTalent(tid, nil, nil, true, nil, false, true)
      end
      self:unlearnTalent(tid)
      if not game.creating_player then
	game.log(game.flash.BAD, 'You lose talent '..t.name..'.')
      end
    end
  end
  for _, tid in ipairs(self.target_talents) do
    if not self:getTalentFromId(tid) then
      print('[PLAYER] recalcEverything:  skipping unknown talent '..tid)
    elseif not self:knowTalent(tid) then
      self:learnTalent(tid)
      if not game.creating_player then
	local t = self:getTalentFromId(tid)
	game.log(game.flash.GOOD, 'You gain talent '..t.name..'.')
      end
    end
  end

  -- Clear and restart this for any sustains that confer talents (notably
  -- mimicry cloaks).
  self.target_talents = {}

  -- We do this after the above, so that any sustained talents we lost will
  -- have been deactivated by now.
  self:applySustainedTalents()

  -- See if we gained anything from that...
  for _, tid in ipairs(self.target_talents) do
    if not self:getTalentFromId(tid) then
      print('[PLAYER] recalcEverything:  skipping unknown talent '..tid)
    elseif not self:knowTalent(tid) then
      self:learnTalent(tid)
      if not game.creating_player then
	local t = self:getTalentFromId(tid)
	game.log(game.flash.GOOD, 'You gain talent '..t.name..'.')
      end
    end
  end

  -- *Now* we're done with this.
  self.target_talents = nil

  -- That should be everything that could increase our light radius.
  self.lite = math.min(self.lite, 5)

  self:applyFlags()
  self:applyCrossEffects()
  TacticExplorDialog:apply(self)

  local stat_to_h = self:getStatRelated('dex_to_hit') + self:getStatRelated('str_to_hit')
  local why = ('%+d from DEX, %+d from INT'):format(self:getStatRelated('dex_to_hit'), self:getStatRelated('str_to_hit'))
  self:adjustDis('to_h', stat_to_h, true, why)

  if self:has('NO_MAGIC') then
    self.saving_throw = math.min(self.saving_throw, 95)
  end
  self.saving_throw = self.saving_throw <= 10 and 10 or self.saving_throw + 10

  -- Pre-exponentiate this, as that's how we'll need it.
  self.skill_stealth = util.bound(self.skill_stealth, 0, 30)
  self.stealth_exp = util.lshift(1, 30 - self.skill_stealth)

  self.adj_digging = math.max(self.adj_digging, 1)

  do
    -- Hack(TM):  Wielding a COULD2H weapon with a shield penalizes melee
    -- to-hit and to-dam (cf. src/xtra1.c line 3798-3817).
    -- TODO This will be much trickier with multi-armed possessors.
    local ww, ss = self:getInven('MAIN_HAND'), self:getInven('OFF_HAND')
    for i = 1, math.max(#ww, #ss) do
      local w = ww[i]
      if w and w.flags and w.flags.COULD2H and ss[i] then
	local tmp = math.abs(math.floor((w.to_h or 0)/2))
	self.to_h_melee = (self.to_h_melee or 0) - tmp
	tmp = math.abs(math.floor((w.to_d or 0)/2))
	tmp = tmp + (w.combat and math.floor((w.combat[1]*w.combat[2])/2) or 0)
	self.to_d_melee = (self.to_d_melee or 0) - tmp
      end
    end
  end

  self.speed = self.speed + self:speedDeltaForWeight()
  self:computeGlobalSpeed()
  self:recalcHP()
  self:recalcMana()
  self:recalcSanity()

  self:playerFOV()
  util.juggle_can_see(self)
  self:updateMainShader()

  self.changed = true
  self.recalcing_everything = nil
  print('[PLAYER] recalcEverything:  done')
end

function _M:on_quest_grant(quest)
  self:installCallbacks(quest, 'callbacks', quest.id, 'quest')

  local key = util.getFormattedKeyBinds('SHOW_QUESTS', "' or '")
  game.logPlayer(game.player, "#LIGHT_GREEN#Accepted quest '%s'! #LAST#(Press '%s' to see the quest log)", quest.name, key)
  game.bignews:saySimple(60, "#LIGHT_GREEN#Accepted quest '%s'!#LAST#", quest.name)
end

function _M:on_quest_status(quest, status, sub)
  if quest:isEnded() then
    self:uninstallCallbacks(quest, 'callbacks', quest.id)
  end

  local key = util.getFormattedKeyBinds('SHOW_QUESTS', "' or '")
  if sub then
    game.logPlayer(game.player, "#LIGHT_GREEN#Quest '%s' status updated! #LAST#(Press '%s' to see the quest log)", quest.name, key)
    game.bignews:saySimple(60, "#LIGHT_GREEN#Quest '%s' status updated!#LAST#", quest.name)
  elseif status == engine.Quest.COMPLETED then
    game.logPlayer(game.player, "#LIGHT_GREEN#Quest '%s' completed! #LAST#(Press '%s' to see the quest log)", quest.name, key)
    game.bignews:saySimple(60, "#LIGHT_GREEN#Quest '%s' completed!#LAST#", quest.name)
  elseif status == engine.Quest.DONE then
    game.logPlayer(game.player, "#LIGHT_GREEN#Quest '%s' is done! #LAST#(Press '%s' to see the quest log)", quest.name, key)
    game.bignews:saySimple(60, "#LIGHT_GREEN#Quest '%s' is done!#LAST#", quest.name)
  elseif status == engine.Quest.FAILED then
    game.logPlayer(game.player, "#LIGHT_GREEN#Quest '%s' is failed! #LAST#(Press '%s' to see the quest log)", quest.name, key)
    game.bignews:saySimple(60, "#LIGHT_GREEN#Quest '%s' is failed!#LAST#", quest.name)
  end
end

function _M:onHealSanity(value, src)
  if self.sanity < self.max_sanity then
    local feel = value <  5 and 'a little better' or
		 value < 15 and 'better' or
		 value < 35 and 'much better' or
				'very good'
    game.logPlayer(self, 'You feel '..feel..'.')
  end
  return value
end

-- Recalc everything when temporary effects are added or removed.
function _M:on_temporary_effect_added(eff_id, e, p)
  mod.class.Actor.on_temporary_effect_added(self, eff_id, e, p)
  if not self.recalcing_everything then
    self:recalcEverything()
  end
end

function _M:on_temporary_effect_removed(eff_id, e, p)
  mod.class.Actor.on_temporary_effect_removed(self, eff_id, e, p)
  self:runStop("temporary effect expired")
  self:restStop("temporary effect expired")
  if not self.recalcing_everything then
    self:recalcEverything()
  end
end

function _M:resistDam(damtype, p)
  p = p or {}
  -- We assume p.or_immune and p.count won't be used simultaneously.
  if p.or_immune and self:immuneDam(damtype) then return true end

  local dt = DamageType:get(damtype)
  local perm = dt.is_resist and dt:is_resist(self)
  perm = perm or self.resist_dam[damtype]
  local temp = self.tmp_resist_dam[damtype]
  if p.count then return (perm and 1 or 0) + (temp and 1 or 0) end
  return perm or temp
end

function _M:immuneDam(damtype)
  local dt = DamageType:get(damtype)
  if dt.is_immune then
    local imm, msg = dt:is_immune(self)
    if imm then return imm, msg end
  end
  return self.immune_dam[damtype]
end

function _M:vulnDam(damtype)
  local dt = DamageType:get(damtype)
  if dt.is_vulnerable then
    local vuln, msg = dt:is_vulnerable(self)
    if vuln then return vuln, msg end
  end
  return self.vuln_dam[damtype]
end
