-- $Id: StatsDisplay.lua 4222 2021-06-02 02:23:38Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local FontPackage = require 'engine.FontPackage'
local ActorStats = require 'mod.class.interface.ActorStats'
local TacticExplorDialog = require 'mod.dialogs.TacticExplorDialog'

module(..., package.seeall, class.make)

-- Static
function _M:prefFont(fontname, fontsize)
  local font_mono, size_mono = FontPackage:getFont("mono_small", "mono")
  fontname = fontname or font_mono or '/data/font/DroidSansMono.ttf'
  fontsize = fontsize or size_mono or 10
  return core.display.newFont(fontname, fontsize)
end

-- Static
function _M:prefWidth(fontname, fontsize)
  local font = self:prefFont(fontname, fontsize)
  -- 25 charas wide, plus 5 pixel borders
  return font:size('MMMMMMMMMMMMMMMMMMMMMMMMM') + 10
end

function _M:init(actor, x, y, w, h, bgcolor, fontname, fontsize)
  self.actor = actor
  self.bgcolor = bgcolor or { 0, 0, 0 }
  self.font = self:prefFont(fontname, fontsize)
  self.font_h = self.font:lineSkip()

  self.zones = {}
  self.cache = {}
  setmetatable(self.cache, {__mode='v'})

  self:resize(x, y, w, h)
end

function _M:resize(x, y, w, h)
  self.display_x, self.display_y = math.floor(x), math.floor(y)
  self.w, self.h = math.floor(w), math.floor(h)
  self.surface = core.display.newSurface(w, h)
  self.texture, self.texture_w, self.texture_h = self.surface:glTexture()
  if self.actor then
    self.actor.changed = true
  end

  local cw, ch = self.font:size(' ')
  self.font_w = cw
end

function _M:display()
  if not self.actor or not self.actor.changed or self.actor.creating_player then
    return self.surface
  end
  self.surface:erase(self.bgcolor[1], self.bgcolor[2], self.bgcolor[3])

  self.zones = {}
  local aa = self.actor
  local x, y = 5, 5

  for i, stat in ipairs(ActorStats.stats) do
    x, y = self:drawStat(x, y, aa, i, stat)
    --[[
    local c = aa:isStatMaxed(stat) and 'LIGHT_BLUE' or
	      aa:hasEffect(aa['EFF_LOSE_'..stat]) and 'ORANGE' or
	      aa:isStatDrained(stat) and 'YELLOW' or 'LIGHT_GREEN'
    local stat_fmt = aa:getStat(stat, { fmt='default' })
    stat_fmt = '#'..c..'#'..stat..': '..stat_fmt..'#LAST#'
    x, y = self:drawLine(x, y, stat_fmt)
    ]]
  end

  x, y = self:drawLine(x, y)
  x, y = self:drawResource(x, y, 'HP:   ', aa.life, aa.max_life)
  if aa.max_mana then
    x, y = self:drawResource(x, y, 'Mana: ', aa.mana, aa.max_mana)
  end
  if aa.gold then
    local str = ('#GOLD#Gold:   %d#LAST#'):format(aa.gold)
    x, y = self:drawLine(x, y, str)
  end
  if aa.last_piety_regen then
    local c = aa.praying and '#LIGHT_BLUE#' or aa.piety < 0 and '#YELLOW#' or '#LIGHT_GREEN#'
    local str = ('%sPiety: % d#LAST#'):format(c, aa.piety)
    x, y = self:drawLine(x, y, str)
  end

  local spd_desc = aa.speed > 0 and 'Fast' or aa.speed < 0 and 'Slow' or 'Normal'
  local c_spd = aa.speed > 0 and '#00FF00#' or aa.speed < 0 and '#YELLOW#' or '#SLATE#'
  local spd_str = (c_spd..'Speed:  '..spd_desc..' (%+d)'):format(aa.speed)
  x, y = self:drawLine(x, y, spd_str)

  -- TODO Find a better place to put this.
  local c_lev = aa.level < aa.level_hiwater and '#YELLOW#' or '#LIGHT_GREEN#'
  local lev_str = ("Level:  %d"):format(aa.level)
  lev_str = c_lev..lev_str..'#LAST#'
  x, y = self:drawLine(x, y, lev_str)

  local c_exp = aa.exp < aa.max_exp and '#YELLOW#' or '#LIGHT_GREEN#'
  if aa.level >= aa.max_level then
    x, y = self:drawLine(x, y, c_exp..'Exp:    ********#LAST#')
  else
    local mode = config.settings.tome2.exp_display
    local xlev = aa:getExpChart(aa.level)
    local xlevp1 = aa:getExpChart(aa.level+1)
    local str
    if mode == 'Absolute' then
      str = string.format('Exp:    %0.2f', aa.exp)
    elseif mode == 'Percent' then
      local pct = math.floor((aa.exp - xlev)/(xlevp1 - xlev) * 100)
      str = string.format('Exp:    %d%%', pct)
    else
      str = string.format('Exp:    %.2f', xlevp1 - aa.exp)
    end
    str = c_exp..str..'#LAST#'
    x, y = self:drawLine(x, y, str)
  end

  local sym_i = aa:getInven('SYMBIOTE')
  local sym_o = sym_i and sym_i[1]
  if sym_o and sym_o.symbiote and not sym_o.symbiote.dead then
    x, y = self:drawResource(x, y, 'SymHP:', sym_o.symbiote.life, sym_o.symbiote.max_life)
  end

  local c_food = aa.food < util.FOOD.starve and '#RED#Starving' or
		 aa.food < util.FOOD.fainting and '#LIGHT_RED#Fainting' or
		 aa.food < util.FOOD.weak and '#ORANGE#Weak' or
		 aa.food < util.FOOD.hungry and '#YELLOW#Hungry' or
		 aa.food >= util.FOOD.full and '#LIGHT_GREEN#Full' or
		 aa.food >= util.FOOD.max and '#YELLOW#Gorged' or nil
  if config.settings.tome2.food_display then
    local c, s = string.match(c_food or '#LIGHT_GREEN#OK', '^#([%a_]+)#(.+)')
    c_food = ('#%s#Food:   %d (%s)'):format(c, aa.food, s)
  end
  if c_food then
    c_food = c_food .. '#LAST#'
    x, y = self:drawLine(x, y, c_food)
  end

  -- TODO Find a better place to put this.
  if game.zone and game.level and game.level.level then
    y = y + self.font_h
    local str = game.zone.name
    if not game.zone.town and not game.zone.wilderness and not game.zone.worldmap then
      str = ("%s (%d)"):format(str, game.zone:level_adjust_level(game.level))
    end
    str = game.level.data.level_name or str
    str = '#0000FF#'..str..'#LAST#'
    x, y = self:drawLine(x, y, str)
  end

  local tac_exp_desc = TacticExplorDialog:getCurDesc(self.actor)
  tac_exp_desc = '#GREY#'..tac_exp_desc..'#LAST#'
  x, y = self:drawLine(x, y, tac_exp_desc)
  x, y = self:drawLine(x, y)

  for t_id, p in pairs(aa.sustain_talents) do
    local def = aa.talents_def[t_id]
    local desc = def.sust_display and def.sust_display(p) or def.name
    x, y = self:drawStatus(x, y, desc, def.name, true, 'sustain', t_id, p)
  end

  for eff_id, eff in pairs(aa.tmp) do
    local def = aa:getEffectFromId(eff_id)
    local desc = def.display and def.display(eff) or def.desc
    local good = def.status == 'beneficial'
    if not def.no_dur then desc = desc..' ('..eff.dur..')' end
    x, y = self:drawStatus(x, y, desc, def.desc, good, 'effect', eff_id, eff)
  end

  self.surface:updateTexture(self.texture)
  return self.surface
end

function _M:addZone(x, y, txt, title, type, id, src)
  if not self.cache[txt] then
    local w, h = self.font:size(txt)
    self.cache[txt] = { w=w, h=h }
  end
  local c = self.cache[txt]
  local zone = {
    x=x, y=y, w=c.w, h=c.h, type=type, id=id, src=src, text=txt, title=title
  }
  table.insert(self.zones, zone)
end

function _M:drawLine(x, y, text)
  if x > 5 then
    -- Start at new line
    x, y = 5, y + self.font_h
  end
  if text then
    self.surface:drawColorStringBlended(self.font, text, x, y, 255, 255, 255)
  end
  return 5, y + self.font_h
end

function _M:drawResource(x, y, pfx, amt, max_amt)
  local amt_fmt = tostring(amt)
  local amt_fmt2 = ('%.2f'):format(amt)
  amt_fmt = amt_fmt:len() < amt_fmt2:len() and amt_fmt or amt_fmt2

  local red_cutoff = max_amt * config.settings.tome2.hp_warning_pct/100
  local c = amt < red_cutoff and '#RED#' or
	    amt < max_amt and '#YELLOW#' or '#LIGHT_GREEN#'
  local str = c..pfx..'  '..amt_fmt..'/'..max_amt..'#LAST#'
  return self:drawLine(x, y, str)
end

function _M:drawStatus(x, y, text, title, good, type, id, src)
  -- Crop status text to 12 chars to make sure it fits.
  text = text:sub(1, 12)
  local display_text = (good and '#LIGHT_GREEN#' or '#RED#') .. text .. '#LAST#'
  self.surface:drawColorStringBlended(self.font, display_text, x, y, 255, 255, 255)
  self:addZone(x, y, text, title, type, id, src)
  if x == 5 then
    -- Two status blurbs per line.
    local half_w = self.font:size('MMMMMMMMMMMMM')
    return x + half_w, y
  else
    return 5, y + self.font_h
  end
end

function _M:drawStat(x, y, aa, i, stat)
  local c = aa:isStatMaxed(stat) and '#LIGHT_BLUE#' or
	    aa:hasEffect(aa['EFF_LOSE_'..stat]) and '#ORANGE#' or
	    aa:isStatDrained(stat) and '#YELLOW#' or '#LIGHT_GREEN#'
  local stat_fmt = aa:getStat(stat, { fmt='default' })
  stat_fmt = c..stat..': '..stat_fmt..'#LAST#'

  local half = math.floor((#ActorStats.stats + 1)/2)
  local half_w = self.font:size('MMMMMMMMMMMMM')
  local dx = (i > half) and half_w or 0
  local dy = (i - 1 - (i > half and half or 0)) * self.font_h
  self.surface:drawColorStringBlended(self.font, stat_fmt, x+dx, y+dy, 255, 255, 255)

  if i < #ActorStats.stats then return x, y end
  return 5, y + half * self.font_h
end

function _M:toScreen()
  self:display()
  self.texture:toScreenFull(self.display_x, self.display_y, self.w, self.h, self.texture_w, self.texture_h)
end

function _M:onMouse(button, mx, my, click, on_over, on_click)
  local aa = self.actor

  mx, my = mx - self.display_x, my - self.display_y
  for i, z in ipairs(self.zones) do
    if mx >= z.x and mx < z.x+z.w and my >= z.y and my < z.y+z.h then
      if click and on_click then
	if on_click(z.type, z.id, z.src) then click = false end
      end
      if button == 'left' and click then
	-- Probably shouldn't do anything
      elseif button == 'right' and click then
	-- Right-click to cancel beneficial effects.
	if z.type == 'effect' then
	  local def = aa:getEffectFromId(z.id)
	  if def.status == 'beneficial' or config.settings.cheat then
	    local Dialog = require 'engine.ui.Dialog'
	    local function cb(ok)
	      if ok then aa:removeEffect(z.id) end
	    end
	    local desc = def.display and def.display(eff) or def.desc
	    desc = ("Cancel effect '%s'?"):format(desc)
	    Dialog:yesnoPopup('Cancel Effect', desc, cb)
	  end
	end
      else
	aa.changed = true
	local oldsel = self.cur_sel
	self.cur_sel = i
	if on_over and self.cur_sel ~= oldsel then
	  local text = tstring{{'color','GOLD'}, {'font','bold'}, (z.title or z.text), {'font','normal'}, {'color','LAST'}, true}
	  if z.type == 'effect' then
	    local desc = aa:callEffect(z.id, 'describe')
	    if desc then text:merge(desc) end
	    local def = aa:getEffectFromId(z.id)
	    if def.status == 'beneficial' then
	      text:add(true, '---', true, {'font','italic'}, 'You can cancel this effect early by right-clicking on it.', {'font','normal'})
	    end
	  elseif z.type == 'sustain' then
	    local desc = aa:callTalent(z.id, 'sustain_describe', z.src) or
			 aa:getTalentFullDescription(aa:getTalentFromId(z.id))
	    if desc then text:merge(desc) end
	  end
	  on_over(text)
	end
      end
      return
    end
  end
  self.cur_sel = nil
end
