-- $Id: T2DungeonLevel.lua 3894 2020-04-21 02:40:39Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2020 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Map = require 'mod.class.Map'
local T2LevelGenerator = require 'mod.class.generator.map.T2LevelGenerator'
local T2DungeonRooms = require 'mod.class.generator.map.T2DungeonRooms'

module(..., package.seeall, class.inherit(T2LevelGenerator, T2DungeonRooms))

-- We'll turn these off once we're sure everything is working...
_M.debug = {
  tunnels = true,
  roomgen = true,
}

-- L in N chance of generating an unusual room, where L is depth.
_M.unusual_rooms = 194

-- We set up rooms in 11x11 blocks.
_M.block_h = 11
_M.block_w = 11

function _M:init(zone, map, level, data)
  T2LevelGenerator.init(self, zone, map, level, data)

  -- Pre-create some generators for rooms to use.
  self.actor_gen = zone:getGenerator('actor', level, level.data.generator.actor)
  self.obj_gen = zone:getGenerator('object', level, level.data.generator.object)
end

-- Roughly mimics level_generate_dungeon() in src/generate.c
function _M:layoutMap(lev, old_lev)
  self.cur_lev = lev
  self.dun_lev = self.zone:level_adjust_level(self.level, 'map')
  self.rooms = {}
  self.doors = {}
  self.has_crowded_room = false
  self.avoid = {}
  local cavern, destroyed = false, false

  -- TODO dungeon town level (cf. src/generate.c line 6768)

  -- Empty levels.
  if self.data.flags.EMPTY or (config.settings.tome2.allow_empty_levels and rng.chance(15)) then
    -- Pass this back to the caller so it will know to light the level.
    self.empty_level = true
    self:report_precog('Arena level')
    -- Refill the level with floor tiles.
    self:fill_level(empty_level, self.data.fill_step, lev)
  end

  -- Cavern in the middle of the level. (cf. src/generate.c line 6792)
  if self.data.flags.CAVERN and rng.range(1, math.floor(lev/2)) > 30 then
    cavern = true
    self:report_precog('Cavern on level')
    self:build_cavern()
  end

  -- TODO destroyed level (except for quest levels and small levels)
  --      (cf. src/generate.c line 6806)
  --[[
  destroyed = self.dun_lev > 10 and rng.chance(18) 
  ]]

  self.room_grid = {
    w = math.floor(self.map.w / _M.block_w),
    h = math.floor(self.map.h / _M.block_h)
  }
  for y = 0, self.room_grid.h - 1 do self.room_grid[y] = {} end

  -- Equivalent of hook HOOK_BUILD_ROOM1, used by princess quests and
  -- Thrain in Dol Guldur (cf. src/generate.c line 6855).
  local hook = {
    'MapGenerator:T2DungeonLevel:firstRoom',
    gen = self,
    gy = rng.range(0, self.room_grid.h - 1),
    gx = rng.range(0, self.room_grid.w - 1),
    lev = lev,
    old_lev = old_lev,
    spots = self.spots,
    done = false,
    no_down = false
  }
  self:triggerHook(hook)
  if hook.no_down then self.data.no_down_stairs = true end

  for _ = 1, util.getval(self.data.nb_rooms or 50) - (hook.done and 1 or 0) do
    local gy = rng.range(0, self.room_grid.h - 1)
    local gx = rng.range(0, self.room_grid.w - 1)
    -- Arrange x%3==1 if align_rooms is set.
    if config.settings.tome2.align_rooms then gx = gx + 1 - (gx % 3) end
    if self.debug.roomgen then print(('[MAPGEN] Trying to place room at %d,%d'):format(gx,gy)) end

    -- We have to do this in a somewhat convoluted manner; the T2 C code
    -- makes extensive use of 'continue' in this loop, for which Lua has no
    -- counterpart.
    local done_room = false

    -- Destroyed levels (cf. src/generate.c line 6875)
    if destroyed then
      -- "Destroyed levels are boring", according to the T2 code.
      local k = rng.range(1, 100)
      if not cavern and k < self.dun_lev then
	done_room = self:createRoom('crypt', gx, gy)
      elseif self.data.flags.CIRCULAR_ROOMS then
	done_room = self:createRoom('circular', gx, gy)
      else
	done_room = self:createRoom('basic', gx, gy)
      end
      -- The T2 code assumes a room was created here.
      done_room = true
    end

    -- Vaults and other fancy rooms (except on town levels)
    -- (cf. src/generate.c line 6903)
    -- TODO Not on dungeon town levels.
    local wgts1, wgts2 = 0, 0
    for _, def in ipairs(self.room_types) do
      wgts1 = wgts1 + (def.wgt_unusual or 0)
      wgts2 = wgts2 + (def.wgt_very_unusual or 0)
    end
    if self.debug.roomgen then print(('[MAPGEN]   room weights %d, %d'):format(wgts1, wgts2)) end

    local function pick_room(total_wgt, field, unusual)
      local k = rng.range(1, total_wgt)
      local dbg = {}
      for _, def in ipairs(self.room_types) do
	if def[field] then table.insert(dbg, def.id .. '=' .. def[field]) end
      end
      dbg = table.concat(dbg, '/')
      if self.debug.roomgen then print(('[MAPGEN]   pick_room(%s, %s) %d/%d from (%s)'):format(field, tostring(unusual), k, total_wgt, dbg)) end
      for _, def in ipairs(self.room_types) do
	if done_room then return end
	if def[field] then
	  if k <= def[field] and self:createRoom(def.id, gx, gy, unusual) then
	    done_room = true
	    return
	  else
	    k = k - def[field]
	  end
	end
      end
    end

    local ironman = game.state.birth.ironman_rooms
    if ironman then
      -- Always try for a greater vault first if ironman rooms is set.
      if self.debug.roomgen then print('[MAPGEN]   attempting greater vault for ironman rooms') end
      done_room = self:createRoom('greater_vault', gx, gy)
      if not done_room then
	if self.debug.roomgen then print('[MAPGEN]     failed generating greater vault; falling through...') end
      end
    end
    -- Failing that, try for an unusual room.
    if not done_room and not self.zone.town and rng.range(1, self.unusual_rooms) < self.dun_lev then
      if self.debug.roomgen then print('[MAPGEN]   maybe unusual room?') end
      -- Double-check for very unusual room
      -- (cf. src/generate.c lines 6910-6951)
      -- [Slight Hack(TM):  In the T2 code, even if we pass the random
      -- check, it only works out to a 60% chance of creating a very
      -- unusual room.]
      if (ironman or rng.range(1, self.unusual_rooms) < self.dun_lev) and rng.percent(60) then
	if self.debug.roomgen then print('[MAPGEN]   ooh, maybe very unusual room?') end
	pick_room(wgts2, 'wgt_very_unusual', 'very')
      end
      -- Failing that, back to "just" unusual rooms.
      if not done_room then
	if self.debug.roomgen then print('[MAPGEN]   trying unusual room') end
	pick_room(wgts1, 'wgt_unusual', true)
      end
    end
    if not done_room then
      -- "Attempt a trivial room"
      if self.data.flags.CAVE then
	if self.debug.roomgen then print('[MAPGEN]   trying cave room') end
	done_room = self:createRoom('fractal_cave', gx, gy)
      elseif self.data.flags.CIRCULAR_ROOMS then
	if self.debug.roomgen then print('[MAPGEN]   trying circular room') end
	done_room = self:createRoom('fractal_cave', gx, gy)
	done_room = self:createRoom('circular', gx, gy)
      else 
	if self.debug.roomgen then print('[MAPGEN]   trying basic room') end
	done_room = self:createRoom('basic', gx, gy)
      end
    end
    if not done_room then
      if self.debug.roomgen then print('[MAPGEN]   hmm, we did not make a room in this loop...') end
    end
  end

  -- Force at least one room.
  while #self.rooms == 0 do self:createRoom('basic', 0, 0) end

  -- Connect rooms.
  table.shuffle(self.rooms)
  for i = 1, #self.rooms do
    local prev = i == 1 and #self.rooms or i - 1
    self:connectRooms(prev, i)
  end

  -- Place doors in intersection spaces we found while tunneling.
  if not self.data.flags.NO_DOORS then
    for _, d in ipairs(self.doors) do
      self:try_doors(d.x, d.y)
    end
  end

  -- TODO magma streamers, quartz streamers, sand streamers
  --      (cf. src/generate.c line 7050)
  if destroyed then
    -- TODO destroy level (cf. src/generate.c line 7075)
  end
  -- TODO create dungeon town (cf. src/generate.c line 7078)
  -- TODO water and lava rivers (cf. src/generate.c line 7084)
  -- TODO tree/water/lava streamers (cf. src/generate.c line 7128)

  -- Don't need these anymore.
  self.avoid = nil
  self.room_grid = nil
  self.rooms = nil
  self.doors = nil

  self:set_level_boundaries(false)

  -- TODO get player starting position (cf. src/generate.c line 7232)
  -- (maybe not; this may get handled by stair placement in parent generate())
end

-- A utility method to bifurcate various map-generation-related messages to
-- both the logfile output and, if the player has precognition, the game
-- log.
function _M:report_precog(msg, ...)
  print('[MAPGEN] precognition:  ' .. msg:format(...))
  if game.player and game.player:has('PRECOGNITION') then
    game.log(('[precognition] '..msg):format(...))
  end
end

-- Roughly mimics room_alloc() in src/generate.c
-- width and height are in map coordinates; gx and gy are in room_grid
-- coordinates.
-- Returns the center of the allocated room in map coordinates.
function _M:roomAlloc(width, height, crowded, gx, gy)
  local gw = math.floor((width - 1)/_M.block_w) + 1
  local gh = math.floor((height - 1)/_M.block_h) + 1
  if gw > self.room_grid.w or gh > self.room_grid.h then
    if gw > self.room_grid.w then
      if self.debug.roomgen then print(('[MAPGEN] %dx%d room too wide (%d room grids vs. %d)'):format(width, height, gw, self.room_grid.w)) end
    end
    if gw > self.room_grid.w then
      if self.debug.roomgen then print(('[MAPGEN] %dx%d room too tall (%d room grids vs. %d)'):format(width, height, gh, self.room_grid.h)) end
    end
    return false
  end

  -- Shift the room left and up if it's off-map.
  gx = math.min(gx, self.room_grid.w - gw)
  gy = math.min(gy, self.room_grid.h - gh)

  -- Check that the space is empty
  for ggy = gy, gy + gh - 1 do
    for ggx = gx, gx + gw - 1 do
      if self.room_grid[ggy][ggx] then
	if self.debug.roomgen then print(('[MAPGEN] %dgx%dg room @g%d,g%d overlaps room @g%d,g%d'):format(gw, gh, gx, gy, ggx, ggy)) end
	return false
      end
    end
  end

  local ctr = {
    x = gx*_M.block_w + math.floor(gw*_M.block_w/2),
    y = gy*_M.block_h + math.floor(gh*_M.block_h/2),
  }
  self.rooms[#self.rooms+1] = {
    w = width, h = height, cx = ctr.x, cy = ctr.y
  }

  -- Reserve the space
  for ggy = gy, gy + gh - 1 do
    for ggx = gx, gx + gw - 1 do
      self.room_grid[ggy][ggx] = true
    end
  end
  self.has_crowded_room = crowded or self.has_crowded_room

  -- TODO check room boundary to see if room will cut off cavern
  -- [cf. src/generate.c, room_alloc(), line 2033]

  return ctr
end

-- Roughly mimics room_build() in src/generate.c
function _M:createRoom(rtype, gx, gy, unusual)
  local rdef = self.room_types[rtype]
  if not rdef then
    if self.debug.roomgen then print(("[MAPGEN] ??? no room type '%s'"):format(rtype)) end
    return false
  end

  -- Check min depth for room.
  if self.dun_lev < rdef.min_depth and not self.data.flags.IRONMAN_ROOMS then
    if self.debug.roomgen then print(("[MAPGEN] depth %d too shallow for room type '%s' with min depth %d"):format(self.dun_lev, rtype, rdef.min_depth)) end
    return false
  end

  -- Don't create a 'crowded' room if we've already got one on the level.
  if self.has_crowded_room and rdef.crowded then
    if self.debug.roomgen then print(("[MAPGEN] room type '%s' is crowded, and we already have one of those"):format(rtype)) end
    return false
  end

  if self.debug.roomgen then print(("[MAPGEN]   createRoom('%s', %d, %d, %s)"):format(rtype, gx, gy, tostring(unusual))) end
  return rdef:build(self, gx, gy, unusual)
end

-- Roughly mimics correct_dir() in src/generate.c
local function tunnel_dir(x1, y1, x2, y2)
  local dx = x1 == x2 and 0 or x1 < x2 and 1 or -1
  local dy = y1 == y2 and 0 or y1 < y2 and 1 or -1
  if dx ~= 0 and dy ~= 0 then
    -- If diagonal, choose between vertical or horizontal randomly.
    if rng.chance(2) then dx = 0 else dy = 0 end
  end
  return dx, dy
end

-- Roughly mimics rand_dir() in src/generate.c
local function rand_hv_dir()
  local dx, dy = 0, rng.chance(2) and 1 or -1
  if rng.chance(2) then dx, dy = dy, dx end
  return dx, dy
end

-- Roughly mimics build_tunnel() in src/generate.c
function _M:connectRooms(room1, room2)
  local x1, y1 = self.rooms[room1].cx, self.rooms[room1].cy
  local x2, y2 = self.rooms[room2].cx, self.rooms[room2].cy
  local xs, ys = x1, y1
  local tunnels, walls = {}, {}
  local max_loop = 0

  local dx, dy = tunnel_dir(x1, y1, x2, y2)
  local tx, ty
  local no_door = false

  if self.debug.tunnels then print(('[MAPGEN] starting tunnel %d,%d -> %d,%d, dir %d,%d'):format(x1, y1, x2, y2, dx, dy)) end
  while x1 ~= x2 or y1 ~= y2 do
    if max_loop > 2000 then
      if self.debug.tunnels then print('[MAPGEN] tunnel loop aborted at 2000 iters') end
      break
    end
    max_loop = max_loop + 1

    -- Chance of tunnel turning
    if rng.percent(30) then
      dx, dy = tunnel_dir(x1, y1, x2, y2)
      if self.debug.tunnels then print(('[MAPGEN]   turning to dir %d,%d at %d,%d'):format(dx, dy, x1, y1)) end
      if rng.percent(10) then
	-- Or maybe completely random.
	dx, dy = rand_hv_dir()
	if self.debug.tunnels then print(('[MAPGEN]     turning again to dir %d,%d'):format(dx, dy)) end
      end
    end

    tx, ty = x1+dx, y1+dy
    while not self:in_bounds(tx, ty) do
      -- Don't wander off the edge...
      dx, dy = tunnel_dir(x1, y1, x2, y2)
      if self.debug.tunnels then print(('[MAPGEN]     %d,%d is off map; changing direction to %d,%d'):format(tx, ty, dx, dy)) end
      if rng.percent(10) then
	-- Or maybe completely random.
	dx, dy = rand_hv_dir()
	if self.debug.tunnels then print(('[MAPGEN]       turning again to dir %d,%d'):format(dx, dy, x1, y1)) end
      end
      tx, ty = x1+dx, y1+dy
    end

    local g = self.map(tx, ty, Map.TERRAIN)
    local skip = g.define_as == 'PERMA_WALL' or		-- Dungeon edge
		 self.map.attrs(tx, ty, 'vault_outer') or	-- Vault wall
		 self.avoid[tx + ty*self.map.w]		-- Marked to avoid
							-- (see below)
    if not skip then
      if self:is_room_wall(tx, ty) then
	-- Hit a room wall.
	local nx, ny = tx+dx, ty+dy
	local gn = self.map(nx, ny, Map.TERRAIN)
	skip = gn.define_as == 'PERMA_WALL' or
	       self.map.attrs(nx, ny, 'vault_outer') or
	       self:is_room_wall(nx, ny) or
	       self.avoid[nx + ny*self.map.w]
	if not skip then
	  -- Good location; save it, and mark adjacent walls to avoid.
	  if self.debug.tunnels then print(('[MAPGEN]   tunneling through room wall %d,%d -> %d,%d'):format(x1, y1, tx, ty)) end
	  x1, y1 = tx, ty
	  walls[#walls+1] = { x=tx, y=ty }
	  for ax = tx-1, tx+1 do
	    for ay = ty-1, ty+1 do
	      if self:is_room_wall(ax, ay) then
		if self.debug.tunnels then print(('[MAPGEN]     marking room wall %d,%d to avoid'):format(ax, ay)) end
		self.avoid[ax + ay*self.map.w] = true
	      end
	    end
	  end
	else
	  if self.debug.tunnels then
	    local reason =
	      (gn.define_as == 'PERMA_WALL' and 'perma-wall') or
	      (self.map.attrs(nx, ny, 'vault_outer') and 'vault wall') or
	      (self:is_room_wall(nx, ny) and 'room wall') or
	      (self.avoid[tx + ty*self.map.w] and 'marked to avoid') or '???'
	    print(('[MAPGEN]   skipping bad room wall (%s) at %d,%d'):format(reason, tx, ty))
	  end
	end
      elseif self.map.attrs(tx, ty, 'room') then
	-- In a room; pass through quickly
	if self.debug.tunnels then print(('[MAPGEN]   advancing through room interior %d,%d -> %d,%d'):format(x1, y1, tx, ty)) end
	x1, y1 = tx, ty
      elseif self.is_fill[g.define_as] then
	-- Fill wall; tunnel through it
	if self.debug.tunnels then print(('[MAPGEN]   tunneling through fill wall %d,%d -> %d,%d'):format(x1, y1, tx, ty)) end
	x1, y1 = tx, ty
	tunnels[#tunnels+1] = { x=tx, y=ty }
	-- Allow door in next grid
	no_door = false
      else
	-- Corridor intersection or overlap
	if self.debug.tunnels then print(('[MAPGEN]   intersecting corridor %d,%d -> %d,%d'):format(x1, y1, tx, ty)) end
	x1, y1 = tx, ty
	-- Remember that a door is allowed here, and forbid a door in the
	-- next grid.
	if not no_door then
	  self.doors[#self.doors+1] = { x=tx, y=ty }
	  no_door = true
	end

	-- Hack:  Allow pre-emptive tunnel termination if it's gone far
	-- enough.
	if rng.range(0, 99) >= 15 then
	  local dxs = x1 < xs and xs - x1 or x1 - xs
	  local dys = y1 < ys and ys - y1 or y1 - ys
	  if dxs > 10 or dys > 10 then
	    if self.debug.tunnels then print(('[MAPGEN]   early corridor termination at %d,%d'):format(x1, y1)) end
	    break
	  end
	end
      end
    else
      if self.debug.tunnels then
	local reason =
	  (g.define_as == 'PERMA_WALL' and 'perma-wall') or
	  (self.map.attrs(tx, ty, 'vault_outer') and 'vault wall') or
	  (self.avoid[tx + ty*self.map.w] and 'grid marked to avoid') or
	  'mysterious bad grid'
	print(('[MAPGEN]   skipping %s at %d,%d'):format(reason, tx, ty))
      end
    end
  end
  if self.debug.tunnels and x1 == x2 and y1 == y2 then print(('[MAPGEN] finished tunnel at %d,%d'):format(x2, y2)) end

  -- Actually make the tunnel.
  self:dumpMap(('tunnel %d,%d -> %d,%d'):format(xs, ys, x2, y2), {
    ['*'] = tunnels,
    ['X'] = { { x=xs, y=ys } },
    ['Y'] = { { x=x2, y=y2 } },
  })
  for _, t in ipairs(tunnels) do
    -- TODO Allow placing water here, if 'water' parameter is provided
    --      (cf. src.generate.c line 6375)
    self.map(t.x, t.y, Map.TERRAIN, self:resolve('floor'))
  end

  -- Punch through the room walls, possibly placing doors.
  self:dumpMap('doors for preceding tunnel', { ['@'] = walls })
  for _, w in ipairs(walls) do
    if not self.data.flags.NO_DOORS and rng.percent(90) then
      self:place_random_door(w.x, w.y)
    else
      self.map(w.x, w.y, Map.TERRAIN, self:resolve('floor'))
    end
  end
end

function _M:is_room_wall(x, y)
  local g = self.map(x, y, Map.TERRAIN)
  return g.define_as == self.data.outer_wall and self.map.attrs(x, y, 'room')
end

-- Roughly mimics place_random_door() in src/generate.c
function _M:place_random_door(x, y)
  local tmp = rng.range(1, 1000)
  if tmp <= 300 then		-- 30% chance open door
    self.map(x, y, Map.TERRAIN, self:resolve('OPEN_DOOR', nil, true))
  elseif tmp <= 400 then	-- 10% chance broken door
    self.map(x, y, Map.TERRAIN, self:resolve('BROKEN_DOOR', nil, true))
  elseif tmp <= 600 then	-- 20% chance secret door
    -- Slight Hack(TM):  Pass ourself to spell.wall_grid() so it doesn't
    -- have to create another map generator.
    spell.map_generator = self
    self:place_secret_door(x, y)
    spell.map_generator = nil
  elseif tmp <= 900 then	-- 30% chance closed door
    self.map(x, y, Map.TERRAIN, self:resolve('DOOR', nil, true))
  elseif tmp <= 999 then	-- 9.9% chance locked door
    -- We count on resolvers in LOCKED_DOOR to set door_lock_power.
    self.map(x, y, Map.TERRAIN, self:resolve('LOCKED_DOOR', nil, true))
  else				-- 0.1% chance jammed door
    -- We count on resolvers in JAMMED_DOOR to set door_jam_power.
    self.map(x, y, Map.TERRAIN, self:resolve('JAMMED_DOOR', nil, true))
  end
end

-- Roughly mimics place_secret_door() in src/generate.c
function _M:place_secret_door(x, y)
  local g = self.map(x, y)
  if g and g:hasFlags('WALL') then
    -- Dodgy Hack(TM), part 1:  If we're replacing a wall grid, pass it
    -- through to spell.wall_grid() via spell.use_wall_grid to be used as
    -- the secret door's replace_display.
    spell.use_wall_grid = g
  else
    -- Dodgy Hack(TM), part 2:  Otherwise, pass ourself to spell.wall_grid()
    -- via spell.map_generator so that it can use us to set up a suitable
    -- wall replace_display for the secret door.
    spell.map_generator = self
  end
  local gg = self:resolve('SECRET_DOOR', nil, true)
  self.map(x, y, Map.TERRAIN, gg)
  gg:check("addedToLevel", self.level, x, y)
  gg:check("on_added", self.level, x, y)
  -- Dodgy Hack(TM), part 3:  Clean these up behind ourself.
  spell.use_wall_grid = nil
  spell.map_generator = nil
end

-- Roughly mimics try_doors() in src/generate.c
function _M:try_doors(x, y)
  if self.data.flags.NO_DOORS then return end

  -- Find all the cardinally adjacent spots where we can place doors.
  local dd = {}
  for _, cc in ipairs(self.cc) do
    if self:in_bounds(x+cc.x, y+cc.y) and
       not self.map:checkEntity(x+cc.x, y+cc.y, Map.TERRAIN, 'wall') and
       not self.map.attrs(x+cc.x, y+cc.y, 'room') and
       self:allow_door(x+cc.x, y+cc.y)
    then
      dd[#dd+1] = { x=x+cc.x, y=y+cc.y }
    end
  end

  if not rng.chance(4) then	-- 75% chance place doors on all ok spots.
    for _, d in ipairs(dd) do
      if rng.percent(90) then
	self:place_random_door(d.x, d.y)
      end
    end
  else				-- 25% chance something else:
    if #dd == 4 then		--   At a crossroad, place two secret doors
      while #dd > 2 do table.remove(dd, rng.range(1, #dd)) end
    elseif #dd > 1 then		--   At a T-junction or regular spot, place one secret door
      while #dd > 1 do table.remove(dd, rng.range(1, #dd)) end
    end
    for _, d in ipairs(dd) do
      self:place_secret_door(d.x, d.y)
    end
  end
end

-- Roughly mimics possible_doorway() in src/generate.c
function _M:allow_door(x, y)
  if self:adjacent_corridors(x, y) < 2 then return false end
  -- Vertical walls.
  if self.map:checkEntity(x, y-1, Map.TERRAIN, 'wall') and
     self.map:checkEntity(x, y+1, Map.TERRAIN, 'wall')
  then
    return true
  end
  -- Horizontal walls
  if self.map:checkEntity(x-1, y, Map.TERRAIN, 'wall') and
     self.map:checkEntity(x+1, y, Map.TERRAIN, 'wall')
  then
    return true
  end
  -- Nope.
  return false
end

-- Build a rectangle of the specified wall type.
-- Roughly duplicates build_rectangle() in src/generate.c
function _M:build_rectangle(x1, y1, x2, y2, wall, lit, attrs)
  attrs = attrs or {}
  local lev = self.cur_lev
  -- Top and bottom walls
  for x = x1, x2 do
    self.map(x, y1, Map.TERRAIN, self:resolve(wall, nil, false, lev))
    if lit then self.map.lites(x, y1, true) end
    for _, f in ipairs(attrs) do self.map.attrs(x, y1, f, true) end
    self.map(x, y2, Map.TERRAIN, self:resolve(wall, nil, false, lev))
    if lit then self.map.lites(x, y2, true) end
    for _, f in ipairs(attrs) do self.map.attrs(x, y2, f, true) end
  end
  -- Left and right walls
  for y = y1, y2 do
    self.map(x1, y, Map.TERRAIN, self:resolve(wall, nil, false, lev))
    if lit then self.map.lites(x1, y, true) end
    for _, f in ipairs(attrs) do self.map.attrs(x1, y, f, true) end
    self.map(x2, y, Map.TERRAIN, self:resolve(wall, nil, false, lev))
    if lit then self.map.lites(x2, y, true) end
    for _, f in ipairs(attrs) do self.map.attrs(x2, y, f, true) end
  end
end

-- Place a secret door in one of four locations relative to the provided
-- x/y coordinates.  Typically used in conjunction with :build_rectangle()
-- above to place a secret door in the middle of one of the four walls of
-- the rectangle.
function _M:place_rect_secret_door(x1, y1, x2, y2, cx, cy)
  local k = rng.range(1, 4)
  local sx = k == 3 and x1 or k == 4 and x2 or cx
  local sy = k == 1 and y1 or k == 2 and y2 or cy
  self:place_secret_door(sx, sy)
end

-- Fill a rectangle with floor.
-- Roughly duplicates code from build_type1() et al. in src/generate.c
function _M:build_floor(x1, y1, x2, y2, floor, lit, attrs)
  attrs = attrs or {}
  for y = y1, y2 do
    for x = x1, x2 do
      self.map(x, y, Map.TERRAIN, self:resolve(floor, nil, false, self.cur_lev))
      if lit then self.map.lites(x, y, true) end
      for _, f in ipairs(attrs) do self.map.attrs(x, y, f, true) end
    end
  end
end

-- Fill a rectangle with walls.
-- Roughly duplicates code from build_type1() et al. in src/generate.c
function _M:fill_wall(x1, y1, x2, y2, wall, lit, attrs)
  attrs = attrs or {}
  for y = y1, y2 do
    for x = x1, x2 do
      self.map(x, y, Map.TERRAIN, self:resolve(wall, nil, false, self.cur_lev))
      if lit then self.map.lites(x, y, true) end
      for _, f in ipairs(attrs) do self.map.attrs(x, y, f, true) end
    end
  end
end

-- Attempt to place the specified number of slightly higher-leveled
-- monsters in a radius-1 area around the specified x/y coordinates.
-- Roughly duplicates vault_monsters() in src/generate.c.
function _M:place_vault_mon(cx, cy, num)
  -- Precompute candidate grids.
  local grids = core.fov.circle_grids(cx, cy, 1, true)
  local gg = {}
  for x, yy in pairs(grids) do
    for y, _ in pairs(yy) do
      gg[#gg+1] = { x=x, y=y }
    end
  end

  for k = 1, num do
    -- Try nine times for each monster.
    for i = 1, 9 do
      local xy = rng.table(gg)
      if self.map:checkEntity(xy.x, xy.y, Map.TERRAIN, 'empty_bold') then
	self.actor_gen:generateOne{ x=xy.x, y=xy.y, add_levels=2 }
	break
      end
    end
  end
end

-- Attempt to place the specified number of traps within the specified
-- range of of the provided x/y center coordinates.
-- Roughly duplicates vault_traps() and vault_traps_aux() in src/generate.c
-- (apart from actually placing traps, of course...)
function _M:place_vault_traps(cx, cy, dx, dy, num)
  -- Precompute candidate grids.
  local gg = {}
  for x = cx - dx, cx + dx do
    for y = cy - dy, cy + dy do
      if self.map:isBound(x, y) then gg[#gg+1] = { x=x, y=y } end
    end
  end

  for k = 1, num do
    -- Try five times for each trap.
    for i = 1, 5 do
      local xy = rng.table(gg)
      if self.map:checkEntity(xy.x, xy.y, Map.TERRAIN, 'naked_bold') then
	-- TODO Actually place a trap here.
	break
      end
    end
  end
end

-- Place an object at the specified location.
-- Roughly duplicates place_object() in src/object2.c
function _M:place_object(x, y, t)
  t = t and table.clone(t, true) or {}
  t.x, t.y, t.filter = x, y, {add_levels=0}
  if not self.map:isBound(x, y) then return end
  if not self.map:checkEntity(x, y, Map.TERRAIN, 'clean_bold') then return end

  self.obj_gen:generateOne(t)
end

-- Attempt to place the specified number of objects in a pre-set range
-- around the provided center x/y coordinates.
-- Roughly duplicates vault_objects() in src/generate.c
function _M:place_vault_obj(cx, cy, num)
  -- Precompute candidate grids.
  local gg = {}
  for x = cx - 3, cx + 3 do
    for y = cy - 2, cy + 2 do
      if self.map:isBound(x, y) then gg[#gg+1] = { x=x, y=y } end
    end
  end

  for k = 1, num do
    -- Try eleven times for an empty place.
    for i = 1, 11 do
      local xy = rng.table(gg)
      if self.map:checkEntity(xy.x, xy.y, Map.TERRAIN, 'clean_bold') then
	-- 75% object, 25% gold.
	if rng.percent(75) then
	  self:place_object(xy.x, xy.y)
	else
	  self.obj_gen:generateGold{ x=xy.x, y=xy.y }
	end
	break
      end
    end
  end
end

-- Roughly mimics add_door() in src/generate.c
function _M:try_vault_door(x, y)
  local t  = self.map:checkEntity(x, y, Map.TERRAIN, 'define_as')
  local tl = self.map:checkEntity(x-1, y, Map.TERRAIN, 'define_as')
  local tr = self.map:checkEntity(x+1, y, Map.TERRAIN, 'define_as')
  local tu = self.map:checkEntity(x, y-1, Map.TERRAIN, 'define_as')
  local td = self.map:checkEntity(x, y+1, Map.TERRAIN, 'define_as')
  -- We want a wall/floor configuration like:
  --    ?#?      ?.?
  --    .#.  or  ###  [we don't care what ? is]
  --    ?#?      ?.?
  if t ~= self.data.outer_wall then return end
  if tl == self.data.outer_wall and tr == self.data.outer_wall and self:get_is_floor(x, y-1) and self:get_is_floor(x, y+1) then
    self:place_secret_door(x, y)
    -- The T2 code describes this as "set boundarys so don't get wide
    -- doors"; I think it means that by changing the adjacent outer_wall
    -- grids to fill, we prevent this method from placing another door on
    -- them.
    gen.map(x-1, y, Map.TERRAIN, gen:resolve('fill', nil, false, gen.dun_lev))
    gen.map(x+1, y, Map.TERRAIN, gen:resolve('fill', nil, false, gen.dun_lev))
  elseif tu == self.data.outer_wall and td == self.data.outer_wall and self:get_is_floor(x-1, y) and self:get_is_floor(x+1, y) then
    self:place_secret_door(x, y)
    -- As above.
    gen.map(x, y-1, Map.TERRAIN, gen:resolve('fill', nil, false, gen.dun_lev))
    gen.map(x, y+1, Map.TERRAIN, gen:resolve('fill', nil, false, gen.dun_lev))
  end
end

-- Roughly mimics fill_treasure() in src/generate.c
function _M:fill_treasure(x1, y1, x2, y2, power)
  local cx, cy = math.floor((x1+x2)/2), math.floor((y1+y2)/2)
  local size = math.abs(x2-x1) + math.abs(y2-y1)

  for x = x1, x2 do
    for y = y1, y2 do
      -- What we put here is based on distance from vault center and on
      -- power (1=easiest, 10=hardest).
      local val = (core.fov.distance(cx, cy, x, y) * 100)/size + rng.range(1, 10) - power
      -- Sometimes we put nothing here.
      if rng.range(1, 100) - power*3 > 50 then val = 20 end
      -- We place stuff on floor, shallow water or shallow lava.
      local t = self.map:checkEntity(x, y, Map.TERRAIN, 'define_as')
      if self:get_is_floor(x, y) or t == 'SHALLOW_WATER' or t == 'SHALLOW_LAVA' then
	if val < 0 then
	  -- Very tough monster + very good treasure
	  self.actor_gen:generateOne{ x=x, y=y, filter={add_levels=40} }
	  self.obj_gen:generateOne{ x=x, y=y, filter={add_levels=20} }
	elseif val < 5 then
	  -- Tough monster + good treasure
	  self.actor_gen:generateOne{ x=x, y=y, filter={add_levels=20} }
	  self.obj_gen:generateOne{ x=x, y=y, filter={add_levels=10} }
	elseif val < 10 then
	  -- Monster
	  self.actor_gen:generateOne{ x=x, y=y, filter={add_levels=9} }
	elseif val < 17 then
	  -- Empty space
	elseif val < 23 then
	  -- Object or trap
	  if rng.percent(25) then
	    self.obj_gen:generateOne{ x=x, y=y }
	  else
	    -- TODO place trap
	  end
	elseif val < 30 then
	  -- Monster and trap
	  self.actor_gen:generateOne{ x=x, y=y, filter={add_levels=5} }
	  -- TODO place trap
	elseif val < 40 then
	  -- Monster or object
	  if rng.percent(50) then
	    self.actor_gen:generateOne{ x=x, y=y, filter={add_levels=3} }
	  else
	    self.obj_gen:generateOne{ x=x, y=y, filter={add_levels=7} }
	  end
	elseif val < 50 then
	  -- TODO Place trap
	else
	  -- 20% monster, 40% trap, 20% object, 20% blank space
	  if rng.percent(20) then
	    self.actor_gen:generateOne{ x=x, y=y }
	  elseif rng.percent(50) then
	    -- TODO place trap
	  elseif rng.percent(50) then
	    self.obj_gen:generateOne{ x=x, y=y }
	  end
	end
      end
    end
  end
end

-- Roughly mimics dist2() in src/generate.c, a "wrong" distance formula
-- used to generate certain special room types, notably crypts.
function _M:dist2(x1, y1, x2, y2, h1, h2, h3, h4)
  local dx, dy = math.abs(x2 - x1), math.abs(y2 - y1)
  -- As per the original T2 code:  "Basically this works by taking the
  -- normal pythagorean formula and using an expansion to express this in a
  -- way without the square root.  This approximate formula is then
  -- perturbed to give the distorted results.  (I found this by making a
  -- mistake when I was trying to fix the circular rooms.)"
  if dx >= 2*dy then return dx + math.floor((dy*h1) / h2) end
  if dy >= 2*dx then return dy + math.floor((dx*h1) / h2) end
  -- 128/181 is approx. 1/sqrt(2)
  return math.floor(((dx + dy) * 128)/181) + (math.floor(dx*dx / (dy*h3)) + math.floor(dy*dy / (dx*h3))) * h4
end

-- Roughly mimics r_visit() in src/generate.c, to build a random maze in
-- the provided area via depth-first recursive descent.
function _M:r_visit(x1, y1, x2, y2, node, dir, visited)
  -- Dimensions of the visited vertex array.
  local m = math.floor(x2 - x1)/2 + 1
  local n = math.floor(y2 - y1)/2 + 1

  -- Mark the node visited and set it to floor.
  visited[node] = true
  local x = 2*(node % m) + x1
  local y = 2*math.floor(node/m) + y1
  self.map(x, y, Map.TERRAIN, self:resolve('floor', nil, false, self.cur_lev))

  -- Set up order of adjacent node visits.
  local adj = {}
  for i = 1, 4 do adj[i] = i - 1 end
  table.shuffle(adj)
  -- 1 in 3 chance to ignore the specified direction and pick a random
  -- initial direction; otherwise the specified direcition is the initial
  -- direction.
  if rng.chance(3) then
    dir = adj[1]
  else
    table.removeFromList(adj, dir)
    table.insert(adj, 1, dir)
  end

  for _, dd in adj do
    if dd == 0 then
      -- (0,+) - check for bottom boundary.
      if math.floor(node/m) < n - 1 and not visited[node + m] then
	self.map(x, y+1, Map.TERRAIN, self:resolve('floor', nil, false, self.cur_lev))
	self:r_visit(x1, y1, x2, y2, node + m, dir, visited)
      end
    elseif dd == 1 then
      -- (0,-) - check for top boundary.
      if math.floor(node/m) > 0 and not visited[node - m] then
	self.map(x, y-1, Map.TERRAIN, self:resolve('floor', nil, false, self.cur_lev))
	self:r_visit(x1, y1, x2, y2, node - m, dir, visited)
      end
    elseif dd == 2 then
      -- (+,0) - check for right boundary.
      if node % m < m - 1 and not visited[node + 1] then
	self.map(x+1, y, Map.TERRAIN, self:resolve('floor', nil, false, self.cur_lev))
	self:r_visit(x1, y1, x2, y2, node + 1, dir, visited)
      end
    elseif dd == 3 then
      -- (0,0) - check for left boundary.
      if node % m > 0 and not visited[node - 1] then
	self.map(x-1, y, Map.TERRAIN, self:resolve('floor', nil, false, self.cur_lev))
	self:r_visit(x1, y1, x2, y2, node - 1, dir, visited)
      end
    end
  end
end

-- Roughly mimics build_cavern() in src/generate.c
function _M:build_cavern()
  local grd, roug, cutoff
  local done, lit, room = false, self.dun_lev <= rng.range(1, 25), false
  local x0, y0 = math.floor((self.map.w - 1)/2), math.floor((self.map.h - 1)/2)
  local xsize, ysize = x0*2, y0*2

  while not done do
    -- *sigh* Lua doesn't have a << operator.
    grd = util.lshift(1, rng.range(5, 8))
    roug = rng.range(1, 8) * rng.range(1, 4)
    cutoff = math.floor(xsize/2)
    local hmap = self.generate_hmap(x0, y0, xsize, ysize, grd, roug, cutoff)
    done = self:generate_fracave(x0, y0, xsize, ysize, hmap, cutoff, lit, false)
  end
end

-- Roughly mimics add_outer_wall() in src/generate.c, used by various
-- special rooms to add outer walls to a floored region.
-- Notes:
--   - Uses the 'room' map attribute to mark areas it has already covered.
--     This attribute should be .unattr()'d from the region of interest.
--   - Uses the special WALL_XTRA grid entity (and its 'xtra_hack' marker
--     field) to detect the edges of regions to be set to outer wall.
function _M:add_outer_wall(x, y, lit, x1, y1, x2, y2)
  if not self:in_bounds(x, y) then return end
  -- Grids we've already checked will have the 'room' attribute.
  if self.map.attrs(x, y, 'room') then return end
  -- To wit...
  self.map.attrs(x, y, 'room', true)
  if lit then self.map.lites(x, y, true) end

  if self:get_is_floor(x, y) then
    for xx = x - 1, x + 1 do
      for yy = y - 1, y + 1 do
	if xx >= x1 and xx <= x2 and yy >= y1 and yy <= y2 then
	  self:add_outer_wall(xx, yy, lit, x1, y1, x2, y2)
	end
      end
    end
  elseif self.map:checkEntity(x, y, Map.TERRAIN, 'xtra_hack') then
    -- Set bounding walls
    self.map(x, y, Map.TERRAIN, self:resolve('outer_wall', nil, false, self.cur_lev))
  end
end
