-- $Id: AlchemyRecipesDialog.lua 3038 2017-08-27 02:03:30Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Dialog = require 'engine.ui.Dialog'
local List = require 'engine.ui.List'
local Button = require 'engine.ui.Button'
local TextzoneList = require 'engine.ui.TextzoneList'
local Separator = require 'engine.ui.Separator'

module(..., package.seeall, class.inherit(Dialog))

function _M:init(actor, o, recipes)
  self.actor = actor
  self.base_obj = o
  self.recipes = recipes

  self.ess_order = {}
  for _, oo in ipairs(game.zone.object_list) do
    if oo.tval == 4 and oo.sval then table.insert(self.ess_order, oo) end
  end
  table.sort(self.ess_order, function(a, b) return a.sval < b.sval end)
  self.ess_order = table.extract_field(self.ess_order, 'define_as', ipairs)

  local title = self.base_obj and 'Select Alchemy Recipe' or 'Known Alchemy Recipes'
  Dialog.init(self, title, game.w*0.8, game.h*0.8)

  self:generateList()
  self.c_list = List.new {
    width = math.floor(self.iw/2 - 10),
    height = self.ih - 20,
    list = self.list,
    scrollbar = true,
    select = function(item, sel) end,
    fct = function(item, sel) self:selectRecipe(item) end,
  }

  self.c_create = Button.new {
    text = 'Create',
    fct = function()
      if self.current_item and self.current_item.recipe then
	self.actor:talentDialogReturn(self.current_item.recipe)
	game:unregisterDialog(self)
      end
    end,
  }

  self.c_desc = TextzoneList.new {
    width = math.floor(self.iw/2 - 10),
    height = self.ih - 30 - self.c_create.h,
    scrollbar = true,
  }

  local sep = Separator.new{ dir='horizontal', size=self.ih-10 }
  self:loadUI {
    { left=0, top=0, ui=self.c_list },
    { hcenter=0, top=5, ui=sep },
    { right=0, top=0, ui=self.c_desc },
    { right=0, bottom=0, ui=self.c_create, hidden=true }
  }
  self:setFocus(self.c_list)
  self:setupUI()

  self.key:addBinds {
    EXIT = function() game:unregisterDialog(self) end,
  }
end

function _M:setEntryName(item, sel)
  if sel then
    item.name = '#LIGHT_GREEN#' .. item.base_name .. '#LAST#'
  elseif not item.can_create then
    item.name = '#RED#' .. item.base_name .. '#LAST#'
  else
    item.name = item.base_name
  end
end

function _M:canCreate(def)
  if def.essences and self.actor:numRecipeEssences(def) == 0 then return false end
  if def.ingredient then
    local inv = self.actor:getInven('INVEN')
    local o = util.findInInventoryBy(inv, function(o, _, _)
      return game.zone:checkFilter(o, tag, 'object')
    end)
    if not o then return false end
  end
  if def.skill and self.actor:getSkillLevel('ALCHEMY') < def.skill then return false end
  return true
end

function _M:generateList()
  local list = {}

  local function list_entry(def)
    local item = {
      base_name = def.name,
      recipe = def,
      can_create = self:canCreate(def),
    }
    self:setEntryName(item)
    return item
  end

  if not self.recipes then
    -- Default to all known recipes, with category subheaders.
    list[#list+1] = { name='#GREY##{bold}#Object recipes:#{normal}##LAST#' }
    for _, def in ipairs(self.actor.alchemy_obj_recipe_def) do
      if self.actor:knowsAlchemyRecipe(def) then
	list[#list+1] = list_entry(def)
      end
    end
    list[#list+1] = { name='#GREY##{bold}#Ego recipes:#{normal}##LAST#' }
    for _, def in ipairs(self.actor.alchemy_ego_recipe_def) do
      if self.actor:knowsAlchemyRecipe(def) then
	list[#list+1] = list_entry(def)
      end
    end
    -- Artifact flag recipes get sub-subheaders.
    local flags, any_flags = {}, false
    for _, g in ipairs(self.actor.alchemy_flag_grp_def) do
      flags[g.tag] = {}
      for _, def in ipairs(self.actor.alchemy_flag_recipe_def.by_grp[g.tag]) do
	if self.actor:knowsAlchemyRecipe(def) then
	  table.insert(flags[g.tag], def)
	  any_flags = true
	end
      end
    end
    if any_flags then
      list[#list+1] = { name='#GREY##{bold}#Artifact flag recipes:#{normal}##LAST#' }
      for _, g in ipairs(self.actor.alchemy_flag_grp_def) do
	if #flags[g.tag] > 0 then
	  list[#list+1] = { name='#GREY##{italic}#' .. g.name .. ':#{normal}##LAST#' }
	  for _, def in ipairs(flags[g.tag]) do
	    list[#list+1] = list_entry(def)
	  end
	end
      end
    end
  else
    list[#list+1] = { name='#GREY##{bold}#Applicable recipes:#{normal}##LAST#' }
    for _, def in ipairs(self.recipes) do
      list[#list+1] = list_entry(def)
    end
  end

  self.list = list
end

function _M:selectRecipe(item)
  if not item or not item.recipe then
    self:getUIElement(self.c_create).hidden = true
    return
  end
  if self.current_item then self:setEntryName(self.current_item) end
  self.current_item = item
  self:setEntryName(item, true)

  self.c_desc:switchItem(item.recipe.define_as, self:getDescription(item.recipe))
  self:getUIElement(self.c_create).hidden = not self.base_obj or not item.can_create
end

function _M:getDescription(def)
  local desc = '#{bold}##LIGHT_BLUE#' .. def.name .. '#LAST##{normal}#\n'

  if def.desc then desc = desc .. '\n' .. def.desc end
  if def.base_obj then
    desc = desc .. '\n#{bold}#Required base object:#{normal}#'
    desc = desc .. '\n' .. self:listItem(def.base_obj)
  end
  if def.essences then
    desc = desc .. '\n#{bold}#Required essences:#{normal}#'
    for _, ess in ipairs(self.ess_order) do
      if def.essences[ess] then
	desc = desc .. '\n' .. self:listItem(ess, def.essences[ess])
      end
    end
  end
  if def.exp then
    if def.multiple then
      local t = {}
      for i = 1, 5 do
	t[i] = ('%+d=%d'):format(i, util.alchemy.exp_needed_for_recipe(def, i))
      end
      t = table.concat(t, ', ')
      desc = desc .. ('\n#{bold}#Required artifact exp:#{normal}#  %d base\n\t(%s)'):format(def.exp, t)
    else
      desc = desc .. ('\n#{bold}#Required artifact exp:#{normal}#  %d'):format(def.exp)
    end
  end
  if def.skill then
    local c = self.actor:getSkillLevel('ALCHEMY') >= def.skill and 'GREEN' or 'RED'
    desc = desc .. ('\n#{bold}#Required Alchemy skill:#{normal}#  #LIGHT_%s#%d#LAST#'):format(c, def.skill)
  end
  if def.ingredient then
    local ing = def.ingredient
    desc = desc .. '\n#{bold}#Required ingredient:#{normal}#'
    if def.keep_open then
      desc = desc .. '\n' .. self:listItem(ing.match, game.player:getSkillLevel('ALCHEMY'), ing.name, ing.plural_name)
    else
      desc = desc .. '\n' .. self:listItem(ing.match, 1, ing.name)
      if def.multiple then
	desc = desc .. ' (one per pval)'
      end
    end
  end
  if def.ego_def then
    local Ego = require 'mod.class.Ego'
    local ego = Ego:get(def.ego_def)
    if ego then
      desc = desc .. '\n#{bold}#Can be applied to:#{normal}#'
      for _, oo in ipairs(game.zone.object_list) do
	if ego:canApplyTo(oo, true) then
	  desc = desc .. '\n· ' .. oo:getName{no_count=true, force_qty=1, force_id=true, no_mods=true, no_pval=true}
	end
      end
    end
  end

  return desc:toTString()
end

function _M:listItem(tag, n, name, plural_name)
  n = n or 1
  local inv, oo = self.actor:getInven('INVEN'), nil
  if type(tag) == 'table' then
    oo = util.findInInventoryBy(inv, function(o, _, _)
      return game.zone:checkFilter(o, tag, 'object')
    end)
  else
    oo = self.actor:findInInventoryBy(inv, 'define_as', tag)
  end
  local c = oo and oo:getNumber() >= n and 'GREEN' or 'RED'
  if not name and type(tag) == 'string' then
    local o = game.zone.object_list[tag]
    if not o then
      return ('· #LIGHT_RED#%d ???(%s)#LAST#'):format(n, tag)
    end
    name = o:getName{force_qty=n, force_id=true, no_mods=true, no_pval=true}
  end
  if plural_name and n > 1 then
    name = ('%d %s'):format(n, plural_name)
  end
  return ('· #LIGHT_%s#%s#LAST#'):format(c, name or '???')
end
