-- $Id: AutomatizerMakeClause.lua 3677 2019-09-18 02:28:35Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2018-2019 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Dialog = require 'engine.ui.Dialog'
local Checkbox = require 'engine.ui.Checkbox'
local RadioButton = require 'mod.dialogs.ui.RadioButton'
local ShortDropdown = require 'mod.dialogs.ui.ShortDropdown'
local OpaqueDropdown = require 'mod.dialogs.ui.OpaqueDropdown'
local Textbox = require 'engine.ui.Textbox'
local Textzone = require 'engine.ui.Textzone'
local TextzoneList = require 'engine.ui.TextzoneList'
local Numberbox = require 'engine.ui.Numberbox'
local Button = require 'engine.ui.Button'
local SubGroup = require 'mod.dialogs.ui.SubGroup'
local CardLayout = require 'mod.dialogs.ui.CardLayout'
local Automatizer = require 'mod.class.Automatizer'
local AutomatizerOneOfEditor = require 'mod.dialogs.AutomatizerOneOfEditor'

module(..., package.seeall, class.inherit(Dialog))

_M.simple_tests = {
  { name='Object is identified', test='is_identified' },
  { name='Player knows talent:', test='has_talent', has_arg=true },
  { name='Player has ability:', test='has_ability', has_arg=true },
  { name='Addon is installed:', test='has_addon', has_arg=true },
}
_M.values = {
  { name='Object name', value='name', valtype='string' },
  { name='Object inscription', value='inscription', valtype='string' },
  { name='Object ASCII symbol', value='symbol', valtype='string' },
  { name='Object type', value='type', valtype='string' },
  { name='Object subtype', value='subtype', valtype='string' },
  { name='Object type/subtype', value='full_type', valtype='string' },
  { name='Object internal ID', value='define_as', valtype='string' },
  { name='Object discount', value='discount', valtype='number' },
  { name='Object status', value='status', valtype='string', group='status', grp_only=true },
  { name='Player race', value='race', valtype='string', group='race' },
  { name='Player subrace', value='subrace', valtype='string', group='subrace' },
  { name='Player class', value='class', valtype='string', group='class' },
  { name='Player god', value='god', valtype='string', group='god' },
  { name='Player descriptor:', value='descriptor', valtype='string', has_arg=true },
  { name='Player level', value='level', valtype='number' },
  { name='Player skill:', value='skill', valtype='number', has_arg=true },
}
_M.ops = {
  { name='equals', op='==', for_valtype={string=true, number=true} },
  { name='does not equal', op='!=', for_valtype={string=true, number=true} },
  { name='is less than', op='<', for_valtype={number=true} },
  { name='is less than or equal to', op='<=', for_valtype={number=true} },
  { name='is greater than', op='>', for_valtype={number=true} },
  { name='is greater than or equal to', op='>=', for_valtype={number=true} },
  { name='is between', op='inrange', for_valtype={number=true} },
  { name='contains', op='contains', for_valtype={string=true}, no_grp_only=true },
  { name='matches', op='matches', for_valtype={string=true}, no_grp_only=true },
  { name='is one of', op='oneof', for_valtype={string=true} },
}

function _M:init(orig_clause, apply_cb)
  self.orig_clause = orig_clause
  self.apply_cb = apply_cb

  local title = 'Modify Test'
  if not orig_clause or orig_clause.test == 'placeholder' then
    title = 'New Test'
  end
  Dialog.init(self, title, 1, 1)

  -- Build caches of dropdown items by relevant values.
  local simple_by_test, values_by_name, ops_by_test = {}, {}, {}
  for _, item in ipairs(self.simple_tests) do
    simple_by_test[item.test] = item
  end
  for _, item in ipairs(self.values) do
    values_by_name[item.value] = item
  end
  for _, item in ipairs(self.ops) do
    ops_by_test[item.op] = item
  end

  -- First pass:  decide which radio button to select by default.
  local init_sel = 3
  if orig_clause then
    if orig_clause.test == 'and' then
      init_sel = 1
    elseif orig_clause.test == 'or' then
      init_sel = 2
    elseif simple_by_test[orig_clause.test] then
      init_sel = 3
    elseif ops_by_test[orig_clause.test] then
      init_sel = 4
    end
  end

  self.rb_and = RadioButton.new {
    title = '"And" group (all sub-tests must match)',
    radio_group = 'clause_type',
    default = (init_sel == 1),
  }
  self.rb_or = RadioButton.new {
    title = '"Or" group (at least one sub-test must match)',
    radio_group = 'clause_type',
    default = (init_sel == 2),
  }
  self.rb_simple = RadioButton.new {
    title = 'Simple test:',
    radio_group = 'clause_type',
    default = (init_sel == 3),
  }
  self.dd_simple = OpaqueDropdown.new {
    width = 250,
    list = self.simple_tests,
    nb_items = #self.simple_tests,
    fct = function(item) self:selectSimple(item) end,
    alpha = 1,
  }
  self.cb_simple_negate = Checkbox.new {
    title = 'Is not true',
    default = false,
  }
  self.tb_simple_arg = Textbox.new {
    title = '',
    text = '',
    chars = 20,
    max_len = 99,
    fct = function() end,
  }
  self.rb_compar = RadioButton.new {
    title = 'Comparison test:',
    radio_group = 'clause_type',
    default = (init_sel == 4),
  }
  self.dd_value = OpaqueDropdown.new {
    width = 250,
    list = self.values,
    nb_items = 12,
    fct = function(item) self:selectValue(item) end,
    scrollbar = true,
    alpha = 1,
  }
  self.tb_value_arg = Textbox.new {
    title = '',
    text = '',
    chars = 20,
    max_len = 99,
    fct = function() end,
  }
  self.dd_op = OpaqueDropdown.new {
    width = 250,
    list = {
      -- This will get updated in :selectValue().
      { name = 'equals', op='==' },
      { name = 'does not equal', op='!=' },
    },
    nb_items = 2,
    fct = function(item) self:selectOp(item) end,
    alpha = 1,
  }
  self.cb_compar_negate = Checkbox.new {
    title = 'Is not true',
    default = false,
  }
  self.nb_target = Numberbox.new {
    title = '',
    number = 0,
    min = 0,
    max = 999,
    chars = 3,
    fct = function() end,
  }
  self.nb_min = Numberbox.new {
    title = '',
    number = 0,
    min = 0,
    max = 999,
    chars = 3,
    fct = function() end,
  }
  self.nb_max = Numberbox.new {
    title = ' and ',
    number = 0,
    min = 0,
    max = 999,
    chars = 3,
    fct = function() end,
  }
  self.tb_target = Textbox.new {
    title = '',
    text = '',
    chars = 20,
    max_len = 99,
    fct = function() end,
  }
  self.dd_grpsel = OpaqueDropdown.new {
    width = 175,
    fct = function() end,
    -- These will be populated in :updateGroupSelector().
    list = { { name='Placeholder', value='' } },
    nb_items = 1,
    alpha = 1,
  }
  self.tb_target_g = Textbox.new {
    title = '',
    text = '',
    chars = 20,
    max_len = 99,
    fct = function() end,
  }
  self.sd_grpsel_t = ShortDropdown.new {
    text = 'Select',
    -- This will be populated in :updateGroupSelector().
    list = { { name='Placeholder', value='' } },
    max_shown = 8,
    fct = function(item)
      if item and item.value and #item.value > 0 then
	self.tb_target_g:setText(item.value)
      end
    end,
    alpha = 1,
  }
  self.tl_oneof = TextzoneList.new {
    width = 175,
    height = self.font_h + 6,
    weakstore = true,
  }
  self.b_oneof = Button.new {
    text = 'Edit',
    fct = function() self:editOneOf() end,
  }

  self.cards = CardLayout.new {
    empty = {},
    number = {
      { right=0, top=0, ui=self.nb_target },
    },
    nb_minmax ={
      { left=0, top=0, ui=self.nb_min },
      { left=self.nb_min, top=0, ui=self.nb_max },
    },
    text = {
      { right=0, top=0, ui=self.tb_target },
    },
    grpsel = {
      { left=0, top=0, ui=self.dd_grpsel },
    },
    text_with_grpsel = {
      { left=0, top=0, ui=self.tb_target_g },
      { right=0, top=0, ui=self.sd_grpsel_t },
    },
    oneof = {
      { left=0, top=0, ui=self.tl_oneof },
      { right=0, top=0, ui=self.b_oneof },
    },
  }
  self.cards:set_initial_card('empty')

  self.b_ok = Button.new {
    text = 'OK',
    fct = function() self:finish() end,
  }
  self.b_cancel = Button.new {
    text = 'Cancel',
    fct = function() self.key:triggerVirtual('EXIT') end,
  }

  self:loadUI {
    { left=0, top=0, ui=self.rb_and },
    { left=0, top=self.rb_and, ui=self.rb_or },
    { left=0, top=self.rb_or, ui=self.rb_simple },
    { left=20, top=self.rb_simple, ui=self.dd_simple },
    { left=self.dd_simple, top=self.rb_simple, ui=self.cb_simple_negate },
    { left=20, top=self.dd_simple, ui=self.tb_simple_arg, hidden=true },
    { left=0, top=self.tb_simple_arg, ui=self.rb_compar },
    { left=20, top=self.rb_compar, ui=self.dd_value },
    { left=self.dd_value, top=self.rb_compar, ui=self.tb_value_arg, hidden=true },
    { left=20, top=self.dd_value, ui=self.dd_op },
    { left=self.dd_op, top=self.dd_value, ui=self.cb_compar_negate },
    { left=40, top=self.dd_op, ui=self.cards:current_card() },
    { left=0, bottom=0, ui=self.b_ok },
    { left=self.b_ok, bottom=0, ui=self.b_cancel },
  }
  self:setupUI(true, true)
  self:selectSimple(self:dropdownSel('dd_simple'), true)
  self:selectValue(self:dropdownSel('dd_value'), true)

  -- Second pass:  populate dropdowns and fields according to the original
  -- clause.
  if orig_clause then
    if init_sel == 3 then
      if orig_clause.test ~= 'placeholder' then
	ui_util.selDropdownByField(self.dd_simple, 'test', orig_clause.test)
	if simple_by_test[orig_clause.test].has_arg then
	  self.tb_simple_arg:setText(orig_clause[1])
	end
      end
    end
    if init_sel == 4 then
      local val_sel = values_by_name[orig_clause[1].clause]
      ui_util.selDropdownByField(self.dd_value, 'value', orig_clause[1].clause)
      ui_util.selDropdownByField(self.dd_op, 'op', orig_clause.test)
      if val_sel.has_arg then
	self.tb_value_arg:setText(orig_clause[1][1])
      end
      if orig_clause.test == 'inrange' then
	self.nb_min.number = orig_clause[2]
	self.nb_min:updateText(true)
	self.nb_max.number = orig_clause[3]
	self.nb_max:updateText(true)
      elseif orig_clause.test == 'oneof' then
	self.oneof_list = {}
	for i = 2, #orig_clause do self.oneof_list[i-1] = orig_clause[i] end
	self:updateOneOfDesc()
      elseif val_sel.grp_only then
	ui_util.selDropdownByField(self.dd_grpsel, 'value', orig_clause[2])
      elseif val_sel.group then
	self.tb_target_g:setText(orig_clause[2])
      elseif val_sel.valtype == 'number' then
	self.nb_target.number = orig_clause[2]
	self.nb_target:updateText(true)
      else
	self.tb_target:setText(orig_clause[2])
      end
    end
  end

  self.key:addBinds {
    EXIT = function() game:unregisterDialog(self) end
  }
end

function _M:finish()
  local errmsg = nil
  if self.rb_simple.checked then
    local test_sel = self:dropdownSel('dd_simple')
    if test_sel.has_arg then
      local ok = self.tb_simple_arg.text and #self.tb_simple_arg.text > 0
      if not ok then
	errmsg = 'Missing parameter for '..test_sel.name..' test.'
      end
    end
  elseif self.rb_compar.checked then
    local val_sel = self:dropdownSel('dd_value')
    if val_sel.has_arg then
      local ok = self.tb_value_arg.text and #self.tb_value_arg.text > 0
      if not ok then
	errmsg = 'Missing parameter for '..test_sel.name..' test.'
      end
    end
    if self.cards.cur_card == 'oneof' and (not self.oneof_list or #self.oneof_list == 0) then
      errmsg = 'No values specified for '..test_sel.name..'.'
    end
    if self.cards.cur_card == 'grpsel' and not self:dropdownSel('dd_grpsel') then
      errmsg = 'Comparison with '..val_sel.name..' needs target value.'
    end
    if self.cards.cur_card == 'text_with_grpsel' then
      local ok = self.tb_target_g.text and #self.tb_target_g.text > 0
      if not ok then
	errmsg = 'Comparison with '..val_sel.name..' needs target value.'
      end
    end
  end
  if errmsg then
    Dialog:simplePopup('Missing value', errmsg)
    return
  end

  if self.apply_cb then
    local clause = self:buildClause()
    self.apply_cb(clause)
  end
  self.key:triggerVirtual('EXIT')
end

function _M:dropdownSel(tag)
  return self[tag].c_list.list[self[tag].c_list.sel]
end

function _M:selectSimple(item, no_select)
  if not item then return end
  self:toggleDisplay(self.tb_simple_arg, item.has_arg)
  if not no_select and not self.rb_simple.checked then
    self.rb_simple:select()
  end
end

function _M:selectValue(item, no_select)
  if not item then return end
  self:toggleDisplay(self.tb_value_arg, item.has_arg)
  if not no_select and not self.rb_compar.checked then
    self.rb_compar:select()
  end

  local ops = table.greplist(self.ops, function(op)
    return op.for_valtype[item.valtype] and not (op.no_grp_only and item.grp_only)
  end)
  self.dd_op.list = ops
  self.dd_op.nb_items = math.min(#ops, 8)
  self.dd_op.scrollbar = (#ops > 8)
  local ui = self:getUIElement(self.dd_op)
  if ui and ui.x then
    self.dd_op:positioned(ui.x, ui.y, self.display_x + ui.x, self.display_y + ui.y)
  end
  self:selectOp(self:dropdownSel('dd_op'), true)
end

function _M:selectOp(item, no_select)
  if not item or not item.op then return end
  if not no_select and not self.rb_compar.checked then
    self.rb_compar:select()
  end

  local val_sel = self:dropdownSel('dd_value')
  local new_card = item.op == 'inrange' and 'nb_minmax' or
		   item.op == 'oneof' and 'oneof' or
		   val_sel.grp_only and not item.no_grp_only and 'grpsel' or
		   val_sel.group and not item.no_grp_only and 'text_with_grpsel' or
		   val_sel.valtype == 'number' and 'number' or 'text'
  self.cards:show_card(new_card, self)
  if new_card == 'grpsel' or new_card == 'text_with_grpsel' then
    self:updateGroupSelector()
  elseif new_card == 'oneof' then
    -- TODO Maybe remember these between values.
    self.oneof_list = {}
    self:updateOneOfDesc()
  end
end

function _M:updateGroupSelector()
  local val_sel = self:dropdownSel('dd_value')
  if not val_sel.group then return end

  local list = { { name='Placeholder', value='' } }
  local function list_cb(_, v) return { name=tostring(v), value=v } end
  local vals_src = Automatizer.test_allowed_values[val_sel.group]
  if type(vals_src) == 'table' then
    list = table.maplist(list_cb, vals_src)
  elseif type(vals_src) == 'function' then
    list = table.maplist(list_cb, vals_src())
  end

  if self.cards.cur_card == 'grpsel' then
    self.dd_grpsel.list = list
    self.dd_grpsel.nb_items = math.min(#list, 8)
    self.dd_grpsel.scrollbar = (#list > 8)
  elseif self.cards.cur_card == 'text_with_grpsel' then
    self.sd_grpsel_t.list = list
    self.sd_grpsel_t.scrollbar = #list > 8
    self.sd_grpsel_t.list_w = nil	-- Recompute for new list
    self.tb_target_g:setText('')
  end

  -- Slight Hack(TM):  Since this selector is in a card, call :positioned()
  -- on the entire card to propagate the right x/y down to the selector.
  local ui = self:getUIElement(self.cards:current_card())
  if ui and ui.x then
    self.cards:current_card():positioned(ui.x, ui.y, self.display_x + ui.x, self.display_y + ui.y)
  end
end

function _M:editOneOf()
  local function cb(vals)
    if vals and #vals > 0 then
      self.oneof_list = vals
      self:updateOneOfDesc()
    end
  end
  local val_sel = self:dropdownSel('dd_value')
  game:registerDialog(AutomatizerOneOfEditor.new(val_sel.name, val_sel.group, val_sel.grp_only, table.clone(self.oneof_list or {}), cb))
end

function _M:updateOneOfDesc()
  if not self.oneof_list or #self.oneof_list == 0 then
    self.tl_oneof:switchItem('@NONE@', '#{italic}##GREY#--no values--#LAST##{normal}#')
  else
    local text = ('%d item%s'):format(#self.oneof_list, #self.oneof_list == 1 and '' or 's')
    for n = #self.oneof_list, 1, -1 do
      local cand = table.concat(self.oneof_list, ', ', 1, n)
      if n < #self.oneof_list then
	cand = cand .. ' + ' .. (#self.oneof_list - n) .. ' more'
      end
      if self.font:size(cand) <= 175 then
	text = cand
	break
      end
    end
    self.tl_oneof:switchItem(text, text)
  end
end

function _M:buildClause()
  if self.rb_and.checked then
    return { test='and', { test='placeholder' } }
  end
  if self.rb_or.checked then
    return { test='or', { test='placeholder' } }
  end
  if self.rb_simple.checked then
    local item = self:dropdownSel('dd_simple')
    local ret = { test=item.test }
    if item.has_arg then ret[1] = self.tb_simple_arg.text end
    if self.cb_simple_negate.checked then ret = { test='not', ret } end
    return ret
  end
  if self.rb_compar.checked then
    local val_sel = self:dropdownSel('dd_value')
    local op_sel = self:dropdownSel('dd_op')
    local ret = { test=op_sel.op, { clause=val_sel.value } }
    if val_sel.has_arg then ret[1][1] = self.tb_value_arg.text end
    if self.cards.cur_card == 'nb_minmax' then
      ret[2] = self.nb_min.number
      ret[3] = self.nb_max.number
    elseif self.cards.cur_card == 'oneof' then
      for i, v in ipairs(self.oneof_list or {}) do ret[i+1] = v end
    elseif self.cards.cur_card == 'number' then
      ret[2] = self.nb_target.number
    elseif self.cards.cur_card == 'grpsel' then
      local grp_sel = self:dropdownSel('dd_grpsel')
      ret[2] = grp_sel.value
    elseif self.cards.cur_card == 'text_with_grpsel' then
      ret[2] = self.tb_target_g.text
    else
      ret[2] = self.tb_target.text
    end
    if self.cb_compar_negate.checked then ret = { test='not', ret } end
    return ret
  end
  return { test='placeholder' }
end
