-- $Id: MapMenu.lua 4247 2021-07-21 02:42:47Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Dialog = require 'engine.ui.Dialog'
local List = require 'engine.ui.List'
local Map = require 'mod.class.Map'

module(..., package.seeall, class.inherit(Dialog))

function _M:init(mx, my, tmx, tmy, extra)
  self.tmx = util.bound(tmx, 0, game.level.map.w - 1)
  self.tmy = util.bound(tmy, 0, game.level.map.h - 1)
  self.on_player = tmx == game.player.x and tmy == game.player.y
  self.extra = extra

  self:generateList()
  self.__showup = false

  Dialog.init(self, 'Actions', 1, 100, mx, my)

  self.c_list = List.new {
    width = math.max(self.max_w + 10, self.font_bold:size(self.title)),
    nb_items = math.min(15, #self.list),
    scrollbar = #self.list > 15,
    list = self.list,
    fct = function(item) self:use(item) end,
    select = function(item)self:select(item) end,
  }

  self:loadUI {
    { left=0, top=0, ui=self.c_list },
  }
  self:setupUI(true, true, function(w, h)
    self.force_x = mx - w / 2
    self.force_y = my - self.h + (self.ih + self.c_list.fh / 3)
    if self.force_y + h > game.h then self.force_y = game.h - h end
  end)

  local function exit_cb(button, x, y, xrel, yrel, bx, by, event)
    if (button == "left" or button == "right") and event == "button" then
      self.key:triggerVirtual("EXIT")
    end
  end
  local function act_cb(button, x, y, xrel, yrel, bx, by, event)
    if button == "right" and event == "button" then
      self.key:triggerVirtual("EXIT")
    else
      self:mouseEvent(button, x, y, xrel, yrel, bx, by, event)
    end
  end
  self.mouse:reset()
  self.mouse:registerZone(0, 0, game.w, game.h, exit_cb)
  self.mouse:registerZone(self.display_x, self.display_y, self.w, self.h, act_cb)
  self.key:addBinds {
    EXIT = function() game:unregisterDialog(self) end,
  }
end

function _M:unload()
  Dialog.unload(self)
  game:targetMode(false, false)
  self.exited = true
end

function _M:use(item)
  if not item then return end
  game:unregisterDialog(self)
  game:targetMode(false, false)

  if item.action == 'move_to' then
    game.player:mouseMove(self.tmx, self.tmy, true)
  elseif item.action == 'debug_teleport' then
    game.player:move(self.tmx, self.tmy, true)
  elseif item.action == 'rename' then
    game.player:playerSetPetName(item.entity)
  elseif item.action == 'visuals' then
    local VisualsDialog = require 'mod.dialogs.VisualsDialog'
    local GameOptions = require 'mod.dialogs.GameOptions'
    local d = VisualsDialog:create(item.type)
    d:selectEntity(item.entity)
    d.unload = function(dlg_self)
      local val = dlg_self:get_value()
      if type(val) ~= 'nil' then
	local key = 'ascii_'..item.type..'_symbols'
	config.settings.tome2[key] = val
	VisualsDialog:updateForEntityType(item.type, val)
	game.level.map:redisplay()
	GameOptions:saveOptions()
      end
    end
    game:registerDialog(d)
  elseif item.action == 'debug-inspect' then
    local DebugConsole = require 'engine.DebugConsole'
    local d = DebugConsole.new()
    game:registerDialog(d)
    DebugConsole.line = "=__uids["..item.entity.uid.."]"
    DebugConsole.line_pos = #DebugConsole.line
    d.changed = true
  elseif item.action == 'debug-mapattrs' then
    local DebugConsole = require 'engine.DebugConsole'
    local d = DebugConsole.new()
    game:registerDialog(d)
    DebugConsole.line = "=game.level.map.attrs["..(self.tmx + self.tmy*game.level.map.w).."]"
    DebugConsole.line_pos = #DebugConsole.line
    d.changed = true
  elseif item.action == 'debug-spots' then
    local spots = {}
    for _, spot in ipairs(game.level.spots) do
      if spot.x == self.tmx and spot.y == self.tmy then
	spots[#spots+1] = { name = spot.type..'/'..spot.subtype }
      end
    end
    local text = ('Spots at map location %d,%d:'):format(self.tmx, self.tmy)
    ui_util.listPopup('Spots', text, spots, 300)
  elseif item.action == 'debug-inventory' then
    local d
    local function post_use_cb(_, _, _, _, stop)
      d:generate()
      d:generateList()
      if stop then self:unregisterDialog(d) end
    end
    local function use_cb(o, inven, item_, button, event)
      if o then
	local UseItemDialog = require 'mod.dialogs.UseItemDialog'
	local center = event == 'button'
	local ud = UseItemDialog.new(center, item.actor, o, item_, post_use_cb)
	self:registerDialog(ud)
      end
    end
    d = item.actor:showEquipInven(item.actor.name..': Inventory', nil, use_cb)
  elseif item.action == 'debug-trap-known' then
    item.entity:setKnown(game.player, item.known)
    game.level.map:updateMap(self.tmx, self.tmy)
  elseif item.action == 'debug-trap-remove' then
    game.level.map:remove(self.tmx, self.tmy, Map.TRAP)
    if item.entity.removed then
      item.entity:removed(self.tmx, self.tmy, game.player)
    end
  elseif item.action == 'debug-place-trap' then
    local list = { {name='Random'} }
    for _, t in ipairs(game.zone.trap_list) do
      list[#list+1] = { name=t.name, tag=t.define_as }
    end
    local function cb(sel)
      if sel then
	trap_util.generator():generateOne{x=self.tmx, y=self.tmy, name=sel.tag}
      end
    end
    Dialog:listPopup('Place Trap', 'Select trap type to place:', list, 600, game.h*0.8, cb)
  elseif item.action == 'debug-secret-door' then
    game.level.map(self.tmx, self.tmy, Map.TERRAIN, game.zone.grid_list.DOOR)
    game.level.map:updateMap(self.tmx, self.tmy)
  elseif item.action == 'debug-place-feat' then
    local list = {
      { name='Rubble', feat='RUBBLE' },
      { name='Fountain', feat='FOUNTAIN' },
      { name='Void jumpgate', feat='VOID_JUMPGATE' },
      { name='Altar of Darkness', feat='ALTAR_OF_DARKNESS' },
    }
    local function cb(sel)
      if sel then
	local gen = game.zone:getGenerator('map', game.level, game.level.data.generator.map)
	if sel.feat == 'FOUNTAIN' then gen:buildPotionCache() end
	gen:place_special_feature(self.tmx, self.tmy, sel.feat, true, game.level.level)
      end
    end
    Dialog:listPopup('Place Special Feature', 'Select feature type to place:', list, 600, game.h*0.8, cb)
  end
end

local olditem = nil
function _M:select(item)
  if self.exited then return end
  if not item then return end
  if olditem and olditem == item then return end
  if not item.set_target then game:targetMode(false, false) return end

  game:targetMode(true, nil, nil, item.set_target)
  game.target:setSpot(self.tmx, self.tmy, "forced")
end

function _M:generateList()
  local list = {}
  local p = game.player
  local g = game.level.map(self.tmx, self.tmy, Map.TERRAIN)
  local t = game.level.map(self.tmx, self.tmy, Map.TRAP)
  local o = game.level.map(self.tmx, self.tmy, Map.OBJECT)
  local a = game.level.map(self.tmx, self.tmy, Map.ACTOR)
  local ok_trap = not t and game.level.map:checkEntity(self.tmx, self.tmy, Map.TERRAIN, 'naked_bold') and not game.level.map:checkEntity(self.tmx, self.tmy, Map.TERRAIN, 'hasFlags', 'DOOR')

  if g and not self.on_player then
    list[#list+1] = {
      name = 'Move to location',
      action = 'move_to',
      color = colors.simple(colors.ANTIQUE_WHITE),
    }
    if config.settings.cheat then
      list[#list+1] = {
	name = 'Teleport to location',
	action = 'debug_teleport',
	color = colors.simple(colors.LIGHT_BLUE),
      }
    end
  end
  if a and a.faction == 'pets' and not a.unique then
    list[#list+1] = {
      name = 'Change pet/companion name',
      action = 'rename',
      entity = a,
      color = colors.simple(colors.ANTIQUE_WHITE),
    }
  end
  if g then
    list[#list+1] = {
      name = 'Modify visuals [terrain]',
      action = 'visuals',
      entity = g,
      type = 'grid',
      color = colors.simple(colors.ANTIQUE_WHITE),
    }
  end
  if a and not self.on_player then
    list[#list+1] = {
      name = 'Modify visuals [actor]',
      action = 'visuals',
      entity = a,
      type = 'actor',
      color = colors.simple(colors.ANTIQUE_WHITE),
    }
  end
  if o then
    list[#list+1] = {
      name = 'Modify visuals [object]',
      action = 'visuals',
      entity = o,
      type = 'object',
      color = colors.simple(colors.ANTIQUE_WHITE),
    }
  end
  if g and config.settings.cheat then
    if g.define_as == 'SECRET_DOOR' then
      list[#list+1] = {
	name = 'Reveal secret door',
	action = 'debug-secret-door',
	color = colors.simple(colors.LIGHT_BLUE),
      }
    end
    list[#list+1] = {
      name = 'Lua inspect [terrain]',
      action = 'debug-inspect',
      entity = g,
      color = colors.simple(colors.LIGHT_BLUE),
    }
    list[#list+1] = {
      name = 'Lua inspect [map attrs]',
      action = 'debug-mapattrs',
      color = colors.simple(colors.LIGHT_BLUE),
    }
    for _, spot in ipairs(game.level.spots) do
      if spot.x == self.tmx and spot.y == self.tmy then
	list[#list+1] = {
	  name = 'Level spots',
	  action = 'debug-spots',
	  color = colors.simple(colors.LIGHT_BLUE),
	}
	break
      end
    end
  end
  if a and config.settings.cheat then
    list[#list+1] = {
      name = 'Lua inspect [actor]',
      action = 'debug-inspect',
      entity = a,
      color = colors.simple(colors.LIGHT_BLUE),
    }
    list[#list+1] = {
      name = 'Show inventory',
      action = 'debug-inventory',
      actor = a,
      color = colors.simple(colors.YELLOW),
    }
  end
  if o and config.settings.cheat then
    list[#list+1] = {
      name = 'Lua inspect [object]',
      action = 'debug-inspect',
      entity = o,
      color = colors.simple(colors.LIGHT_BLUE),
    }
  end
  if t and config.settings.cheat then
    list[#list+1] = {
      name = 'Lua inspect [trap]',
      action = 'debug-inspect',
      entity = t,
      color = colors.simple(colors.LIGHT_BLUE),
    }
    local known = t:knownBy(game.player)
    list[#list+1] = {
      name = known and 'Unlearn trap' or 'Learn trap',
      action = 'debug-trap-known',
      entity = t,
      known = not known,
      color = colors.simple(colors.LIGHT_BLUE),
    }
    list[#list+1] = {
      name = 'Remove trap',
      action = 'debug-trap-remove',
      entity = t,
      color = colors.simple(colors.LIGHT_BLUE),
    }
  end
  if ok_trap and config.settings.cheat then
    list[#list+1] = {
      name = 'Place trap',
      action = 'debug-place-trap',
      color = colors.simple(colors.LIGHT_BLUE),
    }
    list[#list+1] = {
      name = 'Place special feature',
      action = 'debug-place-feat',
      color = colors.simple(colors.LIGHT_BLUE),
    }
  end

  self.max_w = 0
  for _, v in ipairs(list) do
    local w, _ = self.font:size(v.name)
    self.max_w = math.max(self.max_w, w)
  end

  self.list = list
end
