-- $Id: UseItemDialog.lua 3834 2020-03-29 02:42:36Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2020 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require "engine.class"
local Dialog = require "engine.ui.Dialog"
local GetQuantity = require 'engine.dialogs.GetQuantity'
local GetText = require 'engine.dialogs.GetText'
local List = require "engine.ui.List"

module(..., package.seeall, class.inherit(Dialog))

function _M:init(center_mouse, actor, object, item, inven, floor, on_use)
  self.actor = actor
  self.object = object
  self.inven = inven
  self.floor = floor
  self.item = item
  self.on_use = on_use

  self:generateList()
  local name = object:getName()
  Dialog.init(self, name, 1, 1)

  local w = self.font_bold:size(name)
  local list = List.new {
    width = math.max(w, self.max_w) + 10,
    nb_items = #self.list,
    list = self.list,
    fct = function(item) self:use(item) end,
  }

  local center_cb = function(w, h)
    if center_mouse then
      local mx, my = core.mouse.get()
      self.force_x = mx - w / 2
      self.force_y = my - (self.h - self.ih + list.fh / 3)
    end
  end
  self:loadUI { { left=0, top=0, ui=list } }
  self:setupUI(true, true, center_cb)

  self.key:addBinds {
    EXIT = function() game:unregisterDialog(self) end
  }
end

function _M:use(item)
  if not item then return end
  game:unregisterDialog(self)
  print(("[USE ITEM] name='%s' action=%s talent=%s"):format(item.name, item.action or 'nil', item.talent_id or 'nil'))

  if item.action == 'use_item' then
    self.actor:playerUseItem(self.object, self.item, self.inven, self.floor)
    self.on_use(self.inven, self.floor, self.item, self.object, true)
  elseif item.action == 'eat' then
    self.actor:playerEat(self.object, self.item, self.inven, self.floor)
    self.on_use(self.inven, self.floor, self.item, self.object, true)
  elseif item.action == 'fill' then
    -- cf. do_cmd_refill_lamp() in src/cmd3.c
    local amt = util.getval(self.object.oil, self.object)
    local inv = self.actor:getInven('LITE')
    -- We've already checked that this exists.
    local lite = inv[1]
    -- TODO Find someplace to define this instead of hard-coding it.
    lite.lite_fuel = math.min(15000, lite.lite_fuel + amt)
    local name = lite:getName{no_count=true, no_mods=true, no_pval=true, no_inscr=true}
    game.log(lite.lite_fuel >= 15000 and 'Your %s is full.' or 'You fuel your %s.', name)
    self.actor:getOneObject(self.inven, self.item, self.floor, {vocal_left=true})
    -- Slight Hack(TM):  Refilling a lantern uses only half a turn.
    self.actor:useEnergy(game.energy_to_act/2)
    self.on_use(self.inven, self.floor, self.item, self.object, true)
  elseif item.action == 'retorch' then
    -- cf. do_cmd_refill_torch() in src/cmd3.c
    local inv = self.actor:getInven('LITE')
    -- We've already checked that this exists.
    local lite = inv[1]
    -- TODO Find someplace to define this instead of hard-coding it.
    lite.lite_fuel = math.min(4000, lite.lite_fuel + self.object.lite_fuel + 5)
    game.log('You combine the torches')
    local name = lite:getName{no_count=true, no_mods=true, no_pval=true, no_inscr=true}
    game.log(lite.lite_fuel >= 4000 and 'Your %s is fully fueled.' or 'Your %s glows more brightly.', name)
    self.actor:getOneObject(self.inven, self.item, self.floor, {vocal_left=true})
    -- Slight Hack(TM):  Combining torches uses only half a turn.
    self.actor:useEnergy(game.energy_to_act/2)
    self.on_use(self.inven, self.floor, self.item, self.object, true)
  elseif item.action == 'browse' then
    local UseTalents = require('mod.dialogs.UseTalents')
    local d = UseTalents.new(self.actor, self.object)
    game:registerDialog(d)
    self.on_use(self.inven, self.floor, self.item, self.object, true)
  elseif item.action == 'wear' then
    self.actor:doWear(self.inven, self.floor, self.item, self.object)
    self.on_use(self.inven, self.floor, self.item, self.object, true)
  elseif item.action == 'takeoff' then
    self.actor:doTakeoff(self.inven, self.item, self.object)
    self.on_use(self.inven, self.floor, self.item, self.object, true)
  elseif item.action == 'drop' then
    local function on_use_cb()
      self.on_use(self.inven, self.floor, self.item, self.object, false)
    end
    if self.object:getNumber() > 1 then
      local n = self.object:getNumber()
      local function nb_cb(qty)
	qty = util.bound(qty, 1, n)
	self.actor:doDrop(self.inven, self.item, on_use_cb, qty)
      end
      game:registerDialog(GetQuantity.new('Drop how many?', '1 to '..n, n, n, nb_cb))
    else
      self.actor:doDrop(self.inven, self.item, on_use_cb)
    end
  elseif item.action == 'pickup' then
    self.actor:pickupFloor(self.item, true)
    self.actor:useEnergy()
    self.on_use(self.inven, self.floor, self.item, self.object, false)
  elseif item.action == 'destroy' then
    -- cf. do_cmd_destroy() in src/cmd3.c
    local function destroy_cb(n, force)
      if not n or n <= 0 then return end
      local ref_name = self.object:getName{force_qty=n}
      if not force and self.object:hasConfig('ask_destroy') then
	local function check_cb(ok)
	  -- 'Yes' is on the 'No' button below.
	  if not ok then destroy_cb(n, true) end
	end
	Dialog:yesnoLongPopup('Destroy Item Check', ('Are you sure you want to destroy %s?'):format(ref_name), 600, check_cb, 'No', 'Yes', true)
	return
      end
      if self.object:isArtifact() then
	game.log('You cannot destroy %s.', ref_name)
	self.object:forAllStack(function(so)
	  so.pseudo_id = so:isCursed() and 'terrible' or 'special'
	end)
      else
	local xo = self.actor:getOneObject(self.inven, self.item, self.floor, {num=n, vocal_left=true})
	local name = xo:getName()
	if game:hasEntity(xo) then game:removeEntity(xo) end
	game.log('You destroy %s.', name)
      end
      self.on_use(self.inven, self.floor, self.item, self.object, false)
    end
    local function stage1_cb(n)
      if not n or n <= 0 then return end
      local function stage2_cb(ok) if ok then destroy_cb(n) end end
      Dialog:yesnoPopup('Destroy Item', ('Destroy %s?'):format(self.object:getName{force_qty=n}), stage2_cb)
    end
    if self.object:getNumber() > 1 then
      local n = self.object:getNumber()
      game:registerDialog(GetQuantity.new('Destroy how many?', '1 to '..n, n, n, destroy_cb))
    else
      stage1_cb(1)
    end
  elseif item.action == 'inscr-add' then
    local function cb(text)
      if text then self.object.inscription = text end
      self.on_use(self.inven, self.floor, self.item, self.object, false)
    end
    game:registerDialog(GetText.new('Inscription', 'Text', 1, 99, cb))
  elseif item.action == 'inscr-edit' then
    local function cb(text)
      if text then self.object.inscription = text end
      self.on_use(self.inven, self.floor, self.item, self.object, false)
    end
    local gt = GetText.new('Edit Inscription', 'New text', 1, 99, cb)
    gt.c_box:setText(self.object.inscription)
    game:registerDialog(gt)
  elseif item.action == 'inscr-remove' then
    self.object.inscription = nil
    self.on_use(self.inven, self.floor, self.item, self.object, false)
  elseif item.action == 'automat-quickrule' then
    local QuickAutomatizerRule = require 'mod.dialogs.QuickAutomatizerRule'
    local d = QuickAutomatizerRule.new(self.object)
    game:registerDialog(d)
    self.on_use(self.inven, self.floor, self.item, self.object, true)
  elseif item.action == 'automat-test' then
    local function rule_cb(rule_item)
      if not rule_item or not rule_item.name then return end
      local AutomatizerRules = require 'mod.dialogs.AutomatizerRules'
      local d = AutomatizerRules.new()
      game:registerDialog(d)
      d:selectRule(rule_item.name, self.object)
    end
    local Automatizer = require 'mod.class.Automatizer'
    local list = {}
    for _, rule in ipairs(Automatizer.atm_rules) do
      list[#list+1] = { name=rule.name }
    end
    ui_util.listPopup('Test rule', 'Select rule to test:', list, 500, rule_cb)
    self.on_use(self.inven, self.floor, self.item, self.object, true)
  elseif item.action == 'configure' then
    local ConfigureItemDialog = require 'mod.dialogs.ConfigureItemDialog'
    local d = ConfigureItemDialog.new(self.object)
    game:registerDialog(d)
    self.on_use(self.inven, self.floor, self.item, self.object, true)
  elseif item.action == 'debug-inspect' then
    local DebugConsole = require"engine.DebugConsole"
    local d = DebugConsole.new()
    game:registerDialog(d)
    DebugConsole.line = "=__uids["..self.object.uid.."]"
    DebugConsole.line_pos = #DebugConsole.line
    d.changed = true
  elseif item.action == 'debug-full-id' then
    self.object:identify('full', self.actor)
    self.on_use(self.inven, self.floor, self.item, self.object, false)
  elseif item.talent_id then
    local t = self.actor:getTalentFromId(item.talent_id)
    if t.use_item.close_before then
      self.on_use(self.inven, self.floor, self.item, self.object, true)
    end
    if self.actor:preUseTalent(t) then
      -- Wrap this in a coroutine in case use_item.action() uses targeting.
      local co = coroutine.create(function()
	local ret = t.use_item.action(self.actor, t, self.object, self.inven, self.item, self.floor)
	if self.actor:postUseTalent(t, ret) then
	  self.actor:startTalentCooldown(t)
	end
      end)
      local ok, err = coroutine.resume(co)
      if not ok and err then
	print(debug.traceback(co))
	error(err)
      end
    end
    if not t.use_item.close_before then
      self.on_use(self.inven, self.floor, self.item, self.object, t.use_item.close_after)
    end
  elseif item.callback then
    item.callback(self.actor, self.object, self.inven, self.item, self.floor, self.on_use)
  end
end

function _M:generateList()
  local list = {}
  local is_worn = self.inven and self.inven ~= self.actor.INVEN_INVEN
  local is_carried = self.inven and self.inven == self.actor.INVEN_INVEN

  if self.object:canUseObject() then
    local titles = self.object:useTitles()
    list[#list+1] = { name=titles[1], action='use_item' }
  end
  if self.object.food then
    local title = self.object.food.bottle and 'Drink' or 'Eat'
    list[#list+1] = { name=title, action='eat' }
  end
  if self.object:check('can_refuel', 'oil') and not is_worn then
    -- Make sure we're wielding a refuelable light before offering to
    -- refill it.
    local inv = self.actor:getInven('LITE')
    if inv and #inv > 0 and inv[1]:refueled_by('oil') then
      list[#list+1] = { name='Fill lantern', action='fill' }
    end
  end
  if self.object:check('can_refuel', 'torch') and not is_worn then
    -- As above.
    local inv = self.actor:getInven('LITE')
    if inv and #inv > 0 and inv[1]:refueled_by('torch') then
      list[#list+1] = { name='Combine with current torch', action='retorch' }
    end
  end
  if spell.item_spells(self.object, self.actor, true) then
    list[#list+1] = { name='Browse spells', action='browse' }
  end
  if (is_carried or self.floor) and self.object.slot then
    list[#list+1] = { name='Wear/wield', action='wear' }
  end
  if is_worn then
    list[#list+1] = { name='Take off', action='takeoff' }
  end
  if is_carried or is_worn then
    list[#list+1] = { name='Drop', action='drop' }
  end
  if self.floor then
    list[#list+1] = { name='Pick up', action='pickup' }
  end
  -- Allow the 'destroy' option for unknown artifacts; it'll fail but
  -- pseudo-ID it as {special} or {terrible}.  Disallow it for cursed
  -- CURSE_NO_DROP items completely. [cf. do_cmd_destroy() in src/cmd3.c]
  if not (self.object:hasFlag('CURSE_NO_DROP') and self.object:isCursed()) and not (self.object:isIDd() and self.object:isArtifact()) then
    list[#list+1] = { name='Destroy', action='destroy' }
  end
  if is_worn or is_carried then
    if self.object.inscription then
      list[#list+1] = { name='Edit inscription', action='inscr-edit' }
      list[#list+1] = { name='Remove inscription', action='inscr-remove' }
    else
      list[#list+1] = { name='Add inscription', action='inscr-add' }
    end
  end
  list[#list+1] = { name='Configure...', action='configure' }
  if not self.object.unique and not is_worn then
    list[#list+1] = { name='Create destroy/pick-up rule', action='automat-quickrule', color=colors.simple(colors.YELLOW) }
    list[#list+1] = { name='Test automatizer rule', action='automat-test', color=colors.simple(colors.YELLOW) }
  end
  if config.settings.cheat then
    list[#list+1] = { name='Lua inspect', action='debug-inspect', color=colors.simple(colors.LIGHT_BLUE) }
    list[#list+1] = { name='*Identify*', action='debug-full-id', color=colors.simple(colors.LIGHT_BLUE) }
  end
  local inven = self.actor:getInven(self.inven)
  local where = self.floor and 'floor' or inven.worn and 'equip' or 'inven'
  for t_id, _ in pairs(self.actor.talents) do
    if self.actor:knowTalent(t_id) then
      local t = self.actor:getTalentFromId(t_id)
      if t.use_item then
	local allow_from = t.use_item.allow_from or { inven=true, equip=true, floor=true }
	if allow_from[where] and t.use_item.filter(self.object) and self.actor:preUseTalent(t, true, true) then
	  list[#list+1] = {
	    name = util.getval(t.use_item.name, self.actor, t, self.object, self.inven, self.item, self.floor) or t.name,
	    talent_id = t_id,
	  }
	end
      end
    end
  end
  if self.object.on_use_item_dialog then
    self.object:on_use_item_dialog(self.actor, self.inven, self.item, self.floor, list)
  end

  self:triggerHook {
    'UseItemMenu:generate',
    actor = self.actor,
    object = self.object,
    inven = self.inven,
    item = self.item,
    floor = self.floor,
    menu = list,
  }

  self.max_w = 0
  self.max_h = 0
  for i, v in ipairs(list) do
    local w, h = self.font:size(v.name)
    self.max_w = math.max(self.max_w, w)
    self.max_h = self.max_h + self.font_h
  end

  self.list = list
end
