-- $Id: spell-util.lua 4252 2021-07-31 23:22:57Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2021 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

-- Assorted spell-related utility functions, used in spell talents and
-- other places.

local Entity = require 'engine.Entity'
local Map = require 'mod.class.Map'
local DamageType = require 'engine.DamageType'
local Object = require 'mod.class.Object'
local Grid = require 'mod.class.Grid'

spell = {}

-- For use by various wand/staff-related logic.
-- NOTE:  Don't try to hold a reference to it, as it'll get clobbered in use.
spell.stick = {}

spell.use_device = 3	-- cf. USE_DEVICE in src/defines.h; used by rod_fail_check()

-- Aggravate monsters near the 'src' Actor.
function spell.aggravate_near(src)
  local any_woken, any_hasted = false, false
  for _, e in pairs(game.level.entities) do
    -- Skip non-monsters, dead monsters, the source and hte player.
    if e.ai_state and e.ai_state.sleep and not e.dead and e ~= src and e ~= game.player then
      -- Wake up anything within 2*MAX_SIGHT.
      if core.fov.distance(e.x, e.y, src.x, src.y) < 40 then
	e.ai_state.sleep = 0
	e.ai_target.actor = game:getPlayer()
	e.changed = true
	any_woken = true
      end
      -- Speed up monsters in LoS.
      if game.level.map.seens(e.x, e.y) and e.speed and e.base_speed then
	e.speed = math.max(e.speed, e.base_speed + 10)
	e:computeGlobalSpeed()
	-- TODO Potentially anger pets.
	e.changed = true
	any_hasted = true
      end
    end
  end
  if any_hasted then
    game.log('You feel a sudden stirring nearby!')
  elseif any_woken then
    game.log('You hear a sudden stirring in the distance!')
  end
end

-- Try to multiply the provided monster.
function spell.try_breed_monster(e, charm, clone)
  if game.player:hasEffect('EFF_STERILIZE') then
    game.logSeen(e, '%s tries to breed but fails!', e.name:capitalize())
    return false
  end

  local x, y = util.findFreeGrid(e.x, e.y, 1, true, {[Map.ACTOR]=true, [Map.TRAP]=true})
  if not x or not y then return false end

  -- TODO Implement this [cf. multiply_monster() in monster2.c]
  return false
end

-- Returns false if level or zone flags prevent teleportation.
function spell.level_ok_teleport()
  return not ((game.level and game.level.flags and game.level.flags.NO_TELEPORT) or
	      (game.zone and game.zone.flags and game.zone.flags.NO_TELEPORT))
end

-- Teleports the caster.  Checks for level no-teleport flag for player, and
-- tries to enforce a minimum distance.
function spell.teleport(caster, r, force)
  if not force and caster.player and not spell.level_ok_teleport() then
    game.logPlayer(caster, "The level's magic interferes with your teleportation.")
    return
  end
  -- Player with Anchor can prevent monsters from teleporting.
  if caster:has('ANCHOR') or game.player:has('ANCHOR') then
    game.logSeen(caster, [[The space-time continuum can't be disrupted.]])
    return
  end
  if game.zone.worldmap then return end
  -- Timed Roots prevents teleport.
  if caster:hasEffect(caster.EFF_TREE_ROOTS) then return end
  if caster:has('NO_TELE') and not force then
    game.logSeen(game.player, 'A mysterious force prevents %s from teleporting!', caster.player and 'you' or caster:getName{def_art=true, check_seen=true})
    return
  end
  local r_min = caster.player and r/4 or r/2
  caster:teleportRandom(caster.x, caster.y, r, r_min)
end

-- Teleports the player randomly up or down a level.
-- [cf. teleport_player_level() in src/spells1.c]
function spell.player_teleport_level(force)
  if util.is_quest_level() then
    game.logPlayer(game.player, 'There is no effect.')
    return
  end
  if not force and not spell.level_ok_teleport() then
    game.logPlayer(game.player, 'No teleport on special levels...')
    return
  end
  if game.zone.flags and game.zone.flags.NO_EASY_MOVE then
    game.logPlayer(game.player, 'Some powerful force prevents you from teleporting.')
    return
  end
  if game.player:has('ANCHOR') then
    game.logPlayer(game.player, [[The space-time continuum can't be disrupted.]])
    return
  end
  -- TODO not on death-fate levels.
  if game.player:has('TREE_ROOTS') then
    game.logPlayer(game.player, 'Your roots prevent you from teleporting.')
    return
  end
  if game.player:has('NO_TELE') then
    game.logPlayer(game.player, 'A mysterious force prevents you from teleporting!')
    return
  end
  if game.zone.town or game.zone.wilderness or game.zone.worldmap or game.zone.max_level == 1 then
    game.logPlayer(game.player, 'Nothing happens.')
    return
  end
  local up = rng.percent(50)
  if game.level.level == 1 then
    up = false
  elseif game.level.level >= game.zone.max_level or util.is_quest_level(nil, true) then
    up = true
  end
  game.logPlayer(game.player, up and 'You rise up through the ceiling.' or 'You sink through the floor.')
  game:changeLevel(game.level.level + (up and -1 or 1))
end

-- Teleports a targeted opponent.
-- [cf. teleport_away() in src/spells1.c]
function spell.teleport_away(caster, target, r)
  if game.player:has('ANCHOR') then
    game.log("The space-time continuum can't be disrupted.")
    return false
  end
  return target:teleportRandom(target.x, target.y, r, r/2)
end

-- Teleport the caster, loosely weighted in the specified direction.
-- Primarily used for Deathmold movement.
-- [cf. teleport_player_directed() in src/spells1.c]
function spell.teleport_directed(caster, r, dir)
  local dx, dy = util.dirToCoord(dir, caster.x, caster.y)

  if dx == 0 and dy == 0 then
    spell.teport(caster, r)
    return
  end

  -- Timed Roots prevents teleport.
  if caster:hasEffect(caster.EFF_TREE_ROOTS) then return end

  local map = game.level.map
  local x, y = caster.x, caster.y
  local dis, min, done = r, math.floor(r/4), false
  local y_neg = dy < 0 and -1 or 1
  local x_neg = dx < 0 and -1 or 1
  while not done do
    if dis > 200 then
      -- Emergency fallback if we exceed max range.
      spell.teport(caster, r)
      return
    end
    -- Try several times to get a valid location.
    for _ = 1, 500 do
      repeat
	if dy == 0 then
	  y = rng.spread(caster.y + y_neg*math.floor(dis/2), math.floor(dis/2))
	else
	  y = rng.spread(caster.y, dis/3)
	end
	if dx == 0 then
	  x = rng.spread(caster.x + x_neg*math.floor(dis/2), math.floor(dis/2))
	else
	  x = rng.spread(caster.x, dis/3)
	end
	local d = core.fov.distance(caster.x, caster.y, x, y)
      until d >= min and d <= dis
      -- If this is a valid location, we're done.
      if map:isBound(x, y) and map:checkEntity(x, y, Map.TERRAIN, 'empty_bold') then
	done = true
	break
      end
    end
    -- If we didn't find a suitable location, expand our range and try
    -- again.
    dis, min = dis*2, math.floor(min/2)
  end

  -- If we get here, we've found a suitable location.
  caster:move(x, y, true)
end

-- Attempt to move the provided actor in the provided cardinal direction
-- through consecutive walls to the next non-wall space.  If 'dx' and 'dy'
-- are nil, util.select_adjacent() will be used to select the desired
-- direction (in which case the method must be called in a coroutine
-- context).  'quiet' will suppress various log messages if true.  Returns
-- true if movement actually happens.  Note that it is the caller's
-- responsibility to :useEnergy() as needed.
-- [cf. passwall() in src/spells2.c]
function spell.passwall(who, dx, dy, safe, quiet)
  local Map = require 'mod.class.Map'
  local map = game.level.map
  local function ok_cb(x, y)
    return map:checkEntity(x, y, Map.TERRAIN, 'wall')
  end

  if not dx and not dy then
    local x, y = util.select_adjacent(who, ok_cb, 'You are not near any walls.', 'That is not a wall.', true)
    if not x or not y then return false end
    dx, dy = x - who.x, y - who.y
  end
  if dx == 0 and dy == 0 then return false end
  if not ok_cb(who.x + dx, who.y + dy) then
    if not quiet then game.logPlayer(who, 'That is not a wall.') end
    return false
  end
  if not quiet then
    game.logPlayer(who, 'A passage opens, and you step through.')
  end

  local ok = false
  local x, y, lx, ly = who.x, who.y, who.x, who.y
  while stop do
    x, y = x + dx, y + dy
    if not map:isBound(x, y) or map:checkEntity(x, y, Map.TERRAIN, 'permanent') then
      -- Always stop at map edge or before permanent walls.
      x, y = x - dx, y - dy
      ok = false
      break
    end
    if not map(x, y, Map.ACTOR) and not map.attrs(x, y, 'vault') and not map.attrs(x, y, 'no_teleport') then
      -- This locaation could be used as an emergency stop if necessary.
      lx, ly = x, y
      if not map:checkEntity(x, y, Map.TERRAIN, 'wall') then
	-- Out of the wall; we can stop now.
	ok = true
	break
      end
    end
  end

  if not ok then
    -- Exited badly; fall back to last emergency stop location, and
    -- potentially take damage.
    x, y = lx, ly
    if not safe then
      if not quiet then
	game.logPlayer(who, 'You emerge in the wall!')
      end
      who:takeHit(rng.dice(10, 8), map(x, y, Map.TERRAIN), 'becoming one with a wall')
    end
    -- TODO Use zone floor types.
    map(x, y, Map.TERRAIN, game.zone.grid_list.FLOOR)
  end

  who:move(x, y, true)
  who.changed = true
  return true
end

-- Creates a pair of linked jumpgates at the specified x/y coordinates.
-- Caller must do any needed sanity checking to be sure the locations are
-- suitable for jumpgating.
function spell.create_jumpgates(x1, y1, x2, y2)
  local g = game.zone.grid_list.VOID_JUMPGATE
  game.level.map(x1, y1, Map.TERRAIN, g)
  game.level.map(x2, y2, Map.TERRAIN, g)
  game.level.map.attrs(x1, y1, 'pair_gate', { x=x2, y=y2 })
  game.level.map.attrs(x2, y2, 'pair_gate', { x=x1, y=y1 })
end

-- Creates a pair of jumpgates between caster location and a targeted
-- location.  Returns true if targeting was successful, false if targeting
-- failed (placing the destination jumpgate at a random place near the
-- caster), or nil if targeting was canceled.
-- Must be called in a coroutine context, since it uses targeting.
-- Loosely duplicates the ACT_DIM_DOOR activation in do_cmd_activate() in
-- src/cmd6.c
function spell.target_jumpgates(caster, range, fail_rad)
  if not spell.level_ok_teleport() then
    game.logPlayer(caster, 'Not on special levels.')
    return nil
  end
  game.logPlayer(caster, 'You open a Void Jumpgate.  Choose a destination.')
  local tg = {
    type = 'hit',
    default_target = self, 
    range = range,
    nowarning = true,
    nolock = true,
    pass_terrain = true,
    requires_knowledge = false,
  }
  local x, y = self:getTarget(tg)
  if not x or not y then return nil end
  local map, ok = game.level.map, true
  if not map:checkEntity(x, y, Map.TERRAIN, 'empty_bold') or map.attrs(x, y, 'vault') or map.attrs(x, y, 'no_teleport') or core.fov.distance(caster.x, caster.y, x, y) > range or (caster.level and rng.chance(math.floor((caster.level*caster.level)/2))) then
    ok = false
    game.logPlayer(caster, 'You fail to exit the void correctly!')
    local grids = spell.grids_near(caster.x, caster.y, fail_rad, nil, nil, true)
    if #grids > 0 then
      local gxy = rng.table(grids)
      x, y = gxy.x, gxy.y
    else
      x, y = nil, nil
    end
  end
  if x and y then
    spell.create_jumpgates(caster.x, caster.y, x, y)
  end
  return ok
end

-- Check spell resistance of target versus caster spell.
function spell.resists(caster, target)
  if target == game.player then return rng.percent(target.saving_throw) end
  return rng.range(1, 10 + math.max(1, caster.level - 10)) < target.level
end

-- Returns true if the specified damage type is in the specified "category"
-- of grid-changing damage types, as defined by its floor_dam{} sub-table.
function spell.grid_damage(damtype, grp, dam)
  local dt = require('engine.DamageType'):get(damtype)
  if not dt.floor_dam then return false end
  if type(dt.floor_dam[grp]) == 'number' then
    return rng.percent(dt.floor_dam[grp])
  elseif type(dt.floor_dam[grp]) == 'function' then
    return dt.floor_dam[grp](dt, grp, dam)
  else
    return dt.floor_dam[grp]
  end
end

-- Extracts the numerical damage amount out of a possibly table-valued
-- damage parameter.
function spell.dam_num(dam)
  if type(dam) == 'table' then return dam[1] else return dam end
end

-- Returns a random grid type or nil based on provided percentage rules.
function spell.grid_prob(rules)
  if not rules then return nil end
  local p = rng.range(1, 100)
  for grid_type, prob in pairs(rules) do
    if p <= prob then return game.zone.grid_list[grid_type] end
    p = p - prob
  end
  return nil
end

-- Returns a random grid type or nil [as per spell.grid_prob()] to change a
-- floor grid to based on the provided damage type.  Also used directly as
-- the change_for_dmgtype() method of some grid types, hence the signature.
function spell.floor_damage(grid, x, y, damtype, dam)
  local dt = require('engine.DamageType'):get(damtype)
  local rules = dt.floor_dam and dt.floor_dam.floor
  if type(rules) == 'function' then return rules(grid, x, y, damtype, dam) end
  if type(rules) == 'table' then return spell.grid_prob(rules) end
  return nil
end

-- Returns a grid type suitable for the floor of the current zone.
function spell.floor_grid()
  local gen = spell.map_generator or game.zone:getGenerator('map', game.level, game.level.data.generator.map)
  if not gen.data or not gen.data.floor then
    return game.zone.grid_list.FLOOR
  end
  if not gen.data.floor.weighted then
    local idx = gen.data.floor[1]
    -- Slight Hack(TM):  PlasmaWilderness may put multiple grids here; pick
    -- one at random.
    if #gen.data.floor > 1 then idx = rng.table(gen.data.floor) end
    return game.zone.grid_list[idx]
  end
  if not gen.updateWeightCache then return game.zone.grid_list.FLOOR end
  gen:updateWeightCache('floor', game.level.level)
  return game.zone.grid_list[rng.table(gen.data.floor.wgt_cache)]
end

-- Returns a grid type suitable for the walls of the current zone.
function spell.wall_grid()
  if spell.use_wall_grid then
    if type(spell.use_wall_grid) ~= 'string' then
      return spell.use_wall_grid
    end
    if game.zone.grid_list[spell.use_wall_grid] then
      return game.zone.grid_list[spell.use_wall_grid]
    end
    if spell.map_generator then
      local gen = spell.map_generator
      local g = gen:resolve(spell.use_wall_grid, nil, false, gen.cur_lev)
      if g then return g end
    end
    print(("[WALL_GRID] unable to resolve '%s' in spell.wall_grid()"):format(spell.use_wall_grid))
    return game.zone.grid_list.WALL
  end
  local gen = spell.map_generator or game.zone:getGenerator('map', game.level, game.level.data.generator.map)
  if not gen.data or not gen.data.fill then
    return game.zone.grid_list.WALL
  end
  if not gen.data.fill.weighted then
    local idx = gen.data.fill[1]
    -- Slight Hack(TM):  PlasmaWilderness may put multiple grids here; pick
    -- one at random.
    if #gen.data.fill > 1 then idx = rng.table(gen.data.fill) end
    return game.zone.grid_list[idx]
  end
  if not gen.updateWeightCache then return game.zone.grid_list.WALL end
  gen:updateWeightCache('fill', game.level.level)
  return game.zone.grid_list[rng.table(gen.data.fill.wgt_cache)]
end

-- Returns true if the player's skill level(s) are sufficient to cast the
-- spell associated with the provide talent ID.
function spell.spell_skill_level_ok(who, tid)
  local t = who:getTalentFromId(tid)
  if not t or not t.skill or not t.skill_level then return false end

  local s_tags = type(t.skill) == 'table' and t.skill or { t.skill }
  local accum = 0
  local god = who.descriptor and who.descriptor.god
  for _, tag in ipairs(s_tags) do
    local sk = who.skills_def[tag]
    local lvl = who:getSkillLevel(tag, nil, true)
    -- Substitute Sorcery skill if applicable.
    if sk.sorcery_access then
      lvl = math.max(lvl, who:getSkillLevel('SORCERY', nil, true))
    end
    -- Substitute Prayer skill if appropriate.
    local god_conv = god and sk.prayer_access and sk.prayer_access[god]
    if god_conv then
      local god_lev = who:getSkillLevel('PRAYER', nil, true)
      god_lev = god_lev * god_conv[1] / god_conv[2]
      lvl = math.max(lvl, god_lev)
    end
    if lvl <= 0 then return false end
    accum = accum + lvl
  end
  return math.floor(accum/#s_tags) >= t.skill_level
end

-- Computes the effective spell level of the provided spell talent from the
-- provided actor's skill levels.
-- [cf. lib/core/s_aux.lua:get_level_school()]
function spell.get_spell_level(who, t, max, min)
  -- Slight Hack(TM):  Slightly different computation for spells cast from
  -- wands or staffs.
  if spell.stick.source then
    return spell.get_device_spell_level(who, t, max, min, spell.stick.source) 
  end

  max, min = max or 50, min or 1
  if type(t) == 'string' then t = who:getTalentFromId(t) end
  if not t or not t.skill or not t.skill_level then return min end

  local Skills = require 'mod.class.interface.PlayerSkills'

  local s_tags = type(t.skill) == 'table' and t.skill or { t.skill }
  local accum, use_spell_power, bonus = 0, true, 0
  local god = who.descriptor and who.descriptor.god
  for _, tag in ipairs(s_tags) do
    local sk = who.skills_def[tag]
    local lvl = who:getSkillLevel(tag, nil, true)
    -- Substitute Sorcery skill if applicable.
    if sk.sorcery_access then
      lvl = math.max(lvl, who:getSkillLevel('SORCERY', nil, true))
    end
    -- Substitute Prayer skill if appropriate.
    local god_conv = god and sk.prayer_access and sk.prayer_access[god]
    if god_conv then
      local god_lev = who:getSkillLevel('PRAYER', nil, true)
      god_lev = god_lev * god_conv[1] / god_conv[2]
      lvl = math.max(lvl, god_lev)
    end
    -- Apply spell bonus levels from skill definition.  Mainly this is just
    -- Udun's caster-level-based bonus.
    if sk.spell_bonus then bonus = bonus + sk.spell_bonus(who) end
    -- Only use Spell-power skill if all school skills are so flagged.
    use_spell_power = use_spell_power and sk.use_spell_power
    if lvl <= 0 then return min end
    accum = accum + lvl
  end
  accum = math.floor(accum/#s_tags)
  if use_spell_power then
    bonus = bonus + who:getSkillLevel('SPELL_POWER', 20)
  end
  -- Spellpower bonuses from equipment.
  bonus = bonus + (who.spellpower or 0)
  local tmp = accum - t.skill_level + 1
  if tmp >= 1 then tmp = tmp + bonus end
  tmp = math.floor((tmp * max * Skills.skill_step)/Skills.skill_max)
  return math.max(tmp, min)
end

-- Computes the effective spell level of the provided spell talent from the
-- provided actor's Magic-Device skill level.
-- [cf. lib/core/s_aux.lua:get_level_device()]
function spell.get_device_spell_level(who, t, max, min, stick)
  max, min = max or 50, min or 1
  if type(t) == 'string' then t = who:getTalentFromId(t) end
  if not t or not t.skill or not t.skill_level then return min end

  local Skills = require 'mod.class.interface.PlayerSkills'
  local lvl = who:getSkillLevel('MAGIC_DEVICE', nil, true) + stick.use_power.base_level
  lvl = math.min(lvl, stick.use_power.max_level + t.skill_level - 1)

  local tmp = lvl - t.skill_level + 1
  tmp = math.floor((tmp * max * Skills.skill_step)/Skills.skill_max)
  return math.max(tmp, min)
end

-- Computes the cost in mana or piety (detected automatically) to cast the
-- provided spell.  Designed to be used as the .mana() or .piety() method
-- of a spell talent definition.
function spell.mana_cost(who, t)
  -- Vaguely Dodgy Hack(TM):  NPCs don't have mana, and don't use it to
  -- cast their talents.
  if who ~= game.player then return 0 end

  local min, max = t.min_mana, t.max_mana
  if t.min_piety and t.max_piety then min, max = t.min_piety, t.max_piety end
  return min + spell.get_spell_level(who, t, max - min, 0)
end

-- Computes the probability of failure of the provided talent.  Requires
-- the 'base_fail' talent field; also uses fields 'stat', 'skill' and
-- 'skill_level' if present.
-- [cf. src/lua_bind.c:lua_spell_chance()]
-- [cf. assorted talent-using functions in src/cmd7.c]
-- [cf. Actor:preUseTalent(), Actor:getTalentFullDescription()]
function spell.fail_prob(who, t)
  -- Slight Hack(TM):  Slightly different computation for spells cast from
  -- wands or staffs.
  if spell.stick.source then
    return spell.device_fail_prob(who, spell.stick.source) 
  end
  -- Somewhat Bigger Hack(TM):  Completely different computation for NPCs
  -- [cf. make_attack_spell() in src/melee2.c]
  if who ~= game.player then
    -- "Stupid" monsters like jellies never fail.
    if who:has('STUPID') then return 0 end
    -- "Innate" talents never fail.
    if t.innate then return 0 end
    -- Otherwise, decrease with level.
    return math.max(0, 25 - math.floor(((who.level or 1) + 3)/4))
  end

  if not t.base_fail then return 0 end
  local prob = t.base_fail
  if t.is_spell then
    -- Reduce fail rate by caster level
    prob = prob - 3*(who.level - t.skill_level)
  elseif t.skill and t.skill_level then
    -- Reduce fail rate by skill adjustment.
    prob = prob - 3*(who:getSkillLevel(t.skill) - t.skill_level)
  end
  if t.stat then
    -- Reduce fail rate by stat adjustment.
    prob = prob - 3*(who:getStatRelated('fail_reduce', t.stat) - 1)
    end
  if t.mana then
    -- Increase fail rate for insufficient mana.
    local mana = util.getval(t.mana, who, t)
    if mana > who.mana then prob = prob + 5*(mana - who.mana) end
  end
  -- Eru followers get a spellcasting penalty for edged weapons.
  if t.is_spell and who:isFollowing('Eru') and who.icky_weapon then
    prob = prob + 25
  end
  -- Minimum failure rate...
  local minfail = 0
  if t.stat then
    -- ...based on stat, if specified...
    minfail = math.max(minfail, who:getStatRelated('min_fail', t.stat))
  end
  if t.is_spell and not who:has('PERFECT_CASTING') then
    -- ...and, for spells, lack of Perfect Casting.
    min_fail = math.max(minfail, 5)
  end
  prob = math.max(prob, minfail)
  -- Stun makes spells harder to cast.
  local eff = who:hasEffect(who.EFF_STUN)
  if eff and t.is_spell then prob = prob + (eff.dur > 50 and 25 or 15) end
  -- Always a 5% chance of working.
  prob = math.min(prob, 95)

  return math.max(prob, 0)
end

-- Computes the probability of failure of using the provided wand or staff.
-- [cf. src/lua_bind.c:lua_spell_device_chance()]
function spell.device_fail_prob(who, stick)
  local t = spell.stick_talent(stick)
  if not t or not t.base_fail then return 0 end

  local prob = t.base_fail
  local level = spell.get_device_spell_level(who, t, 50, 1, stick)
  -- Reduce failure rate for effective level.
  prob = prob - level + 1
  -- Minimum failure rate.
  local minfail = math.max(0, who:getSkillLevel('MAGIC_DEVICE', 25))
  prob = math.max(prob, minfail)
  -- Stun makes wands/staffs harder to use.
  local eff = who:hasEffect(who.EFF_STUN)
  if eff and t.is_spell then prob = prob + (eff.dur > 50 and 25 or 15) end
  -- Check the EASY_USE flag.
  if stick:hasFlag('flags', 'EASY_USE') then prob = math.floor(prob/3) end
  -- Always a 5% chance of working.
  prob = math.min(prob, 95)

  return math.max(prob, 0)
end

-- Returns true if user failed to use the provided rod.  If 'for_display'
-- is set, returns an equivalent percent fail chance descriptive string.
-- cf. do_cmd_zap_rod() in src/cmd6.c
function spell.rod_fail_check(rod, who, for_display)
  local prob = who.skill_device or 0
  -- Confusion makes this harder.
  if who:has('CONFUSED') then prob = math.floor(prob/2) end
  -- Higher level rod tips are harder to use.
  prob = math.max(1, prob - math.min(rod.rod_tip.level, 50))
  -- Rods of Simplicity help.
  if rod:hasFlag('flags', 'EASY_USE') then prob = prob*10 end
  -- Give everyone a (slight) chance.
  if prob < spell.use_device then
    if for_display then
      -- A 1 in D-prob+1 chance of having a 1 in D chance of success, where
      -- D=spell.use_device.
      local D = spell.use_device
      local pct = 100 - 100/(D * (D - prob + 1))
      return ('%.02f%%'):format(pct)
    elseif rng.chance(spell.use_device - prob + 1) then
      prob = spell.use_device
    end
  end
  if for_display then
    -- A prob-D+1 in prob chance of success, where D=spell.use.device
    local pct = 100*(spell.use_device - 1)/prob
    return (pct < 5 and '%.02f%%' or '%d%%'):format(pct)
  else
    return prob < spell.use_device or rng.range(1, prob) < spell.use_device
  end
end

-- The use() method for rods.
function spell.use_rod(rod, who, inven, floor, item)
  if who:has('ANTIMAGIC_FIELD') then
    local this_dev = rod:getNumber() > 1 and 'these devices' or 'this device'
    game.logPlayer(who, 'Your antimagic field prevents you from using %s.', this_dev)
    return {}
  end

  if spell.rod_fail_check(rod, who) then
    game.logPlayer(who, 'You failed to use the rod properly.')
    return { used=true, no_power=true }
  end
  -- Not enough mana should already have been checked by the time we get
  -- here.
  return rod.use_power.use_rod(rod, who, inven, floor, item)
end

-- The use() method for wands and staffs.
function spell.cast_stick_spell(stick, who, inven, floor, item)
  if who:has('ANTIMAGIC_FIELD') then
    local this_dev = stick:getNumber() > 1 and 'these devices' or 'this device'
    game.logPlayer(who, 'Your antimagic field prevents you from using %s.', this_dev)
    return {}
  end

  -- Hack(TM):  Don't allow use of a stack of wands from the floor.  Don't
  -- ask me, it's what the T2 code does.
  if floor and stick:getNumber() > 1 then
    game.logPlayer(who, 'You must pick up the %ss first.', stick.subtype)
    return {}
  end
  if rng.percent(spell.device_fail_prob(who, stick)) then
    game.logPlayer(who, 'You failed to use the %s properly.', stick.subtype)
    return { used=true, no_power=true }
  end
  -- The T2 code checks for no charges here, after using energy, but
  -- Object:useObject() checks before using energy.  Small favor to the
  -- player... ;)

  local t = spell.stick_talent(stick)
  if not t or not t.base_fail then return {} end
  spell.stick = { source=stick }
  local ret = t.action(who, t)
  if spell.stick.id then
    stick:learnFlavor()
  elseif not stick:isIDd() then
    stick:forAllStack(function(so)
      if not so:isFlavorKnown() then so.pseudo_id_tried_stick = true end
    end)
  end
  spell.stick = {}
  return { used=ret }
end

-- Returns the talent definition for the spell on the provided wand/staff.
function spell.stick_talent(stick)
  if not stick.use_power or not stick.use_power.spell then return nil end
  local Talents = require 'engine.interface.ActorTalents'
  return Talents.talents_def[stick.use_power.spell]
end

-- Constructs and returns a string suitable for a "Stuff will increase with
-- your <skill> skill" postfix of a spell talent description.  Will factor
-- in Sorcery and Prayer skills as appropriate for the talent.
function spell.spell_skill_desc(who, t)
  local Skills = require 'mod.class.interface.PlayerSkills'
  local SK = Skills.skills_def
  local sk = {}
  local g = {}
  local or_sorcery = false
  for i, id in ipairs(type(t.skill) == 'table' and t.skill or { t.skill }) do
    sk[i] = SK[id] and SK[id].name or '???('..id..')'
    or_sorcery = SK[id] and SK[id].sorcery_access or or_sorcery
    for god, _ in pairs(SK[id] and SK[id].prayer_access or {}) do
      g[god] = true
    end
  end
  or_sorcery = or_sorcery and who.skills.SORCERY.mod > 0
  local gg = {}
  for god, _ in pairs(g) do table.insert(gg, god) end
  table.sort(gg)

  -- Slight Hack(TM):  If we're describing a spell on a wand or staff, we
  -- want to use Magic-Device instead.
  if spell.stick.source then sk = { 'Magic-Device' } end
  
  local desc = table.concatNice(sk, ', ', ' and ')
  if or_sorcery then
    table.insert(sk, 'Sorcery')
    desc = table.concatNice(sk, ', ', ' or ')
  end
  desc = desc .. ' skill' .. (#sk > 1 and 's' or '')
  if #gg > 0 then
    gg = table.concatNice(gg, ', ', ' or ')
    desc = desc .. ' (or Prayer if you are following ' .. gg .. ')'
  end
  return ' with your ' .. desc .. '.'
end

-- Returns a faction friendly to the provided actor's faction.  Usually
-- this will be the same faction as the actor's, but for players it will
-- return 'pets' and for actors without a faction it will return 'enemies'.
function spell.friendly_faction(src)
  if not src then return 'enemies' end
  return (src.faction == 'players' and 'pets') or src.faction or 'enemies'
end

-- Attempts to summon the specified number (defaulting to 1) of monsters of
-- the specified level (defaulting to the max level of the current zone)
-- matching the specified filter near the specified target.  filter may
-- contain field 'kin' to match monsters with the same display character
-- as the caster.  If filter is a string, Zone:makeEntityByName() will be
-- used internally.  Returns a list of all monsters summoned, or nil if
-- none.
function spell.summon(caster, tx, ty, filter, num, force_level)
  num = util.getval(num or 1, caster)

  local by_name
  if type(filter) == 'string' then filter = { name=filter } end

  filter = filter and table.clone(filter) or {}
  filter.max_ood = filter.max_ood or 2
  if filter.kin then
    -- Vaguely Dodgy Hack(TM):  Summon Kin summons monsters with the same
    -- display character as the caster.
    filter.not_properties = { 'unique' }
    filter.display = caster.display
  end

  local function ok_summon(xy)
    return not game.level.map:checkEntity(xy[1], xy[2], Map.TERRAIN, 'no_summon')
  end
  local ret = {}
  for _ = 1, num do
    local x, y = util.find_free_grid(tx, ty, 10, true, {[Map.ACTOR]=true}, ok_summon)
    if x and y then
      local gen = game.zone:getGenerator('actor', game.level, game.level.data.generator.actor)
      local m = gen:generateOne { x=x, y=y, filter=filter, ret=true }
      if m then
	m.summoner = caster
	m.summon_time = filter.lastfor
	m.faction = spell.friendly_faction(caster)
	-- Hack(TM):  NPCs that are originally friendly or neutral to the
	-- player cannot be made less so.
	-- [cf.  -- src/monster.c:place_monster_one(), lines 2426-2434
	if m:has('PET') and m:reactionToward(game.player) <= 0 then
	  m.faction = 'friendlies'
	end
	if m:has('NEUTRAL') and m:reactionTo(game.player) < 0 then
	  m.faction = 'neutrals'
	end
	if force_level then
	  m.silent_levelup = true
	  m:forceLevelup(force_level)
	  m.silent_levelup = nil
	end
	table.insert(ret, m)
      end
    end
  end
  return #ret > 0 and ret or nil
end

-- Prompt the user for an item to enchant matching the provided filter and
-- attempt to enchant it as specified.  'which' can contain fields 'to_h',
-- 'to_d', 'to_a' and 'pval', with their values used as the number of times
-- to attempt enchantment.  Must be called in a coroutine context.  Returns
-- true if an item was selected.
-- [cf. obj_util.enchant(), enchant_spell() in src/spells2.c]
function spell.enchant_item(who, title, filter, which)
  if not which or not next(which) then return false end

  local ShowEquipInvenFloor = require 'mod.dialogs.ShowEquipInvenFloor'
  local function action_cb(o, inven, floor, item)
    who:talentDialogReturn({o=o, inven=inven, floor=floor, item=item})
    return true
  end
  local ret = who:talentDialog(ShowEquipInvenFloor.new(title, who, filter, action_cb))
  if not ret then return false end

  game.log('%s %s glow%s brightly!', (ret.floor and 'The' or 'Your'), ret.o:getName{no_count=true, no_mods=true, mo_pval=true, no_inscr=true}, (ret.o:getNumber() > 1 and '' or 's'))

  local any = false
  for f, n in pairs(which) do
    any = obj_util.enchant(ret.o, n, {[f]=true}) or any
  end

  if not any then game.log('The enchantment failed.') end
  return true
end

-- Attempts to disenchant the provided object.  If no object is specified,
-- a random equipment slot will be selected and a random object in that
-- slot of the target's inventory (if any) is disenchanted.
-- [cf. apply_disenchant() in src/spells1.c]
function spell.disenchant(o, target)
  if not o then
    local cands = { 'MAIN_HAND', 'LAUNCHER', 'BODY', 'CLOAK', 'OFF_HAND', 'HELM', 'GLOVES', 'BOOTS' }
    local inven = target:getInven(rng.table(cands))
    if #inven == 0 then return end
    o = rng.table(inven)
  end

  local t_names = target and not target.player and target:getName{def_art=true, check_seen=true}:capitalize().."'s" or 'Your'
  local o_name = o:getName{no_count=true}
  local plural = o:getNumber() ~= 1

  -- If there's nothing to disenchant, don't bother.
  if (not o.to_h or o.to_h <= 0) and
     (not o.to_d or o.to_d <= 0) and
     (not o.to_a or o.to_a <= 0) then
    return
  end

  -- Artifacts have a 71% chance to resist.
  if o.unique and rng.percent(71) then
    game.logSeen(target, '%s %s resist%s disenchantment!', t_names, o_name, plural and '' or 's')
    return
  end

  -- Disenchant to-hit.
  if o.to_h and o.to_h > 0 then
    o.to_h = o.to_h - 1
    if o.to_h > 5 and rng.percent(20) then o.to_h = o.to_h - 1 end
  end
  -- Disenchant to-dam.
  if o.to_d and o.to_d > 0 then
    o.to_d = o.to_d - 1
    if o.to_d > 5 and rng.percent(20) then o.to_d = o.to_d - 1 end
  end
  -- Disenchant armor bonus.
  if o.to_a and o.to_a > 0 then
    o.to_a = o.to_a - 1
    if o.to_a > 5 and rng.percent(20) then o.to_a = o.to_a - 1 end
  end

  game.logSeen(target, '%s %s %s disenchanted!', t_names, o_name, plural and 'were' or 'was')
  -- This will affect unbelievers eventually, so we need to recalc.
  if target.recalcEverything then target:recalcEverything() end
end

-- Lights all grids in the room containing the specified x,y coordinates.
function spell.lite_room(x, y)
  local map = game.level.map
  local function ok_cb(x, y)
    if not map.attrs(x, y, 'room') then return false end
    local g = map(x, y, Map.TERRAIN)
    if not g:floor_bold(x, y) then return nil end
    return true
  end
  local grids = util.flood_grids(x, y, ok_cb)

  -- cf. cave_temp_room_lite() in src/spells2.c
  -- First light all the grids.
  for _, xy in ipairs(grids) do
    map.lites(xy.x, xy.y, true)
  end
  -- Then wake up any monsters in lit grids, with probability 100% for
  -- SMART monsters, 10% for STUPID monsters, and 25% for other monsters.
  for _, xy in ipairs(grids) do
    local a = map(xy.x, xy.y, Map.ACTOR)
    if a and not a.dead and a ~= src and a.ai_state and a.ai_state.sleep then
      local prob = a.flags.SMART and 100 or a.flags.STUPID and 10 or 25
      if rng.percent(prob) and a.ai_state.sleep > 0 then
	a.ai_state.sleep = 0
	a.changed = true
	game.logSeen(a, '%s wakes up.', a:getName{def_art=true, check_seen=true}:capitalize())
      end
    end
  end
end

-- Unlights all grids in the room containing the specified x,y coordinates.
function spell.unlite_room(x, y)
  local map = game.level.map
  local function ok_cb(x, y)
    if not map.attrs(x, y, 'room') then return false end
    local g = map(x, y, Map.TERRAIN)
    if not g:floor_bold(x, y) then return nil end
    return true
  end
  local grids = util.flood_grids(x, y, ok_cb)

  -- cf. cave_temp_room_dark() in src/spells2.c
  -- Unlight all the grids, and forget floor grids.
  for _, xy in ipairs(grids) do
    map.lites(xy.x, xy.y, false)
    local g = map(xy.x, xy.y, Map.TERRAIN)
    if g:plain_floor_bold() then
      map.remembers(xy.x, xy.y, false)
      -- Slight Hack(TM):  Also forget any objects on the grid.
      map:objectsApply(xy.x, xy.y, function(o, _, _, _) o:markForMap(false) end)
    end
  end
end

-- Flush the FOVs of every monster in radius 20 of the target x,y,
-- including the player.  This should be used by effects that add or remove
-- walls that would change FOVs.
function spell.clobber_FOVs(x, y)
  local function cb(_, x, y)
    local a = game.level.map(x, y, Map.ACTOR)
    if a then a.fov_computed = false end
  end
  core.fov.calc_circle(x, y, game.level.map.w, game.level.map.h, 20, function() return false end, cb, nil)
  -- Also recompute FOV for the player.
  game.player:playerFOV()
end

function spell.player_forget_map()
  local map = game.level.map
  map:rememberAll(0, 0, map.w, map.h, false)
  -- Forget objects and traps on floor.
  for x = 0, map.w - 1 do
    for y = 0, map.h - 1 do
      map:checkAllEntities(x, y, 'forget', game.player)
    end
  end
end

function spell.player_amnesia()
  game.player:inventoryApplyAll(function(_, _, o)
    -- Forget non-*ID*'d inventory and equipment.
    -- This will clear pseudo-ID en passant for us.
    if not o:isIDd(true) then o:identify(false, game.player) end
  end)
  game.player:recalcEverything()
  game.player:sortInven()
  spell.player_forget_map()
end

-- A callback, designed to be passed as the 'apply' argument to
-- ActorFOV:computeFOV() [q.v.], that marks grids as remembered according
-- to T2's magic-mapping logic. [cf. map_area() in src/cave.c]
function spell.map_area_callback(x, y)
  local map = game.level.map
  local remember = false
  if not map:checkEntity(x, y, Map.TERRAIN, 'is_map_wall') then
    -- If the grid isn't a wall, remember it if it's not a plain floor.
    if not map:checkEntity(x, y, Map.TERRAIN, 'plain_floor_bold') then
      remember = true
    end
  else
    -- If the grid is a wall, remember it if it is adjacent to any non-wall
    -- grids.
    for xx = x-1, x+1 do
      for yy = y-1, y+1 do
	if map:isBound(xx, yy) and (xx ~= x or yy ~= y) and not map:checkEntity(xx, yy, Map.TERRAIN, 'is_map_wall') then
	  remember = true
	  break
	end
      end
      if remember then break end
    end
  end
  if remember then map.remembers(x, y, true) end
end

-- Returns an unsorted list of grids within in the specified radius (and
-- optionally outside the specified minimum radius) of the specified map
-- coordinates that match the provided match function (defaulting to
-- Grid:naked_bold()).  If 'aggressive' is true and no suitable grid is
-- found with the specified radii, the function will try again with double
-- the radius (up to 200) and half the min radius (down to 0).
-- This roughly mimics get_pos_player() in src/spells1.c.
function spell.grids_near(x, y, radius, min_radius, ok_cb, aggressive)
  ok_cb = ok_cb or function(x, y)
    local map = game.level.map
    return map:checkEntity(x, y, Map.TERRAIN, 'naked_bold') and not map.attrs(x, y, 'vault') and not map.attrs(x, y, 'no_teleport')
  end

  local grids = {}
  local save_cb = function(_, tx, ty)
    if ok_cb(x, y) and (not min_radius or cor.fov.distance(x, y, tx, ty) >= min_radius) then
      grids[#grids+1] = { x=tx, y=ty }
    end
  end
  local block_cb = function(_, lx, ly)
    return not game.level.map:isBound(lx, ly)
  end

  core.fov.calc_circle(x, y, game.level.map.w, game.level.map.h, radius, block_cb, save_cb, nil)
  if #grids == 0 and aggressive and (radius < 200 or (min_radius and min_radius > 0)) then
    local r = math.min(radius*2, 200)
    local mr = min_radius and math.floor(min_radius/2)
    return spell.grids_near(x, y, r, mr, ok_cb, true)
  end
  return grids
end

-- Opens an inventory/equipment/floor dialog for the provided Actor to
-- select an item to identify.  'id_lvl' should be true for normal identify
-- or 'full' for full *identify*, as per Object:identify().  Returns true
-- if an object was selected to identify.
-- NOTE:  Requires a coroutine wrapper enviromment of the type provided by
-- ActorTalents:useTalent() or Player:playerUseItem().
function spell.identify_dialog(who, id_lvl, title)
  local function filter_cb(o) return not o:isIDd(id_lvl == 'full') end
  local function action_cb(o, _, _, _, _, _)
    if not o then
      who:talentDialogReturn(false)
      return true
    end
    o:identify(id_lvl, who)
    if id_lvl == 'full' then
      local Dialog = require 'engine.ui.Dialog'
      Dialog:simpleLongPopup(o:getName(), o:getDesc(), 600, nil, false, 0.8)
    else
      game.logPlayer(who, 'You identify:  ' .. o:getName())
    end
    who:sortInven()
    if who.recalcEverything then who:recalcEverything() end
    who:talentDialogReturn(true)
    return true
  end
  local ShowEquipInvenFloor = require 'mod.dialogs.ShowEquipInvenFloor'
  return who:talentDialog(ShowEquipInvenFloor.new(title, who, filter_cb, action_cb))
end

-- Uncurses all items in the actor's inventory and equipment.  If 'force'
-- is true, heavily cursed items are also uncursed.  Returns true if any
-- item is affected.
function spell.uncurse_inven(who, force)
  local any = false
  local function cb(_, _, o)
    any = o:setCursed(false, force) or any
  end
  who:inventoryApplyAll(cb)
  return any
end

-- Detect traps within radius r around Actor who.  Returns true if any
-- traps were detected.
-- cf. detect_traps() in src/spells2.c
function spell.detect_traps(r, who)
  -- First, do an FOV walk to actually setKnown() the traps.
  local any = false
  local function cb(x, y)
    -- Mark as trap-detected.
    game.level.map.attrs(x, y, 'det_trap', true)
    local t = game.level.map(x, y, Map.TRAP)
    if t then
      t:setKnown(who, true)
      game.level.map:updateMap(x, y)
      game.level.map.remembers(x, y, true)
      any = true
    end
  end
  who:computeFOV(r, 'block_sense', cb, true, true, true)
  -- Then apply a one-turn mod to highlight them on the screen.
  if who.setOneTurnFOV then
    who:setOneTurnFOV {
      r = r,
      block = 'block_sense',
      callback = function(x, y)
	local Map = require 'engine.Map'
	if game.level.map(x, y, Map.TRAP) then
	  game.level.map.seens(x, y, 0.6)
	end
      end
    }
  end
  -- Mention if we found any traps.
  if any then game.logPlayer(who, 'You sense the presence of traps!') end
  -- Set the DET_TRAPS effect if needed.
  if not who:hasEffect(who.EFF_DET_TRAPS) then
    who:setEffect(who.EFF_DET_TRAPS, 1, { silent_gain=true })
  end
  return any
end

-- Detect doors and stairs within radius r around Actor who.  Returns true
-- if anything was detected.
-- cf. detect_doors() and detect_stairs() in src/spells2.c; we combine the
-- two since they seem to always be called together.
function spell.detect_doors_stairs(r, who)
  local any_door = false
  local any_stair = false
  local function cb(x, y)
    local ok = false
    local g = game.level.map(x, y, Map.TERRAIN)
    if g.flags and g.flags.DOOR then
      if g.define_as == 'SECRET_DOOR' then
	-- Replace secret doors with normal closed doors.
	game.level.map(x, y, Map.TERRAIN, game.zone.grid_list.DOOR)
	game.level.map:updateMap(x, y)
      end
      ok = true
      any_door = true
    elseif g.display == '<' or g.display == '>' then
      ok = true
      any_stair = true
    end
    if ok then game.level.map.remembers(x, y, true) end
  end
  who:computeFOV(r, 'block_sense', cb, true, true, true)
  if any_door then
    game.logPlayer(who, 'You sense the presence of doors!')
  end
  if any_stair then
    game.logPlayer(who, 'You sense the presence of ways out of this area!')
  end
  return any_door or any_stair
end

-- Detect non-invisible monsters within radius r around Actor who.  Returns
-- true if any monsters were detected.
-- cf. detect_monsters_normal() in src/spells2.c
function spell.detect_monsters(r, who)
  -- First do an FOV walk to see if we're going to detect anything.
  local any = false
  local function cb(x, y)
    local a = game.level.map(x, y, Map.ACTOR)
    if a and (not a:has('INVISIBLE') or who:has('SEE_INVIS')) then
      any = true
    end
  end
  who:computeFOV(r, 'block_sense', cb, true, true, true)
  -- Then apply a one-turn mod to highlight them on the screen.
  if who.setOneTurnFOV then
    who:setOneTurnFOV {
      r = r,
      block = 'block_sense',
      callback = function(x, y)
	local Map = require 'engine.Map'
	local a = game.level.map(x, y, Map.ACTOR)
	local src = spell.one_turn_fov_actor
	if a and not a:hasEffect(a.EFF_DETECTED) and (not a:has('INVISIBLE') or src:has('SEE_INVIS')) then
	  a:setEffect(a.EFF_DETECTED, 1, { pass=1 })
	  game.level.map.seens(x, y, 0.6)
	end
      end
    }
  end
  -- Mention if we found any monsters.
  if any then game.logPlayer(who, 'You sense the presence of monsters!') end
  return any
end

-- Detect invisible monsters within radius r around Actor who.  Returns
-- true if any monsters were detected.
-- cf. detect_monsters_invis() in src/spells2.c
function spell.detect_invisible_monsters(r, who)
  -- First do an FOV walk to see if we're going to detect anything.
  local any = false
  local function cb(x, y)
    local a = game.level.map(x, y, Map.ACTOR)
    if a and a:has('INVISIBLE') then any = true end
  end
  who:computeFOV(r, 'block_sense', cb, true, true, true)
  -- Then apply a one-turn mod to highlight them on the screen.
  if who.setOneTurnFOV then
    who:setOneTurnFOV {
      r = r,
      block = 'block_sense',
      callback = function(x, y)
	local Map = require 'engine.Map'
	local a = game.level.map(x, y, Map.ACTOR)
	if a and a:has('INVISIBLE') then
	  -- Learn that the detected monster is invisible.
	  if a.learnFlags then a:learnFlags(true, 'flags', 'INVISIBLE') end
	  a:setEffect(a.EFF_DETECTED, 1, { pass=1 })
	  game.level.map.seens(x, y, 0.6)
	end
      end,
    }
  end
  -- Mention if we found any monsters.
  if any then
    game.logPlayer(who, 'You sense the presence of invisible creatures!')
  end
  return any
end

-- Detect monsters within radius r around Actor who that match the provided
-- fct() filter.  fct() will be called with two parameters:  the monster
-- whose detectability is being decided, and the Actor 'who' passed to this
-- function.  Returns true if any monsters were detected.
function spell.detect_monsters_by(r, who, fct)
  -- First do an FOV walk to see if we're going to detect anything.
  local any = false
  local function cb(x, y)
    local a = game.level.map(x, y, Map.ACTOR)
    if a and fct(a, who) then any = true end
  end
  who:computeFOV(r, 'block_sense', cb, true, true, true)
  -- Then apply a one-turn mod to highlight them on the screen.
  if who.setOneTurnFOV then
    who:setOneTurnFOV {
      r = r,
      block = 'block_sense',
      callback = function(x, y)
	local Map = require 'engine.Map'
	local a = game.level.map(x, y, Map.ACTOR)
	local src = spell.one_turn_fov_actor
	if a and fct(a, src) then
	  a:setEffect(a.EFF_DETECTED, 1, { pass=1 })
	  game.level.map.seens(x, y, 0.6)
	end
      end,
    }
  end
  -- Mention if we found any monsters.
  if any then game.logPlayer(who, 'You sense the presence of monsters!') end
  return any
end

-- Detect buried treasure in walls.  Returns true if anything was detected.
-- Loosely based on detect_treasure() in src/spells2.c
function spell.detect_buried_treasure(r, who)
  local any = false
  local hits = {
    MAGMA_VEIN_WITH_TREASURE	= true,
    QUARTZ_VEIN_WITH_TREASURE	= true,
    SANDWALL_WITH_TREASURE	= true,
  }
  local function cb(x, y)
    local g = game.level.map(x, y, Map.TERRAIN)
    if hits[g.define_as] then
      game.level.map.remembers(x, y, true)
      any = true
    end
  end
  who:computeFOV(r, 'block_sense', cb, true, true, true)
  if any then
    game.logPlayer(who, 'You sense the presence of buried treasure!')
  end
  return any
end

-- Detect money objects.  Returns true if anything was detected.
-- Loosely based on detect_objects_gold() in src/spells2.c
function spell.detect_gold(r, who)
  local any = false
  local function floor_cb(o, x, y, i)
    if o.type == 'money' then
      o:markForMap(true)
      any = true
    end
  end
  local function mimic_inv_cb(_, _, o)
    if o.type == 'money' then
      o:markForMap(true)
      any = true
    end
  end
  local function fov_cb(x, y)
    game.level.map:objectsApply(x, y, floor_cb)
    local a = game.level.map(x, y, Map.ACTOR)
    -- Slight Hack(TM):  Also mark gold in mimic inventories.
    if a and a:has('MIMIC') then
      a:inventoryApply(a:getInven('INVEN'), mimic_inv_cb)
    end
  end
  who:computeFOV(r, 'block_sense', fov_cb, true, true, true)
  if any then
    game.level.map:redisplay()
    game.logPlayer(who, 'You sense the presence of treasure!')
  end

  -- Slight Hack(TM):  Detect monsters that look like gold.
  local function mon_cb(a, src) return a.display == '$' end
  any = spell.detect_monsters_by(r, who, mon_cb) or any

  return any
end

-- Detect non-money objects.  Returns true if anything was detected.
-- Loosely based on detect_objects() in src/spells2.c
function spell.detect_objects(r, who)
  local any = false
  local function floor_cb(o, x, y, i)
    if o.type ~= 'money' then
      o:markForMap(true)
      any = true
    end
  end
  local function mimic_inv_cb(_, _, o)
    if o.type == 'money' then
      o:markForMap(true)
      any = true
    end
  end
  local function fov_cb(x, y)
    game.level.map:objectsApply(x, y, floor_cb)
    local a = game.level.map(x, y, Map.ACTOR)
    -- Slight Hack(TM):  Also mark objects in mimic inventories.
    if a and a:has('MIMIC') then
      a:inventoryApply(a:getInven('INVEN'), mimic_inv_cb)
    end
  end
  who:computeFOV(r, 'block_sense', fov_cb, true, true, true)
  if any then
    game.level.map:redisplay()
    game.logPlayer(who, 'You sense the presence of objects!')
  end

  -- Slight Hack(TM):  Detect monsters that look like objects.
  local function mon_cb(a, src)
    return string.find('!=?|', a.display, 1, true)
  end
  any = spell.detect_monsters_by(r, who, mon_cb) or any

  return any
end

-- Loosely based on earthquake() in src/spells2.c
function spell.earthquake(cx, cy, r)
  -- As a courtesy to the player, don't start allowing earthquakes
  -- immediately after we complate a princess quest, or they could clobber
  -- the quest reward.
  if util.is_quest_level(true, true) then return end

  -- Paranoia:  enforce maximum range
  r = math.min(r, 12)
  local map = game.level.map
  local hit = {}
  -- Safe grid for actor to dodge to.
  local function safe_cb(x, y)
    -- TODO Don't allow non-player on glyphs of warding.
    local key = x..','..y
    return map:checkEntity(x, y, Map.TERRAIN, 'empty_bold') and not hit[key]
  end

  local hit_plyr = nil
  for x = cx - r, cx + r do
    for y = cy - r, cy + r do
      if map:isBound(x, y) and core.fov.distance(cx, cy, x, y) <= r then
	map.attrs(x, y, 'room', false)
	map.attrs(x, y, 'vault', false)
	map.lites(x, y, false)
	map.remembers(x, y, false)
	-- 15% chance for a grid to be changed; epicenter is excluded.
	if (x ~= cx or y ~= cy) and rng.percent(15) then
	  local key = x..','..y
	  local a = map(x, y, Map.ACTOR)
	  hit[key] = { x=x, y=y, a=a }
	  -- Remember if we hit the player.
	  if a and a.player then
	    hit_plyr = hit[key]
	    hit_plyr.key = key
	  end
	end
      end
    end
  end

  -- If the provided actor have some suitable wall-passing property, it
  -- need not dodge at all.
  local function passable(a)
    if a:has('NEVER_MOVE') then return false end
    return a:has('PASS_WALL') or a:has('KILL_WALL') or a:has('WRAITH_FORM')
  end

  -- If the player was hit, dodge or do damage.
  if hit_plyr and not passable(game.player) then
    local p = game.player
    local grids = spell.grids_near(p.x, p.y, 1, nil, safe_cb)
    game.log(rng.table {
      'The cave ceiling collapses!',
      'The cave floor twists in an unnatural way!',
      'The cave quakesr!  You are pummeled with debris!',
    })
    local damage = 0
    if #grids == 0 then
      -- Uh-oh, nowhere to escape.
      game.logPlayer(p, 'You are severely crushed!')
      damage = 300
    else
      -- Take no or partial damage.
      local k = rng.range(1, 3)
      if k == 1 then
	game.logPlayer(p, 'You nimbly dodge the blast!')
	damage = 0
      elseif k == 2 then
	game.logPlayer(p, 'You are bashed by rubble!')
	damage = rng.dice(10, 4)
	p:setEffect(p.EFF_STUN, rng.range(1, 50), {merge='combine'})
      else
	game.logPlayer(p, 'You are crushed between the floor and ceiling!')
	damage = rng.dice(10, 4)
	p:setEffect(p.EFF_STUN, rng.range(1, 50), {merge='combine'})
      end
      -- Move the player.
      local xy = rng.table(grids)
      p:move(xy.x, xy.y, true)
    end
    -- We've handled this spot.
    hit[hit_plyr.key] = nil
    -- Whether we moved or not, don't put a wall on the player.
    local key = p.x .. ',' .. p.y
    hit[key] = nil
    -- Semi-wraiths take partial damage, on orders from DG
    if p:has('SEMI_WRAITH') then damage = math.floor(damage/4) end
    if damage > 0 then
      p:takeHit(damage, {name='an earthquake'})
    end
  end

  -- Check for other actors hit by the quake.
  for _, xy in pairs(hit) do
    if xy.a and not passable(xy.a) then
      local grids = spell.grids_near(a.x, a.y, 1, nil, safe_cb)
      game.logSeen(xy.a, '%s wails out in pain!', xy.a:getName{def_art=true, check_seen=true})
      local damage = #grids == 0 and 200 or rng.dice(4, 8)
      -- Good morning.  This is your wakeup call. >;)
      if xy.a.ai_state then xy.a.ai_state.sleep = 0 end
      -- Questionable Hack(TM):  Actors that would be killed by the damage
      -- are instead deleted.  This is why we don't want earthquakes on
      -- princess/lost-sword quest levels.
      -- [TODO Maybe we should try harder to keep such monsters?]
      if damage >= xy.a.life then
	game.level:removeEntity(xy.a)
	-- I think this is what we want to do...
	xy.a:removed()
      else
	xy.a:takeHit(damage, {name='an earthquake'})
	if #grids > 0 then
	  local safe = rng.table(grids)
	  xy.a:move(safe.x, safe.y, true)
	end
      end
    end
  end

  -- Change the affected grids.
  for _, xy in pairs(hit) do
    if map:checkEntity(xy.x, xy.y, Map.TERRAIN, 'valid_bold') then
      local floor = map:checkEntity(xy.x, xy.y, Map.TERRAIN, 'floor_bold')
      -- Destroy objects on floor.
      while map:getObjectTotal(xy.x, xy.y) > 0 do
	local o = map:getObject(xy.x, xy.y, 1)
	map:removeObject(xy.x, xy.y, 1)
	game.level:removeEntity(o)
	if o.removed then o:removed() end
      end
      local new_grid
      if floor then
	-- The T2 code doesn't check for zone wall types.
	-- TODO Maybe we should.
	local t = rng.range(1, 100)
	local def = t < 20 and 'WALL' or t < 70 and 'QUARTZ' or 'MAGMA'
	new_grid = game.zone.grid_list[def]
      else
	-- The T2 code doesn't check for zone floor types either, but it's
	-- easy for us, so we do.
	new_grid = spell.floor_grid()
      end
      if new_grid then
	game.level.map(xy.x, xy.y, Map.TERRAIN, new_grid)
	new_grid:check('addedToLevel', game.level, xy.x, xy.y)
	new_grid:check('on_added', game.level, xy.x, xy.y)
	map:updateMap(xy.x, xy.y)
      end
    end
  end

  spell.clobber_FOVs(cx, cy)
end

-- Destroys a map area, a la Scrolls of *Destruction*.
-- [cf. destroy_area() in src/spells2.c]
function spell.destroy_area(cx, cy, r, full, force_quest)
  if game.level.data.generator.map.flags.NO_GENO then
    game.log('Not on special levels')
    return false
  end
  if util.is_quest_level(true, true) and not force_quest then return end

  local map = game.level.map
  local hit_plyr = false
  local function block_cb(_, x, y) return not map:isBound(x, y) end
  local function mod_cb(_, x, y)
    -- Lose room/vault status.
    map:unattr(x, y, 'room')
    map:unattr(x, y, 'vault')
    -- Forget and unlight.
    map.seens(x, y, false)
    map.remembers(x, y, false)
    map.lites(x, y, false)
    local a = map(x, y, Map.ACTOR)
    if a and a.player then
      -- Don't change player grid, but remember to blind player later.
      hit_plyr = true
    elseif x == cx and y == cy then
      -- Don't hit the epicenter.
    else
      -- Delete monster instead of killing.
      -- TODO don't delete companions
      if a and not a.quest_mon then
	game.level:removeEntity(a)
	-- Put uniques back in circulation.
	a:removed()
      end
      if map:checkEntity(x, y, Map.TERRAIN, 'valid_bold') then
	-- Delete objects on the floor.
	local n = map:getObjectTotal(x, y)
	for i = n, 1, -1 do
	  local o = map:getObject(x, y, i)
	  game.level:removeEntity(o)
	  -- Put artifacts back in circulation.
	  o:removed()
	end
	-- Destroy the grid.
	local t = rng.range(1, 200)
	local f = t <= 20 and 'WALL' or
		  t <= 70 and 'QUARTZ_VEIN' or
		  t <= 100 and 'MAGMA_VEIN' or 'FLOOR'
	map(x, y, Map.TERRAIN, game.zone.grid_list[f])
      end
    end
  end

  core.fov.calc_circle(cx, cy, map.w, map.h, r, block_cb, mod_cb, nil)

  if hit_plyr then
    game.log('There is a searing blast of light!')
    local p = game.player
    if not p:resistDam(DamageType.BLIND) and not p:resistDam(DamageType.LITE) then
      p:setEffect(p.EFF_BLIND, 10 + rng.range(1, 10), {merge='combine'})
    end
  end
  spell.clobber_FOVs(cx, cy)
  map:scheduleRedisplay()
end

-- Genocide all monsters with the specified symbol (if any) and within the
-- specified range (if any) of the caster (if any).  The symbol may be
-- specified as "target" to prompt the player to target a monster whose
-- symbol to use; for this use, the function must be called in a suitable
-- coroutine context.  If 'dam_sides' is specified, does NdM damage to the
-- caster, where N is the number of monsters genocided and M is the value
-- of 'dam_sides'.  Returns true if any monsters were genocided.
function spell.genocide(caster, symbol, range, dam_sides)
  if game.level.data.generator.map.flags.NO_GENO then return false end
  -- TODO no genocide on death fate levels

  if symbol == 'target' then
    game.logPlayer(caster, 'Target a monster type to genocide')
    local tg = { type='hit', range=20, nowarning=true }
    local tx, ty, a = caster:getTarget(tg)
    if not a or not a.display then return false end
    symbol = a.display
  end

  -- Slight Hack(TM):  On the first pass, collect actors to remove.
  local to_remove = {}
  for uid, e in pairs(game.level.entities) do
    if e:getEntityKind() == 'actor' and not e.unique then
      local ok = true
      if symbol then ok = ok and e.display == symbol end
      if range then
	if not caster or not caster.x or not caster.y or not e.x or not e.y then
	  ok = false
	else
	  ok = ok and core.fov.distance(caster.x, caster.y, e.x, e.y) <= range
	end
      end
      if ok then table.insert(to_remove, e) end
    end
  end

  -- Now remove all the actors.  Since none of the removed actors are
  -- unique, I don't think we need to do any special :removed() handling.
  for _, e in ipairs(to_remove) do game.level:removeEntity(e) end

  -- Do damage if requested.
  if caster.takeHit and dam_sides and #to_remove > 0 then
    caster:takeHit(rng.dice(#to_remove, dam_sides), {name='backlash energy from the genocide'})
  end

  return #to_remove > 0
end

-- Returns a list of IDs of spell talents in the provided object's
-- worn_talents{} and carried_talents{} lists.  'a' is a reference actor
-- using the ActorTalents interface, used to obtain talent information.  If
-- 'exist' is true, the function will return true if the object has any
-- spell talents.
function spell.item_spells(o, a, exist)
  local ret = {}
  for _, field in ipairs({ 'worn_talents', 'carried_talents' }) do
    for _, tid in ipairs(o[field] or {}) do
      local t = a:getTalentFromId(tid)
      if t.is_spell then
	if exist then return true end
	ret[#ret+1] = tid
      end
    end
  end
  if exist then return false end
  return ret
end

-- Sets or cancels the RECALL temporary effect as appropriate.
function spell.set_recall(a, d, f)
  if game.level.flags and game.level.flags.NO_RECALL_OUT then
    game.log('You cannot recall from here.')
    return false
  end
  if game.zone.wilderness or game.zone.worldmap then
    game.log('Nothing happens.')
    return false
  end
  if a:hasEffect(a.EFF_TREE_ROOTS) then
    game.log('Your roots prevent the recall.')
    return false
  end
  -- TODO Check for reset recall depth.
  if a:hasEffect(a.EFF_RECALL) then
    a:removeEffect(a.EFF_RECALL)
  else
    local dest = game.state:getRecallDest(game.zone.town)
    if not dest then
      game.log('You have not been anywhere to recall to.')
      return false
    end
    local p = { to_town=not game.zone.town, dest=dest }
    a:setEffect(a.EFF_RECALL, rng.range(0, d-1) + f, p)
  end
  return true
end

-- Tries to move an object from grid tx,ty (of maximum weight wgt) to Actor
-- a's location.
function spell.fetch_object(a, tx, ty, wgt, require_los)
  local os = game.level.map(a.x, a.y, Map.OBJECT)
  if os then
    game.logSeen(a, a:convertTalentMessage(a, '@Source@ is standing on an object and cannot fetch another.'))
    return false
  end

  local o = game.level.map:getObject(tx, ty, 1)
  if not o then
    game.logSeenXY(tx, ty, 'There is no object to fetch there.')
    return false
  end
  if require_los and not a:hasLOS(tx, ty, nil, 18) then
    game.logSeen(a, a:convertTalentMessage(a, '@Source@ does not have line of sight to the object to fetch.'))
    return false
  end
  if o.weight > wgt then
    game.logPlayer(a, '%s is too heavy to move.', o:getName():capitalize())
    return false
  end
  local dropx, dropy = util.drop_loc_near(a.x, a.y, 5)
  if not dropx then
    game.logPlayer(a, 'No place nearby to drop object.')
    return false
  end

  -- Actually move the object.
  game.level.map:removeObject(tx, ty, 1)
  game.level.map:addObject(a.x, a.y, o)
  return true
end

-- cf. self_knowledge() in src/spells2.c
function spell.self_knowledge(who)
  local ret = {}

  if who.dead then
    -- TODO Describe killer and location.
  end

  -- TODO Possessed monster body flags

  for id, def in pairs(who.talents_def) do
    if who:knowTalent(id) and def.self_knowledge_line then
      ret[#ret+1] = def.self_knowledge_line
    end
  end

  local flags = {
    { 'BLOOD_OF_LIFE', 'The Blood of Life flows through your veins.' },
    { 'BLIND', 'You cannot see.' },
    { 'CONFUSED', 'You are confused.' },
    { 'AFRAID', 'You are terrified.' },
    { 'CUT', 'You are bleeding', },
    { 'STUN', 'You are stunned.' },
    { 'POISON', 'You are poisoned.' },
    { 'HALLUCINATION', 'You are hallucinating.' },
    { 'AGGRAVATE', 'You aggravate monsters.' },
    { 'TELEPORT', 'Your position is very uncertain.' },
    { 'BLESSED', 'You feel righteous.' },
    { 'HEROISM', 'You feel heroic.' },
    { 'SUPERHEROISM', 'You are in a battle rage.' },
    { 'PROTECT_EVIL', 'You are protected from evil.' },
    { 'SHIELD', 'You are protected by a mystic shield.' },
    { 'EVASION', 'You are dodging melee attacks.' },
    { 'INVULNERABILITY', 'You are temporarily invulnerable.' },
    { 'CONFUSE_MONSTER', 'Your hands are glowing dull red.' },
    { 'RECALL', 'You will soon be recalled.' },
    { 'infravision', 'Your eyes are sensitive to infrared light.' },
    { 'SEE_INVIS', 'You can see invisible creatures.' },
    { 'MAGIC_BREATH', 'You can breathe without air.' },
    { 'WATER_BREATH', 'You can breathe underwater.' },
    { 'FEATHER', 'You levitate just over the ground.' },
    { 'CLIMB', 'You can climb high mountains.' },
    { 'FREE_ACT', 'You have free action.' },
    { 'REGENERATE', 'You regenerate quickly.' },
    { 'SLOW_DIGEST', 'Your appetite is small.' },
    { 'AUTO_ID', 'You know everything.' },
    { 'HOLD_LIFE', 'You have a firm hold your life force.' },
    { 'REFLECT', 'You reflect arrows and bolts.' },
    { 'SH_FIRE', 'You are surrounded with a fiery aura.' },
    { 'SH_ELEC', 'You are surrounded with electricity.' },
    { 'ANTIMAGIC_FIELD', 'You are surrounded by an anti-magic field.' },
    { 'NO_MAGIC', 'You are surrounded by an anti-magic shell.' },
    { 'WRAITH_FORM', 'You are incorporeal.' },
    { 'NO_TELE', 'You cannot teleport.' },
    -- TODO permanent light
    { 'ANCHOR', 'The space-time continuum cannot be disrupted near you.' },
  }
  for _, f in ipairs(flags) do
    if who:has(f[1]) then ret[#ret+1] = f[2] end
  end

  if who.esp_flags then
    if who.esp_flags.ALL then
      ret[#ret+1] = 'You have ESP.'
    else
      local esp_order = {
	{ 'ORC', 'orcs' },
	{ 'TROLL', 'trolls' },
	{ 'DRAGON', 'dragons' },
	{ 'SPIDER', 'spiders' },
	{ 'GIANT', 'giants' },
	{ 'DEMON', 'demons' },
	{ 'UNDEAD', 'undead' },
	{ 'EVIL', 'evil beings' },
	{ 'ANIMAL', 'animals' },
	{ 'THUNDERLORD', 'thunderlords' },
	{ 'GOOD', 'good beings' },
	{ 'NONLIVING', 'non-living things' },
	{ 'UNIQUE', 'unique beings' },
      }
      for _, f in ipairs(esp_order) do
	if who.esp_flags[f[1]] then
	  ret[#ret+1] = ('You can sense the presence of %s.'):format(f[2])
	end
      end
    end
  end

  local luck = who:getLuckScale(-100, 100)
  local abs_luck = math.abs(luck)
  local lucky = luck > 0 and 'lucky' or 'unlucky'
  local luck_msgs = {
    { 90, 'You are incredibly %s.' },
    { 60, 'You are extremely %s', },
    { 30, 'You are very %s.' },
    {  0, 'You are %s.' },
    { -1, 'You have normal luck.' }
  }
  for _, f in ipairs(luck_msgs) do
    if abs_luck > f[1] then
      ret[#ret+1] = f[2]:format(luck)
      break
    end
  end

  local dams = {
    DamageType.ACID, DamageType.ELEC, DamageType.FIRE, DamageType.COLD,
    DamageType.POISON, DamageType.LITE, DamageType.DARK,
    DamageType.CONFUSION, DamageType.SOUND, DamageType.DISENCHANT,
    DamageType.CHAOS, DamageType.SHARDS, DamageType.NEXUS, DamageType.NETHER,
  }
  for _, dam in ipairs(dts) do
    local dt = DamageType:get(dam)
    if who:immuneDam(dam) then
      ret[#ret+1] = ('You are completely immune to %s.'):format(dt.name)
    elseif who:resistDam(dam, {count=true}) == 2 then
      ret[#ret+1] = ('You resist %s exceptionally well.'):format(dt.name)
    elseif who:resistDam(dam) then
      ret[#ret+1] = ('You are resistant to %s.'):format(dt.name)
    elseif who:vulnDam(dam) then
      ret[#ret+1] = ('You are vulnerable to %s.'):format(dt.name)
    end
  end
  -- A couple special ones.
  if who:resistDam(DamageType.FEAR) then
    ret[#ret+1] = 'You are completely fearless.'
  end
  if who:resistDam(DamageType.BLIND) then
    ret[#ret+1] = 'Your eyes are reistant to blindness.'
  end

  local Stats = require 'mod.class.interface.ActorStats'
  for _, s in ipairs(Stats.stats) do
    if who:isStatSustained(s) then
      ret[#ret+1] = ('Your %s is sustained.'):format(Stats.stat_names[s]:lower())
    end
  end

  if who:hasEffect(who.EFF_BLACK_BREATH) then
    ret[#ret+1] = 'You suffer from Black Breath.'
  end

  local eq_convey = {}
  for iid = 1, #who.inven_def do
    if who.inven[iid] and who.inven_def[iid].is_worn then
      for item, o in ipairs(who.inven[iid]) do
	-- We use actual flags here instead of obvious_flags{}, since we're
	-- presumably gaining magical knowledge.
	table.merge(eq_convey, o.stat_flags or {})
	table.merge(eq_convey, o.pval_flags or {})
      end
    end
  end
  local eq_flags = {
    { 'STR', Stats.stat_names('STR'):lower() },
    { 'INT', Stats.stat_names('INT'):lower() },
    { 'WIS', Stats.stat_names('WIS'):lower() },
    { 'DEX', Stats.stat_names('DEX'):lower() },
    { 'CON', Stats.stat_names('CON'):lower() },
    { 'CHA', Stats.stat_names('CHA'):lower() },
    { 'skill_stealth', 'stealth' },
    { 'skill_searching', 'searching ability' },
    { 'infravision', 'infravision' },
    { 'adj_digging', 'digging ability' },
    { 'speed', 'speed' },
    { 'extra_blows', 'attack speed' },
    { 'extra_crit', 'ability to score critical hits' },
  }
  for _, f in ipairs(eq_flags) do
    if eq_convey[f[1]] then
      ret[#ret+1] = ('Your %s is affected by your equipment.'):format(f[2])
    end
  end

  -- TODO Weapon flags
  local wpn_flags = {
    { 'convey_flags', 'BLESSED', 'has been blessed by the gods.' },
    { 'flags', 'CHAOTIC', 'is branded with the Sign of Chaos.' },
    { 'convey_flags', 'IMPACT', 'can cause earthquakes on impact.' },
    { 'flags', 'VORPAL', 'is very sharp.' },
    { 'flags', 'VAMPIRIC', 'drains life from your foes.' },
    { 'brand_dam', 'BRAND_ACID', 'melts your foes.' },
    { 'brand_dam', 'BRAND_ELEC', 'shocks your foes.' },
    { 'brand_dam', 'BRAND_FIRE', 'burns your foes.' },
    { 'brand_dam', 'BRAND_COLD', 'freezes your foes.' },
    { 'brand_dam', 'BRAND_POIS', 'poisons your foes.' },
    { 'slay_flags', 'SLAY_ANIMAL', 'strikes at animes with extra force.' },
    { 'slay_flags', 'SLAY_EVIL', 'strikes at evil with extra force.' },
    { 'slay_flags', 'SLAY_UNDEAD', 'strikes at undead with holy force.' },
    { 'slay_flags', 'SLAY_DEMON', 'strikes at demons with holy force.' },
    { 'slay_flags', 'SLAY_ORC', 'is especially deadly against orcs.' },
    { 'slay_flags', 'SLAY_TROLL', 'is especially deadly against trolls.' },
    { 'slay_flags', 'SLAY_GIANT', 'is especially deadly against giants.' },
    { 'slay_flags', 'SLAY_DRAGON', 'is especially deadly against dragons.' },
    { 'slay_flags', 'KILL_DRAGON', 'is a great bane of dragons.' },
    { 'slay_flags', 'KILL_DEMON', 'is a great bane of demons.' },
    { 'slay_flags', 'KILL_UNDEAD', 'is a great bane of undead.' },
  }
  local wpn_accum = {}
  local function wpn_cb(_, _, o)
    for _, ff in ipairs(wpn_flags) do
      if o:hasFlag(ff[1], ff[2]) then wpn_accum[ff[2]] = true end
    end
  end
  local inven = who:getInven('MAIN_HAND')
  who:inventoryApply(inven, wpn_cb)
  local wpn_pfx = #inven > 1 and 'One of your weapons ' or 'Your weapon '
  for _, ff in ipairs(wpn_flags) do
    if wpn_accum[ff[2]] then ret[#ret+1] = wpn_pfx .. ff[3] end
  end

  return ret
end

function spell.self_knowledge_dialog(who)
  local Dialog = require 'engine.ui.Dialog'
  local lines = spell.self_knowledge(who)
  lines = table.mapv(function(v) return {name=v} end, lines)
  Dialog:listPopup('Self-Knowledge', 'Your Attributes:', lines, 600, game.h*0.8)
end

-- "Grow" grids of the provided terrain type in a radius around the
-- provided actor.  The optional 'filter' parameter should be either a
-- list of flags [as per Grid:hasFlags()] that the grid must have to be
-- affected or a function that will receive the Grid object and the x/y
-- coordinates as parameters and should return true for grids that can be
-- affected.
-- [cf. grow_things() in src/spells2.c]
function spell.grow_terrain(who, typ, r, filter)
  local map = game.level.map
  local g = game.zone.grid_list[typ]
  local f
  if type(filter) == 'function' then
    f = filter
  elseif type(filter) == 'table' then
    f = function(g, _, _) return g:hasFlags(unpack(filter)) end
  end
  for _ = 1, r*r + 11, 1 do
    -- I assume this is intended to bias towards grids near the center.
    local dx = math.floor((rng.range(0, r*2)-r + rng.range(0, r*2)-r)/2)
    local dy = math.floor((rng.range(0, r*2)-r + rng.range(0, r*2)-r)/2)
    local tx, ty = who.x+dx, who.y + dy
    if map:isBound(tx, ty) and core.fov.distance(who.x, who.y, tx, ty) <= r and map:checkEntity(tx, ty, Map.TERRAIN, 'clean_bold') then
      local g = game.zone.grid_list[typ]
      if not f or f(g, tx, ty) then
	if g.force_clone then g = g:clone() end
	map(tx, ty, Map.TERRAIN, g)
      end
    end
  end
end

-- cf. curse_equipment() in src/melee2.c
function spell.curse_equipment(who, pct, pct_heavy)
  -- Slight Hack(TM):  The original T2 code randomly picks an equipment
  -- slot and punts if the slot is empty; we attempt to replicate this as a
  -- courtesy to the player, including the multiple "extra" slots used for
  -- various Mimicry and Possession handling.
  local slots = {
    MAIN_HAND	= 3,	-- 3 weapons
    LAUNCHER	= 1,	-- 1 bow
    RING	= 6,	-- 6 rings
    AMULET	= 2,	-- 2 amulets
    LITE	= 1,	-- 1 lite
    BODY	= 1,	-- 1 body
    CLOAK	= 1,	-- 1 cloak
    OFF_HAND	= 1,	-- 3 arms
    HELM	= 2,	-- 2 heads
    GLOVES	= 3,	-- 3 hands
    BOOTS	= 2,	-- 2 feets
    SYMBIOTE	= 1,	-- 1 carried monster
    QUIVER	= 1,	-- 1 quiver
    TOOL	= 1,	-- 1 tool
  }
  local pos, total, cands = 0, 0, {}
  for i_id, n in pairs(slots) do
    total = total + n
    local inven = who:getInven(i_id)
    if inven then for i = 1, #inven do cands[i + pos] = inven[i] end end
    pos = pos + n
  end
  local o = cands[rng.range(1, total)]
  -- If we got lucky and hit an empty slot, punt.
  if not o then return end
  -- Check chance of cursing.
  if not rng.percent(pct) then return end
  -- Objects with the BLESSED flag get an extra saving throw.
  if o:hasFlag('convey_flags', 'BLESSED') and rng.range(1, 888) > pct then
    game.logPlayer(who, 'Your %s resist%s cursing!', o:getName{no_count=true}, o:getNumber() == 1 and 's' or '')
    return
  end
  -- Artifacts and ego items have a chance to be heavily cursed.
  local cursed = true
  if (o.ego_names or o.artifact_name) and rng.percent(pct_heavy) then
    cursed = 'heavy'
  end
  -- Slight Hack(TM):  Don't allow this to reduce a heavily cursed item to
  -- just cursed.
  if o.cursed == 'heavy' then cursed = 'heavy' end
  if o:setCursed(cursed, true) then
    game.logPlayer(who, 'There is a malignant black aura surrounding you...')
  end
end

-- "Alter reality" by "re-entering" the current level.
-- (It feels like there should be more sanity checking here, but the T2
-- code doesn't do any...)
-- [cf. alter_reality in src/spells2.c]
function spell.alter_reality()
  game.log('The world changes!')
  game:changeLevel(game.level.level, game.zone)
end


-- Utility function used by the spell methods below.  Sets
-- Grid.project_damage and Object.project_damage to the specified
-- { grid=..., object=... } values in t{}, potentially modified by the
-- damage type definition and the dam{} table passed to the calling spell
-- method, as follows:
--
--   - A damage type definition can include a project_flags{} subtable
--     containing grid=true or grid=false to override the spell-type-based
--     defaults, and similarly for objects.
--
--   - The 'dam' parameter is forwarded from spell.bolt_spell() et al, and
--     may be a table containing grid=true or grid=false to override both
--     the spell-type-based defaults and the damage-type-specified
--     overrides, and similary for objects.
--
-- Returns a corresponding table containing the previous values of same,
-- which can be passed back to this method (sans dt or dam fields) to
-- restore the previous settings..
function spell.set_project(t, dt, dam)
  local ret = { grid=Grid.project_damage, object=Object.project_damage }
  local grid, object = (t and t.grid), (t and t.object)
  if dt then
    local DamageType = require 'engine.DamageType'
    local def = DamageType:get(dt)
    local p = def and def.project_flags or {}
    if type(p.grid) ~= 'nil' then grid = p.grid end
    if type(p.object) ~= 'nil' then object = p.object end
  end
  if type(dam) == 'table' then
    if type(dam.grid) ~= 'nil' then grid = dam.grid end
    if type(dam.object) ~= 'nil' then object = dam.object end
  end
  Grid.project_damage = grid
  Object.project_damage = object
  return ret
end

-- Utility function used by the spell methods below.  Returns values for
-- 'selffire' and 'friendlyfire' based on the initial values provided and
-- any values in the t{} table.  Either value may be true to always hit the
-- affected category (self or friend), false to always hit, or a number
-- from 0 to 100 as a percent chance to hit (thus 0 is equivalent to false
-- and 100 is equivalent to true).
local function get_selffire(t, selffire, friendlyfire)
  if type(t) ~= 'table' then return selffire, friendlyfire end
  if type(t.selffire) ~= 'nil' then selffire = t.selffire end
  if type(t.friendlyfire) ~= 'nil' then friendlyfire = t.friendlyfire end
  print(('[LIAR] selffire=%s friendlyfire=%s'):format(tostring(selffire), tostring(friendlyfire)))
  return selffire, friendlyfire
end

-- Fires a bolt spell from the specified source Actor using the specified
-- talent.
--   = 'range' can be omitted if the talent defines a suitable 'range'
--     field (cf. ActorTalents:getTalentRange()).
--   = 'dt' can be omitted if the talent defines a suitable 'damage_type'
--     field [in the same style as 'range' above].
--   = If x and y are omitted, the player will be prompted for a target (or
--     the NPC AI will choose one).
--   = 'dam' can be a table containing flags passed through to the default
--     projector (cf. notes on setDefaultProjector in data/damage_types.lua).
--   = If 'particle' is passed or the talent defines a suitable .particle()
--     method, it will be used to draw a particle for the spell.  The
--     method should take parameters (dt, src_x, src_y, dst_x, dst_y).
--     [cf. spell.simple_bolt_particle()]
--   = Self-fire and friendly-fire can be controlled via dam.selffire resp.
--     dam.friendlyfire; by default, selffire is false and friendlyfire is
--     nil.  [cf. get_selffire() above]
--   = By default, grids and floor objects will not be affected; this can
--     be overridden by the damage type and the 'dam' parameter as per
--     spell.set_project() [q.v.]
-- If the spell was attempted, returns a table containing the targeted x
-- and y coordinates; returns nil if the player canceled during targeting.
function spell.bolt_spell(src, t, dt, dam, x, y, range, particle)
  local selffire, friendlyfire = get_selffire(dam, false, nil)
  range = range or src:getTalentRange(t)
  particle = particle or (t and t.particle)
  if not dt then
    dt = t.damage_type
    if type(dt) == 'function' then dt = dt(src, t) end
    if not dt then return nil end
  end
  local tg = { type='bolt', range=range, talent=t, selffire=selffire, friendlyfire=friendlyfire }
  util.target_merge_damage_type(tg, dt)
  if not x or not y then x, y = src:getTarget(tg) end
  if not x or not y then return nil end
  local p = type(dam) == 'table' and dam or {}
  local proj = spell.set_project({ grid=false, object=false }, dt, dam)
  src:project(tg, x, y, dt, dam)
  if particle then particle(dt, src.x, src.y, x, y) end
  spell.set_project(proj)
  return { x=x, y=y }
end

-- Fires a ball spell from the specified source Actor using the specified
-- talent.
--   = 'range' can be omitted if the talent defines a suitable 'range'
--     field (cf. ActorTalents:getTalentRange()).
--   = 'r' can be omitted if the talent defines a suitable 'radius' field
--     (cf. ActorTalents:getTalentRadius()).
--   = 'dt' can be omitted if the talent defines a suitable 'damage_type'
--     field [in the same style as 'range' above].
--   = If x and y are omitted, the player will be prompted for a target (or
--     the NPC AI will choose one).
--   = 'dam' can be a table containing flags passed through to the default
--     projector (cf. notes on setDefaultProjector in data/damage_types.lua).
--   = If 'particle' is passed or the talent defines a suitable .particle()
--     method, it will be used to draw a particle for the spell.  The
--     method should take parameters (dt, src_x, src_y, r).
--     [cf. spell.simple_ball_particle()]
--   = Self-fire and friendly-fire can be controlled via dam.selffire resp.
--     dam.friendlyfire; by default, selffire is false and friendlyfire is
--     nil.  [cf. get_selffire() above]
--   = By default, grids and floor objects will be affected; this can be
--     overridden by the damage type and the 'dam' parameter as per
--     spell.set_project() [q.v.]
-- If the spell was attempted, returns a table containing the targeted x
-- and y coordinates; returns nil if the player canceled during targeting.
function spell.ball_spell(src, t, dt, dam, r, x, y, range, particle)
  local selffire, friendlyfire = get_selffire(dam, false, nil)
  r = r or src:getTalentRadius(t)
  range = range or src:getTalentRange(t)
  particle = particle or (t and t.particle)
  if not dt then
    dt = t.damage_type
    if type(dt) == 'function' then dt = dt(src, t) end
    if not dt then return nil end
  end
  local tg = { type='ball', range=range, radius=r, talent=t, selffire=selffire, friendlyfire=friendlyfire }
  util.target_merge_damage_type(tg, dt)
  -- Slight Hack(TM) for traps [cf. trap_util.apply_trap_spec()]
  if type(dam) == 'table' then
    tg.start_x, tg.start_y = dam.start_x, dam.start_y
  end
  if not x or not y then x, y = src:getTarget(tg) end
  if not x or not y then return nil end
  -- Slight Hack(TM):  Record the center of the ball spells as dam.cx and
  -- dam.cy, so that the projector can reduce damage by distance from the
  -- center.
  if type(dam) == 'number' then dam = { dam } end
  dam.cx, dam.cy = x, y
  local proj = spell.set_project({ grid=true, object=true }, dt, dam)
  src:project(tg, x, y, dt, dam)
  if particle then particle(dt, x, y, r) end
  spell.set_project(proj)
  return { x=x, y=y }
end

-- Fires a beam spell from the specified source Actor using the specified
-- talent.
--   = 'range' can be omitted if the talent defines a suitable 'range'
--     field (cf. ActorTalents:getTalentRange()).
--   = 'dt' can be omitted if the talent defines a suitable 'damage_type'
--     field [in the same style as 'range' above].
--   = If x and y are omitted, the player will be prompted for a target (or
--     the NPC AI will choose one).
--   = 'dam' can be a table containing flags passed through to the default
--     projector (cf. notes on setDefaultProjector in data/damage_types.lua).
--   = If 'particle' is passed or the talent defines a suitable .particle()
--     method, it will be used to draw a particle for the spell.  The
--     method should take parameters (dt, src_x, src_y, dst_x, dst_y).
--     [cf. spell.simple_beam_particle()]
--   = Self-fire and friendly-fire can be controlled via dam.selffire resp.
--     dam.friendlyfire; by default, selffire is false and friendlyfire is
--     nil.  [cf. get_selffire() above]
--   = By default, grids and floor objects will not be affected; this can
--     be overridden by the damage type and the 'dam' parameter as per
--     spell.set_project() [q.v.]
-- If the spell was attempted, returns a table containing the targeted x
-- and y coordinates; returns nil if the player canceled during targeting.
function spell.beam_spell(src, t, dt, dam, x, y, range, particle)
  local selffire, friendlyfire = get_selffire(dam, false, nil)
  range = range or src:getTalentRange(t)
  particle = particle or (t and t.particle)
  if not dt then
    dt = t.damage_type
    if type(dt) == 'function' then dt = dt(src, t) end
    if not dt then return nil end
  end
  local tg = { type='beam', range=range, talent=t, selffire=selffire, friendlyfire=friendlyfire }
  util.target_merge_damage_type(tg, dt)
  if not x or not y then x, y = src:getTarget(tg) end
  if not x or not y then return nil end
  local p = type(dam) == 'table' and dam or {}
  local proj = spell.set_project({ grid=false, object=false }, dt, dam)
  src:project(tg, x, y, dt, dam)
  if particle then particle(dt, src.x, src.y, x, y) end
  spell.set_project(proj)
  return { x=x, y=y }
end

-- Fires a bolt spell from the specified source Actor using the specified
-- talent, which will become a beam spell with the specified probability.
--   = 'range' can be omitted if the talent defines a suitable 'range'
--     field (cf. ActorTalents:getTalentRange()).
--   = 'dt' can be omitted if the talent defines a suitable 'damage_type'
--     field [in the same style as 'range' above].
--   = If x and y are omitted, the player will be prompted for a target (or
--     the NPC AI will choose one).
--   = If 'particle1' and 'particle2' are specified, they are used to
--     generate map particles for the beam resp. bolt versions of the
--     spell.  If only 'particle1' is specified, it is used for both.  The
--     talent may also define .bolt_particle() and .beam_particle()
--     methods, or a single .particle() method to use for both.  The
--     methods should take parameters (dt, src_x, src_y, dst_x, dst_y).
--     [cf. spell.simple_bolt_particle(), spell.simple_beam_particle()]
--   = 'dam' can be a table containing flags passed through to the default
--     projector (cf. notes on setDefaultProjector in data/damage_types.lua).
--   = Self-fire and friendly-fire can be controlled via dam.selffire resp.
--     dam.friendlyfire; by default, selffire is false and friendlyfire is
--     nil.  [cf. get_selffire() above]
--   = By default, grids and floor objects will not be affected; this can
--     be overridden by the damage type and the 'dam' parameter as per
--     spell.set_project() [q.v.]
-- If the spell was attempted, returns a table containing the targeted x
-- and y coordinates; returns nil if the player canceled during targeting.
function spell.bolt_or_beam_spell(src, t, prob, dt, dam, x, y, range, particle1, particle2)
  local selffire, friendlyfire = get_selffire(dam, false, nil)
  range = range or src:getTalentRange(t)
  particle1 = particle1 or (t and (t.bolt_particle or t.particle))
  particle2 = particle2 or (t and (t.beam_particle or t.particle)) or particle1
  if not dt then
    dt = t.damage_type
    if type(dt) == 'function' then dt = dt(src, t) end
    if not dt then return nil end
  end
  -- Slight Hack(TM):  We do user targeting in 'beam' mode first for user
  -- convenience, then switch it to 'bolt' when actually projecting the
  -- damage.  (That way, the player can't just keep retrying until they get
  -- a beam...)
  local tg = { type='beam', range=range, talent=t, selffire=selffire, friendlyfire=friendlyfire }
  util.target_merge_damage_type(tg, dt)
  if not x or not y then x, y = src:getTarget(tg) end
  if not x or not y then return nil end
  tg.type = rng.percent(prob) and 'beam' or 'bolt'
  local particle = tg.type == 'beam' and particle2 or particle1
  local p = type(dam) == 'table' and dam or {}
  local proj = spell.set_project({ grid=false, object=false }, dt, dam)
  src:project(tg, x, y, dt, dam)
  if particle then particle(dt, src.x, src.y, x, y) end
  spell.set_project(proj)
  return { x=x, y=y }
end

-- Fires a beam or ball spell in a "ray" in a specific cardinal direction
-- from the specified source Actor using the specified talent, continuing
-- until it hits something.  Beam spells will stop at the first wall they
-- hit; ball spells will stop at the first actor they hit.  Both will stop
-- at range, or just before hitting a non-hostile actor if friendly fire is
-- enabled (see below).
--   = 'range' can be omitted if the talent defines a suitable 'range'
--     field (cf. ActorTalents:getTalentRange()).
--   = 'r' specifies a ball spell with the specified radius if positive;
--     a zero or negative value specifies a beam spell.  Note that the
--     talent's 'radius' field will *not* be used.
--   = 'dt' can be omitted if the talent defines a suitable 'damage_type'
--     field [in the same style as 'range' above].
--   = 'dx' and 'dy' specify the "delta" x/y from the source's x/y; both
--     must be between -1 and 1, and one or both must be non-zero.
--   = The 'particle' parameter, if specified, will be passed through to
--     spell.beam_spell() or spell.ball_spell() [qq.v.] as appropriate, and
--     will be handled as described therein.
--   = 'dam' can be a table containing flags passed through to the default
--     projector (cf. notes on setDefaultProjector in data/damage_types.lua).
--   = Self-fire and friendly-fire can be controlled via dam.selffire resp.
--     dam.friendlyfire; by default, selffire is false and friendlyfire is
--     nil.  [cf. get_selffire() above]
--   = The default handling of grids and floor objects from
--     spell.beam_spell() or spell.ball_spell() [qq.v] will apply, and can
--     be overridden by the damage type and the 'dam' parameter as per
--     spell.set_project() [q.v.]
-- Returns true if the spell was attempted.
function spell.ray_spell(src, t, dt, dam, r, dx, dy, range, particle)
  -- Can't target self.
  if dx == 0 and dy == 0 then return false end
  -- Only target adjacent grids.
  if math.abs(dx) > 1 or math.abs(dy) > 1 then return false end
  local beam_mode = r <= 0

  local selffire, friendlyfire = get_selffire(dam, false, nil)
  range = range or src:getTalentRange(t)
  if not dt then
    dt = t.damage_type
    if type(dt) == 'function' then dt = dt(src, t) end
    if not dt then return nil end
  end
  local l = src:lineFOV(src.x + dx, src.y + dy)
  l:set_corner_block()
  local lx, ly, is_corner_blocked = l:step(true)
  local tx, ty = lx, ly
  local Map = require 'engine.Map'
  local map = game.level.map
  while lx and ly and not is_corner_blocked and core.fov.distance(src.x, src.y, lx, ly) <= range do
    local a = map(lx, ly, Map.ACTOR)
    -- Stop just before non-hostile actor if friendlyfire is true.
    if friendlyfire and a and src:reactionToward(a) >= 0 then break end
    -- Stop on actor in ball mode.
    if a and not beam_mode then
      tx, ty = lx, ly
      break
    end
    -- Stop at wall in beam mode, or just before wall in ball mode.
    if map:checkEntity(lx, ly, Map.TERRAIN, 'block_move') then
      if beam_mode then tx, ty = lx, ly end
      break
    end
    -- Otherwise, keep proceeding along the line.
    lx, ly = l:step(true)
  end
  -- Now that we have a target, fire a beam or ball at it.
  if ball_mode then
    return spell.ball_spell(src, t, dt, dam, r, tx, ty, range, particle)
  else
    return spell.beam_spell(src, t, dt, dam, tx, ty, range, particle)
  end
end

-- Fires beam or ball spells in "rays" in all eight cardinal directions, as
-- per spell.ray_spell() [q.v.].  Returns true if any of the rays fired.
function spell.ray_spells(src, t, dt, dam, r, range, particle)
  local any = false
  for dx = -1, 1 do
    for dy = -1, 1 do
      if dx ~= 0 or dy ~= 0 then
	any = spell.ray_spell(src, t, dt, dam, r, dx, dy, range) or any
      end
    end
  end
  return any
end

-- Fires a spell from the specified source Actor using the specified talent
-- that affects all monsters in line of sight.
--   = 'dt' can be omitted if the talent defines a suitable 'damage_type'
--     field [in the same style as 'range' above].
--   = If a 'filter' method is provided, it will be called with arguments
--     (tgt, src, t, dt, dam), where tgt is the potential target being
--     tested, and should return true if the spell should be applied to the
--     target.  If the talent defines a 'los_filter' function field, it
--     will be used if 'filter' is omitted.  If neither is specified, all
--     actors except ths source actor will be affected.
--   = 'dam' can be a table containing flags passed through to the default
--     projector (cf. notes on setDefaultProjector in data/damage_types.lua).
--   = Self-fire and friendly-fire can be controlled via dam.selffire resp.
--     dam.friendlyfire; by default, selffire is false and friendlyfire is
--     nil.  [cf. get_selffire() above]
--   = Grid and floor objects will not be affected, and this *cannot* be
--     be overridden in the usual manner via spell.set_project() [q.v.]
-- Returns two values:  true if the spell was cast, and true if any visible
-- monster was hit.
function spell.los_spell(src, t, dt, dam, filter)
  local selffire, friendlyfire = get_selffire(dam, false, nil)
  filter = filter or t.los_filter
  if not dt then
    dt = t.damage_type
    if type(dt) == 'function' then dt = dt(src, t) end
    if not dt then return nil, false end
  end
  local seen_hit = false
  local tg = { type='hit', talent=t }
  util.target_merge_damage_type(tg, dt)
  local proj = spell.set_project { grid=false, object=false }
  for _, tgt in ipairs(src.fov.actors_dist or {}) do
    local friendly = src.reactionToward and src:reactionToward(tgt) >= 0
    local hitme = tgt ~= src
    if filter then hitme = filter(tgt, src, t, dt, dam) end
    if friendly and
       ((type(friendlyfire) == 'number' and rng.percent(friendlyfire)) or
	(type(friendlyfire) ~= 'number' and friendlyfire))
    then
      hitme = false
    end
    if hitme then src:project(tg, tgt.x, tgt.y, dt, dam) end
    -- We assume LOS here.
    seen_hit = seen_hit or (hitme and src:canSee(tgt))
  end
  spell.set_project(proj)
  return true, seen_hit
end

-- Fires a 'hit' spell (like a bolt, but not stopped by intervening
-- monsters) from the specified source Actor using the specified talent.
--   = 'range' can be omitted if the talent defines a suitable 'range'
--     field (cf. ActorTalents:getTalentRange()).
--   = 'dt' can be omitted if the talent defines a suitable 'damage_type'
--     field [in the same style as 'range' above].
--   = If x and y are omitted, the player will be prompted for a target (or
--     the NPC AI will choose one).
--   = 'dam' can be a table containing flags passed through to the default
--     projector (cf. notes on setDefaultProjector in data/damage_types.lua).
--   = Self-fire and friendly-fire can be controlled via dam.selffire resp.
--     dam.friendlyfire; by default, selffire is false and friendlyfire is
--     nil.  [cf. get_selffire() above]
--   = By default, grids and floor objects will not be affected; this can
--     be overridden by the damage type and the 'dam' parameter as per
--     spell.set_project() [q.v.]
-- If the spell was attempted, returns a table containing the targeted x
-- and y coordinates; returns nil if the player canceled during targeting.
function spell.hit_spell(src, t, dt, dam, x, y, range)
  local selffire, friendlyfire = get_selffire(dam, false, nil)
  range = range or src:getTalentRange(t)
  if not dt then
    dt = t.damage_type
    if type(dt) == 'function' then dt = dt(src, t) end
    if not dt then return nil end
  end
  local tg = { type='hit', range=range, talent=t, selffire=selffire, friendlyfire=friendlyfire }
  util.target_merge_damage_type(tg, dt)
  if not x or not y then x, y = src:getTarget(tg) end
  if not x or not y then return nil end
  local p = type(dam) == 'table' and dam or {}
  local proj = spell.set_project({ grid=false, object=false }, dt, dam)
  src:project(tg, x, y, dt, dam)
  spell.set_project(proj)
  return { x=x, y=y }
end

-- Fires a "meteor swarm" spell from the specified source Actor using the
-- specified talent.
--   = The 'dt', 'dam' and 'particle' parameters are passed through to
--     spell.ball_spell(); q.v. for handling of those parameters.
--   = The default handling of grids and floor objects from
--     spell.ball_spell() will apply, and can be overridden by the damage
--     type and the 'dam' parameter as per spell.set_project() [q.v.]
-- Returns true if any meteors were fired (which should happen if there are
-- any open grids nearby).
function spell.meteor_spell(src, t, dt, dam, n_balls_base, particle)
  local n_balls = n_balls_base + rng.range(1, n_balls_base)

  -- Pre-compute the list of candidate x,y coordinates.
  local grids = core.fov.circle_grids(src.x, src.y, 5, true)
  local gg = {}
  for x, yy in pairs(grids) do
    for y, _ in pairs(yy) do
      gg[#gg+1] = { x=x, y=y }
    end
  end
  if #gg == 0 then return nil end

  for _ = 1, n_balls do
    local xy = rng.table(gg)
    -- spell.ball_spell() will modify a table dam{} parameter; clone it.
    local ddam = type(dam) == 'table' and table.clone(dam, true) or dam
    spell.ball_spell(src, t, dt, ddam, 2, xy.x, xy.y, nil, particle)
  end
  return true
end

-- Creates and returns an overlay Entity to be used with cloud effects for
-- the specified DamageType.
function spell.cloud_overlay(dt)
  local E = require 'engine.Entity'
  local def = require('engine.DamageType'):get(dt)
  local cc = def.cloud_color or def.color
  if not cc then
    return E.new { alpha=128, display='', color_br=200, color_bg=200, color_bb=200 }
  elseif #cc == 0 then
    return E.new { alpha=128, display='', color_br=cc.r, color_bg=cc.g, color_bb=cc.b }
  elseif #cc == 1 then
    return E.new { alpha=128, display='', color_br=cc[1].r, color_bg=cc[1].g, color_bb=cc[1].b }
  else
    local ShimmerOverlay = require 'mod.class.ShimmerOverlay'
    return ShimmerOverlay.new(cc)
  end
end

-- Fires a cloud spell from the specified source Actor using the specified
-- talent.
--   = 'range' can be omitted if the talent defines a suitable 'range'
--     field (cf. ActorTalents:getTalentRange()).
--   = 'r' can be omitted if the talent defines a suitable 'radius' field
--     (cf. ActorTalents:getTalentRadius()).
--   = 'dt' is the damage type of the initial ball; it can be omitted if
--     the talent defines a suitable 'damage_type' field [in the same style
--     as 'range' above].
--   = 'dt2' is the damage type of the cloud; if omitted, the damage type
--     of the initial ball will be used.
--   = 'dam2' is the cloud damage per turn; if omitted, the ball damage
--     will be used.
--   = If x and y are omitted, the player will be prompted for a target (or
--     the NPC AI will choose one).
--   = 'dam' and 'dam2' can be tables containing flags passed through to
--     the default projector (cf. notes on setDefaultProjector in
--     data/damage_types.lua).
--   = If 'particle' is passed or the talent defines a suitable .particle()
--     method, it will be used to draw a particle for the initial ball part
--     of the spell.  The method should take parameters (dt, src_x, src_y, r).
--     [cf. spell.simple_ball_particle()]
--   = Self-fire and friendly-fire can be controlled via dam.selffire resp.
--     dam.friendlyfire.  By default, selffire is false and friendlyfire is
--     nil.  [cf. get_selffire() above]
--   = By default, grids and floor objects will be affected by the initial
--     ball; this can be overridden by the damage type and the 'dam'
--     parameter as per spell.set_project() [q.v.]
-- If the spell was attempted, returns a table containing the targeted x
-- and y coordinates; returns nil if the player canceled during targeting.
function spell.cloud_spell(src, t, dt, dam, r, dur, dt2, dam2, x, y, range, particle)
  local selffire, friendlyfire = get_selffire(dam, false, nil)
  local p = type(dam) == 'table' and dam or {}
  r = r or src:getTalentRadius(t)
  range = range or src:getTalentRange(t)
  if not dt then
    dt = t.damage_type
    if type(dt) == 'function' then dt = dt(src, t) end
    if not dt then return nil end
  end
  dt2 = dt2 or dt
  dam2 = dam2 or type(dam) == 'table' and table.clone(dam, true) or dam
  local tg = { type='ball', range=range, radius=r, talent=t, selffire=selffire, friendlyfire=friendlyfire }
  util.target_merge_damage_type(tg, dt)
  if not x or not y then x, y = src:getTarget(tg) end
  if not x or not y then return nil end
  _, _, _, x, y = src:canProject(tg, x, y)
  -- Slight Hack(TM):  Record the center of the ball spells as dam.cx and
  -- dam.cy, so that the projector can reduce damage by distance from the
  -- center.  We don't do this for dam2, because cloud damage isn't
  -- supposed to vary by distance from center.  (Which is why we clone()'d
  -- dam above if it's a table, BTW.)
  if type(dam) == 'number' then dam = { dam } end
  dam.cx, dam.cy = x, y
  local proj = spell.set_project({ grid=true, object=true }, dt, dam)
  local grids = src:project(tg, x, y, dt, dam)
  if particle then particle(dt, x, y, r) end
  spell.set_project(proj)
  local overlay = spell.cloud_overlay(dt)
  local e = game.level.map:addEffect(src, x, y, dur, dt2, dam2, r, 5, nil, overlay, nil, selffire, friendlyfire)
  e.name = t.name
  return { x=x, y=y }
end

-- Fires a wall spell from the specified source Actor using the specified
-- talent.
--   = 'range' can be omitted if the talent defines a suitable 'range'
--     field (cf. ActorTalents:getTalentRange()).
--   = 'dt' can be omitted if the talent defines a suitable 'damage_type'
--     field [in the same style as 'range' above].
--   = 'dt2' is the damage type of the cloud; if omitted, the damage type
--     of the initial beam will be used.
--   = 'dam2' is the wall damage per turn; if omitted, the beam damage
--     will be used.
--   = If x and y are omitted, the player will be prompted for a target (or
--     the NPC AI will choose one).
--   = If 'particle' is passed or the talent defines a suitable .particle()
--     method, it will be used to draw a particle for the initial beam part
--     of the spell.  The method should take parameters (dt, src_x, src_y,
--     dst_x, dst_y).
--     [cf. spell.simple_beam_particle()]
--   = 'dam' can be a table containing flags passed through to the default
--     projector (cf. notes on setDefaultProjector in data/damage_types.lua).
--   = Self-fire and friendly-fire can be controlled via dam.selffire resp.
--     dam.friendlyfire.  By default, selffire is false and friendlyfire is
--     nil.  [cf. get_selffire() above]
--   = By default, grids and floor objects will be affected by the intial
--     beam; this can be overridden by the damage type and the 'dam'
--     parameter as per spell.set_project() [q.v.]
-- If the spell was attempted, returns a table containing the targeted x
-- and y coordinates; returns nil if the player canceled during targeting.
function spell.wall_spell(src, t, dt, dam, dur, dt2, dam2, x, y, range, particle)
  local selffire, friendlyfire = get_selffire(dam, false, nil)
  range = range or src:getTalentRange(t)
  particle = particle or (t and t.particle)
  if not dt then
    dt = t.damage_type
    if type(dt) == 'function' then dt = dt(src, t) end
    if not dt then return nil end
  end
  dt2 = dt2 or dt
  dam2 = dam2 or dam
  local tg = { type='beam', range=range, talent=t, selffire=selffire, friendlyfire=friendlyfire }
  util.target_merge_damage_type(tg, dt)
  if not x or not y then x, y = src:getTarget(tg) end
  if not x or not y then return nil end
  _, _, _, x, y = src:canProject(tg, x, y)
  local p = type(dam) == 'table' and dam or {}
  local proj = spell.set_project({ grid=true, object=true }, dt, dam)
  local grids = src:project(tg, x, y, dt, dam)
  if particle then particle(dt, src.x, src.y, x, y) end
  spell.set_project(proj)
  local overlay = spell.cloud_overlay(dt)
  local e = game.level.map:addEffect(src, x, y, dur, dt2, dam2, 0, nil, grids, overlay, nil, selffire, friendlyfire)
  e.name = t.name
  return { x=x, y=y }
end

-- Fires a cone spell from the specified source Actor using the specified
-- talent.
--   = 'r' can be omitted if the talent defines a suitable 'radius' field
--     (cf. ActorTalents:getTalentRadius()).
--   = 'cone_angle' can be omitted to use the default value of 55°.
--   = 'dt' can be omitted if the talent defines a suitable 'damage_type'
--     field [in the same style as 'r' above].
--   = If x and y are omitted, the player will be prompted for a target (or
--     the NPC AI will choose one).
--   = 'dam' can be a table containing flags passed through to the default
--     projector (cf. notes on setDefaultProjector in data/damage_types.lua).
--   = Self-fire and friendly-fire can be controlled via dam.selffire resp.
--     dam.friendlyfire; by default, selffire is false and friendlyfire is
--     nil.  [cf. get_selffire() above]
--   = By default, grids and floor objects will be affected; this can
--     be overridden by the damage type and the 'dam' parameter as per
--     spell.set_project() [q.v.]
-- If the spell was attempted, returns a table containing the targeted x
-- and y coordinates; returns nil if the player canceled during targeting.
function spell.cone_spell(src, t, dt, dam, r, cone_angle, x, y)
  local selffire, friendlyfire = get_selffire(dam, false, nil)
  r = r or src:getTalentRadius(t)
  dt = dt or util.getval(t.damage_type, src, t)
  if not dt then return nil end
  local tg = { type='cone', range=0, radius=r, cone_angle=cone_angle, talent=t, selffire=selffire, friendlyfire=friendlyfire }
  util.target_merge_damage_type(tg, dt)
  if not x or not y then x, y = src:getTarget(tg) end
  if not x or not y then return nil end
  local proj = spell.set_project({ grid=true, object=true }, dt, dam)
  src:project(tg, x, y, dt, dam)
  spell.set_project(proj)
  return { x=x, y=y }
end

-- Quick Hack(TM) to convert "#RRGGBB#" or "#COLORNAME#" from DamageType
-- definition of provided damage type to R/G/B colors.
local function dt_to_rgb(dt)
  local def = DamageType:get(dt)
  local c = def and (def.particle_color or def.text_color)
  if not c then return { 0, 0, 0 } end
  local tstr = c:toTString()
  -- Apparently this results in { '', { 'color', 'COLOR' }, '' } or
  -- { '', { 'color', r, g, b }, '' }.
  if not tstr or not tstr[2] or tstr[2][1] ~= 'color' then
    return { 0, 0, 0 }
  end
  if #tstr[2] == 4 then
    return { tstr[2][2], tstr[2][3], tstr[2][4] }
  end
  if #tstr[2] == 2 then
    return colors.simple(colors[tstr[2][2]]) or { 0, 0, 0 }
  end
  return { 0, 0, 0 }
end

-- Places a map particle drawing a generic "bolt" of the provided damage
-- from the provided source x/y to the provided destination x/y.  Suitable
-- for use as the 'particle' parameter to spell.bolt_spell() or as the
-- .particle() method of a talent definition.
function spell.simple_bolt_particle(dt, src_x, src_y, dst_x, dst_y)
  local rgb = dt_to_rgb(dt)
  game.level.map:particleEmitter(src_x, src_y, 1, 'generic_bolt', { dx=dst_x-src_x, dy=dst_y-src_y, rr=rgb[1], gg=rgb[2], bb=rgb[3] })
end

-- Places a map particle drawing a generic "ball" of the provided damage
-- at the provided source x/y the the provided radius.  Suitable for use as
-- the 'particle' parameter to spell.bolt_spell() or as the .particle()
-- method of a talent definition.
function spell.simple_ball_particle(dt, src_x, src_y, r)
  local rgb = dt_to_rgb(dt)
  game.level.map:particleEmitter(src_x, src_y, r, 'generic_ball', { radius=r, rm=rgb[1], rM=rgb[1], gm=rgb[2], gM=rgb[2], bm=rgb[3], bM=rgb[3], am=80, aM=150 })
end

-- Places a map particle drawing a generic "blast" of the provided damage
-- at the provided source x/y the the provided radius.  Suitable for use as
-- the 'particle' parameter to spell.bolt_spell() or as the .particle()
-- method of a talent definition.
function spell.simple_blast_particle(dt, src_x, src_y, r)
  local rgb = dt_to_rgb(dt)
  game.level.map:particleEmitter(src_x, src_y, r, 'generic_blast', { radius=r, rm=rgb[1], rM=rgb[1], gm=rgb[2], gM=rgb[2], bm=rgb[3], bM=rgb[3], am=35, aM=90 })
end

-- Places a map particle drawing a generic "beam" of the provided damage
-- from the provided source x/y to the provided destination x/y.  Suitable
-- for use as the 'particle' parameter to spell.beam_spell() or as the
-- .particle() method of a talent definition.
function spell.simple_beam_particle(dt, src_x, src_y, dst_x, dst_y)
  local rgb = dt_to_rgb(dt)
  game.level.map:particleEmitter(src_x, src_y, 1, 'generic_beam', { dx=dst_x-src_x, dy=dst_y-src_y, rr=rgb[1], gg=rgb[2], bb=rgb[3] })
end


spell.geomancy = {
  floor_types = {
    [1] = { 'FIRE', 'SAND', 1},
    [2] = { 'FIRE', 'SHALLOW_LAVA', 8},
    [3] = { 'FIRE', 'DEEP_LAVA', 18},

    [4] = { 'WATER', 'SHALLOW_WATER', 1},
    [5] = { 'WATER', 'DEEP_WATER', 8},
    [6] = { 'WATER', 'ICE', 18},

    [7] = { 'EARTH', 'GRASS', 1},
    [8] = { 'EARTH', 'GRASS_WITH_FLOWERS', 8},
    [9] = { 'EARTH', 'DARK_PIT', 18},
  },
  wall_types = {
    [1] = { 'FIRE', 'SANDWALL', 1},

    [2] = { 'WATER', 'TREE', 1},
    [3] = { 'WATER', 'ICE_WALL', 12},

    [4] = { 'EARTH', 'WALL', 1},
  },
  can_tunnel = {
    WALL = true,
    MAGMA_VEIN = true,
    MAGMA_VEIN_WITH_TREASURE = true,
    QUARTZ_VEIN = true,
    QUARTZ_VEIN_WITH_TREASURE = true,
    TREE = true,
    DEAD_TREE = true,
    SANDWALL = true,
    SANDWALL_WITH_TREASURE = true,
    ICE_WALL = true,
  },
}

function spell.geomancy.random_grid_type(t)
  while true do
    local r = rng.table(t)
    if game.player:getSkillLevel(r[1]) >= r[3] and rng.percent(game.player:getSkillLevel(r[1], 100)) then
      return game.zone.grid_list[r[2]]
    end
  end
end

-- Returns a suitable random floor type for transformations caused by Call
-- the Elements 
function spell.geomancy.random_floor()
  return spell.geomancy.random_grid_type(spell.geomancy.floor_types)
end

function spell.geomancy.random_wall()
  return spell.geomancy.random_grid_type(spell.geomancy.wall_types)
end

-- cf. geomancy_dig() in lib/scpt/s_geom.lua
function spell.geomancy.tunnel(sx, sy, dx, dy, length)
  local x, y = sx + dx, sy + dy
  local map = game.level.map

  local function tag(x, y)
    return map:checkEntity(x, y, Map.TERRAIN, 'define_as')
  end

  for _ = 1, length do
    local t = map:checkEntity(x, y, Map.TERRAIN, 'define_as')
    if not spell.geomancy.can_tunnel[t] then break end
    for xx = x - 1, x + 1 do
      for yy = y - 1, y + 1 do
	t = map:checkEntity(xx, yy, Map.TERRAIN, 'define_as')
	if (yy ~= y or xx ~= x) and spell.geomancy.can_tunnel[t] then
	  map(xx, yy, Map.TERRAIN, spell.geomancy.random_wall())
	end
      end
    end
    x, y = x + dx, y + dy
  end

  x, y = x - dx, y - dy
  while y ~= sy or x ~= sx do
    map(x, y, Map.TERRAIN, spell.geomancy.random_floor())
    -- 20% chance to branch
    if rng.chance(5) then
      local ddx, ddy = util.rotate_dxy(dx, dy, rng.chance(2) and -1 or 1)
      spell.geomancy.tunnel(x, y, ddx, ddy, math.floor(length/3))
    end
    x, y = x - dx, y - dy
  end
end


spell.thaumaturgy = {
  destructive_damage_types = {
    DamageType.KILL_WALL,
    DamageType.KILL_WALL,
    DamageType.KILL_WALL,
    DamageType.STONE_WALL,
    DamageType.STONE_WALL,
    DamageType.STONE_WALL,
    DamageType.DESTRUCTION,
    DamageType.DESTRUCTION,
    DamageType.DESTRUCTION,
    DamageType.DESTRUCTION,
  },
  simple_attack_types = {
    DamageType.ARROW,
    DamageType.MISSILE,
    DamageType.MANA,
    DamageType.WATER,
    DamageType.PLASMA,
    DamageType.METEOR,
    DamageType.ICE,
    DamageType.GRAVITY,
    DamageType.INERTIA,
    DamageType.FORCE,
    DamageType.TIME,
    DamageType.ACID,
    DamageType.ELEC,
    DamageType.FIRE,
    DamageType.COLD,
    DamageType.POISON,
    DamageType.LITE,
    DamageType.DARK,
    DamageType.CONFUSION,
    DamageType.SOUND,
    DamageType.SHARDS,
    DamageType.NEXUS,
    DamageType.NETHER,
    DamageType.CHAOS,
    DamageType.DISENCHANT,
  },
  type_names = {
    bolt	= 'Bolt',
    ball	= 'Ball',
    beam	= 'Beam',
    blast	= 'Blast',
    meteor	= 'Area',
    view	= 'View',
  },
  display_entities = {
    bolt	= Entity.new { image='talents/thaumaturgy_bolt.png' },
    ball	= Entity.new { image='talents/thaumaturgy_ball.png' },
    beam	= Entity.new { image='talents/thaumaturgy_beam.png' },
    blast	= Entity.new { image='talents/thaumaturgy_blast.png' },
    meteor	= Entity.new { image='talents/thaumaturgy_meteor.png' },
    view	= Entity.new { image='talents/thaumaturgy_view.png' },
  },
}

-- Creates a thaumaturgy spell of the specified level.
-- [cf. src/spells1.c:generate_spell()]
function spell.thaumaturgy.create(lev)
  local power = rng.range(0, 14)
  local dice = math.max(math.floor(lev/5) + power, 1)
  local sides = math.max(lev + power, 5)
  local mana = math.max(1, lev + math.floor((lev * power)/15))

  local sp = { level=lev, mana=mana }
  local dt_grp = 'simple_attack_types'

  local chance = rng.range(1, 100)
  -- Slight Hack(TM):  Always bolts for level-1 spells.
  if chance < 25 or lev == 1 then
    -- Swap dice and sides for better damage.
    sp.type, sp.dice, sp.sides = 'bolt', sides, dice
  elseif chance < 50 then
    sp.type, sp.dice, sp.sides = 'beam', dice, sides
  elseif chance < 76 then
    sp.type, sp.dice, sp.sides = 'ball', dice, sides
    sp.radius = math.floor(dice/3)
    -- Slight Hack(TM):  Radius-0 ball spells appear to be silently
    -- converted to bolt spells.
    if sp.radius == 0 then
      sp.radius = nil
      sp.type = 'bolt'
    end
  elseif chance < 83 then
    sp.type, sp.dice, sp.sides = 'blast', dice, sides
    sp.radius = math.floor(sides/3)
    dt_grp = 'destructive_damage_types'
  elseif chance < 90 then
    sp.type = 'meteor'
    -- Reduce damage per shot.
    sp.dice = math.max(1, math.floor(dice/5))
    sp.sides = math.max(1, math.floor(sides/5))
    sp.n_balls_base = math.max(4, math.floor(sides/3))
    if rng.percent(25) then dt_grp = 'destructive_damage_types' end
  else
    -- Line-of-sight spells do less damage.
    sp.type, sp.dice, sp.sides = 'view', dice, math.floor(sides/2)
  end

  sp.damtype = rng.table(spell.thaumaturgy[dt_grp])
  sp.name = spell.thaumaturgy.name(sp)
  return sp
end

-- Casts the provided thaumaturgy spell (hence the name).
function spell.thaumaturgy.cast(actor, sp)
  -- Fake up a "talent" to do pre-use check on.
  local t = spell.thaumaturgy.stub_talent(sp)
  if not actor:preUseTalent(t) then return nil end

  local co = coroutine.create(function()
    local dam = rng.dice(sp.dice, sp.sides)
    local ret
    if sp.type == 'bolt' then
      ret = spell.bolt_spell(actor, t, sp.damtype, dam, nil, nil, nil, spell.simple_bolt_particle)
    elseif sp.type == 'beam' then
      ret = spell.beam_spell(actor, t, sp.damtype, dam, nil, nil, nil, spell.simple_beam_particle)
    elseif sp.type == 'ball' then
      ret = spell.ball_spell(actor, t, sp.damtype, dam, sp.radius, nil, nil, nil, spell.simple_ball_particle)
    elseif sp.type == 'blast' then
      ret = spell.ball_spell(actor, t, sp.damtype, dam, sp.radius, actor.x, actor.y, nil, spell.simple_blast_particle)
    elseif sp.type == 'meteor' then
      ret = spell.meteor_spell(actor, t, sp.damtype, dam, sp.n_balls_base, spell.simple_ball_particle)
    elseif sp.type == 'view' then
      ret = spell.los_spell(actor, t, sp.damtype, dam)
    else
      print('[THAUMATURGY] ??? unknown spell type '..tostring(sp.type))
      return nil
    end
    if ret then actor:postUseTalent(t, ret) end
    return ret
  end)
  local ok, err = coroutine.resume(co)
  if not ok and err then print(debug.traceback(co)) error(err) end
  if ok then return err else return false end
end

-- Constructs and returns a simple default name for the provided
-- thaumaturgy spell.
function spell.thaumaturgy.name(sp)
  local dt = DamageType:get(sp.damtype)
  local dam = ('%dd%d'):format(sp.dice, sp.sides)
  if sp.type == 'ball' then
    dam = dam .. ('r%d'):format(sp.radius)
  end
  return ('%s: %s %s'):format(spell.thaumaturgy.type_names[sp.type], dam, dt.name)
end

-- Creates and returns a vaguely talent-definition-shaped table that can be
-- successfully passed to some methods that expect talent definitions.  So
-- far, Actor:getTalentFullDescription(), Actor:preUseTalent() and
-- Actor:postUseTalent() are known to work with it.
function spell.thaumaturgy.stub_talent(sp)
  local info, range, requires_target
  local dtf = util.colorize_damtype(sp.damtype)
  if sp.type == 'bolt' or sp.type == 'beam' then
    info = ('Fires a %s of %s doing %dd%d damage.'):format(sp.type, dtf, sp.dice, sp.sides)
    range = 18
    requires_target = true
  elseif sp.type == 'ball' then
    info = ('Fires a radius-%d ball of %s doing %dd%d damage.'):format(sp.radius, dtf, sp.dice, sp.sides)
    range = 18
    requires_target = true
  elseif sp.type == 'blast' then
    info = ('Does %dd%d %s damage in a radius of %d around the caster.'):format(sp.dice, sp.sides, dtf, sp.radius)
  elseif sp.type == 'meteor' then
    local nb = sp.n_balls_base
    local ret = [[Calls down a meteor storm around the caster.  %d-%d meteors will fall on random open grids in line of sight of the caster, within radius 5; each will do %dd%d %s damage in a radius-2 area.]]
    info = ret:format(nb+1, nb*2, sp.dice, sp.sides, dtf)
  elseif sp.type == 'view' then
    info = ('Does %dd%d %s damage to all monsters in light of sight of the caster.'):format(sp.dice, sp.sides, dtf)
  else
    return ([[#RED#ERROR:#LAST# Unrecognized thaumaturgy spell type '%s']]):format(sp.type)
  end

  return {
    name = sp.name,
    display_entity = spell.thaumaturgy.display_entities[sp.type],
    mode = 'activated',
    is_spell = true,
    skill = 'THAUMATURGY',
    skill_level = sp.level,
    mana = sp.mana,
    base_fail = sp.level + 25,
    stat = 'INT',
    range = range,
    points = 0,
    info = info,
  }
end


-- Assorted methods used by NPC talents.
spell.npc = {}

function spell.npc.summon_action(self, t)
  local tg = { range=self:getTalentRange(t), selffire=false, talent=t }
  local tx, ty = self:getTarget(tg)
  if not tx or not ty then return nil end

  local rlev = math.max(self.level, 1)
  local any = spell.summon(self, tx, ty, t.filter, t.params.num, rlev)
  if t.filter2 and t.params2 then
    -- Annoying Hack(TM):  Some of the really nasty summon spells do
    -- double summons.
    any = spell.summon(self, tx, ty, t.filter2, t.params2.num, rlev) or any
  end

  local msg = '@Source@ tries to cast '..t.name..', but fails.'
  if any then
    local msg = t.params and util.getval(t.params.msg, self) or '@Source@ casts '..t.name..'.'
  end
  game.logSeen(self, self:convertTalentMessage(self, msg))
  return true
end

function spell.npc.summon_info(self, t)
  return 'Summons '..t.params.info_desc..' near the target.'
end

function spell.npc.breathe(self, t)
  local r = self:has('POWERFUL') and 3 or 2
  local dam = math.min(self.life/t.params.div, t.params.max)
  return spell.ball_spell(self, t, t.params.type, dam, r, nil, nil, nil, spell.simple_ball_particle)
end

function spell.npc.breath_info(self, t)
  local r = self:has('POWERFUL') and 3 or 2
  local dam = math.min(self.life/t.params.div, t.params.max)
  local dt = DamageType:get(t.params.type)
  local desc = t.params.desc or dt.name or 'an unknown substance'
  local tmpl = [[Breathes %s for %d damage in a radius of %d]]
  return tmpl:format(desc, dam, r)
end

function spell.npc.bolt_spell(self, t)
  local dam = t.bolt_damage(self)
  return spell.bolt_spell(self, t, t.params.type, dam, nil, nil, nil, spell.simple_bolt_particle)
end

function spell.npc.ball_spell(self, t)
  local dam = t.ball_damage(self)
  return spell.ball_spell(self, t, t.params.type, dam, t.params.r, nil, nil, nil, spell.simple_ball_particle)
end

function spell.npc.get_spell_target(self, t)
  local tg = { type='hit', range=self:getTalentRange(t), talent=t }
  local x, y = self:getTarget(tg)
  local target = game.level.map(x, y, Map.ACTOR)
  return x, y, target
end

function spell.npc.cause_wounds(self, t)
  local x, y, target = spell.npc.get_spell_target(self, t)
  if not target then return end

  local resist_fn = t.target_resist or spell.resists
  if resist_fn(self, target) then
    game.logSeen(target, self:convertTalentMessage(target, '@Target@ resists!'))
  else
    if target == game.player then
      if t.params and t.params.plyr_hit_msg then
	game.logSeen(target, self:convertTalentMessage(target, t.params.plyr_hit_msg))
      end
      if t.curse_chance then
	spell.curse_equipment(target, t.curse_chance[1], t.curse_chance[2])
      end
    end
    target:takeHit(t.spell_damage(target), self, 'cursed to death')
    if t.cut_duration then
      target:setEffect(self.EFF_CUT, t.cut_duration(), {merge='combine', src=self})
    end
  end
  return true
end

function spell.npc.effect_spell(self, t)
  local x, y, target = spell.npc.get_spell_target(self, t)
  if not target then return end

  local resist = false
  if t.params.resist_flag then
    resist = resist or target:has(t.params.resist_flag)
  end
  if t.params.resist_flags then
    for _, flag in pairs(t.params.resist_flags) do
      resist = resist or target:has(flag)
    end
  end
  if t.params.resist_dam then
    resist = resist or target:resistDam(t.params.resist_dam)
  end
  if t.params.resist_unique then
    resist = resist or target.unique
  end

  if resist then
    game.logSeen(target, self:convertTalentMessage(target, t.params.resist_msg or '@Target@ resists!'))
  elseif spell.resists(self, target) then
    game.logSeen(target, self:convertTalentMessage(target, t.params.save_msg or t.params.resist_msg or '@Target@ resists!'))
  else
    target:setEffect(self[t.params.eff_id], t.params.duration(), {merge='combine'})
  end
  return true
end
